/*
    MPEG Maaate: An Australian MPEG audio analysis toolkit
    Copyright (C) 2000 Commonwealth Scientific and Industrial Research Organisation
    (CSIRO), Australia.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#ifdef DRIVER_OSS
#include <sys/soundcard.h>
#define DEV_DSP "/dev/dsp"
#endif

#ifdef DRIVER_SOLARIS_AUDIO
#include <sys/audioio.h>
#define DEV_DSP "/dev/audio"
#endif

#include "SOUNDfile.H"

#define BUFFERSIZE 16500
#define PBUF_SIZE  256
static int dev_dsp = -1;

// prints help information about parameter usage
void 
PrintUsage(char *prog) {
    fprintf(stderr, "\nUsage:\n");
    fprintf(stderr, "%s [-f <outfile.raw> -v -h] <mpeg-audio-file>\n", prog);
    fprintf(stderr, "    -f <outfile.raw> : decode to pcm file only\n");
    fprintf(stderr, "    -v               : verbose\n");
    fprintf(stderr, "    -h               : instruction\n");
    fprintf(stderr, "    -s <start_time>  : start at time\n");
    fprintf(stderr, "    -e <end_time>  : end at time\n");
    exit(1);
}

static int
open_dev_dsp (void)
{
#if defined(DRIVER_OSS) || defined(DRIVER_SOLARIS_AUDIO)
  if((dev_dsp = open(DEV_DSP, O_WRONLY|O_NDELAY, 0)) == -1) {
    perror ("MaaatePlay: unable to open device " DEV_DSP);
    return -1; /* XXX: Flag error */
  }

  return dev_dsp;
#else
  fprintf(stderr, "Warning: No audio device configured\n");
  return -1;
#endif
}

static void
setup_dev_dsp (int stereo, int frequency)
{
  int format;

#ifdef DRIVER_OSS
  // MPEG pcm samples are signed 16 bit
  format=AFMT_S16_LE;

  if (ioctl(dev_dsp, SNDCTL_DSP_SETFMT, &format) == -1) {
    perror("MaaatePlay: Unable to set format");
    exit(-1);
  }

  if(ioctl(dev_dsp, SNDCTL_DSP_STEREO, &stereo) == -1 ) {
    perror("MaaatePlay: Unable to set channels");
  }

  if(ioctl(dev_dsp, SNDCTL_DSP_SPEED, &frequency) == -1 ) {
    perror("MaaatePlay: Unable to set playback frequency");
  }

#elif defined(DRIVER_SOLARIS_AUDIO)

  audio_info_t info;
  AUDIO_INITINFO(&info);
  info.play.precision = 16;	/* cs4231 doesn't handle 16-bit linear PCM */
  info.play.encoding = AUDIO_ENCODING_LINEAR;
  info.play.channels = stereo+1;
  info.play.sample_rate = frequency;
  if(ioctl(dev_dsp, AUDIO_SETINFO, &info) < 0)
      perror("MaaatePlay: Unable to configure audio device");

#endif
}

static void
flush_dev_dsp (void)
{
#ifdef DRIVER_OSS
  if(ioctl (dev_dsp, SNDCTL_DSP_SYNC) == -1) {
    perror ("MaaatePlay: Unable to flush " DEV_DSP);
  }
  //  if(ioctl (dev_dsp, SNDCTL_DSP_POST) == -1) {
  //    perror ("MaaatePlay: Unable to flush " DEV_DSP);
  //  }
#endif

#ifdef DRIVER_SOLARIS_AUDIO
  if(ioctl(dev_dsp, AUDIO_DRAIN, 0) == -1)
      perror("MaaatePlay: Unable to flush " DEV_DSP);
#endif
}


static void
close_dev_dsp (void)
{
  close (dev_dsp);
  dev_dsp = -1;
}

static void
play(short * data, long length)
{
  fd_set fds;
  ssize_t n;
  int written, towrite;

  unsigned char * d = (unsigned char *) data;
  long bytes = length * sizeof(short);
  
  written = 0;
  while (written < bytes) {
    FD_ZERO (&fds);
    FD_SET (dev_dsp, &fds);
    
    if (select (dev_dsp+1, NULL, &fds, NULL, NULL) == 0);
    
    towrite = min ((long) PBUF_SIZE, (bytes - written));
    
    n = write (dev_dsp, d, towrite);
    
    written += n ;
    d += n;
  }
}


int 
main(int argc, char **argv) {
  
  bool verbose = false;
  float start_time = 0.0;
  float end_time = 10000000.0;
  char * mpegfile;
  char * outfile;
  FILE * outpcm = NULL;
  
  // check no. of args
  if (argc < 2) {
    PrintUsage(argv[0]);
  }
  
  // go through parameters and check values
  for (int i=1; i<argc; i++) {
    if (argv[i][0] == '-' && argv[i][1]) {
      switch (argv[i][1]) {
      case 'h': // help
	PrintUsage(argv[0]);
	break;
      case 'v': // help
	verbose=true;
	break;
      case 'f': // output file
	if ((i+1<argc) && isalpha(argv[i+1][0]) 
	    && (argv[i+1][0] != '-')) {
	  outfile = argv[i+1];
	  outpcm = fopen(outfile,"wb");
	}
	break;
      case 's': // start time
	if ((i+1<argc) && !isalpha(argv[i+1][0])) {
	  start_time = atof(argv[++i]);
	}
	break;
      case 'e': // end time
	if ((i+1<argc) && !isalpha(argv[i+1][0])) {
	  end_time = atof(argv[++i]);
	}
	break;
      } // switch
    } else {
      mpegfile = argv[i];
    }
  }


  if (start_time > end_time ) {
    perror ("Maaateplay: That wouldn`t play anything\n");
    exit(1);
  }
    
  // create output buffer
  short buffer[BUFFERSIZE];
  long nr_samples = BUFFERSIZE;

  
  // open SOUND file
  SOUNDfile  *dec = new SOUNDfile(argv[argc-1]);

  if (dec->file_type() == NOFILE) {
    cerr << "Maaateplay: Sorry can`t play that file." << endl;
    exit(1);
  }

  if (start_time >= dec->file_duration() ) {
    cerr << "Maaateplay: That won`t play anything: start_time > file duration" << endl;
    exit(1);
  }

  // open and setup device
  if (outpcm == NULL) {
    if (open_dev_dsp () < 0) {
      /* error */
      exit (1);
    }
    setup_dev_dsp (dec->channels()-1, (int) (dec->sampling_rate()*1000));
  }
  
  //convert time into window nb
  long start = dec->time2window( start_time);

  long end = dec->time2window( end_time);

  // go to start window within soundfile
  if (start > 0) {
    if (!dec->seek_window(start)) {
      cerr <<"MaaatePlay: Error when positioning" << endl;
      dec->seek_window(0);
      start = 0;
    }
  }

  long nb_win = nr_samples / (dec->channels() * dec->timeticks(PCM) * dec->nb_subbands(PCM));
  if ( dec->at_window() + nb_win > end ) {
    nb_win = end - dec->at_window() + 1;
  }
  nr_samples = dec->decode (buffer, nb_win, STEREO);


  if (outpcm != NULL) {
   
    while (nr_samples != 0 && dec->at_window() <= end) {
      fwrite(buffer, sizeof(short), nr_samples, outpcm);  
      nr_samples = dec->decode (buffer, nb_win, STEREO);
    }
  } else {
   
    while (nr_samples != 0  && dec->at_window() <= end) {
      play (buffer, nr_samples);
      if ( dec->at_window() + nb_win > end ) {
	nb_win = end - dec->at_window() + 1;
      }

      nr_samples = dec->decode (buffer, nb_win, STEREO);
    }
  }

  if (verbose)
    printf("Decoded.\n");
  


  // clean up
  if (outpcm == NULL) {
    // flush & close device
    flush_dev_dsp ();
    close_dev_dsp ();
  } else {
    // close output sound file
    fclose (outpcm);
    cout << "PCM file written (16 bit linear, "
	 << dec->sampling_rate() << " kHz samplingrate, "
	 << dec->channels() << " channels): "
	 << outfile << endl;
  }

  delete dec;
  
  exit (0);
}
