/*
    MPEG Maaate: An Australian MPEG audio analysis toolkit
    Copyright (C) 2000 Commonwealth Scientific and Industrial Research Organisation
    (CSIRO), Australia.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "maaateM_brief.H"

#include "SOUNDfile.H"
#include "plugins.H"

using namespace std;

// parameters to control what is to be extracted
static float from          = 0.0;      // start at second 0.0
static float to            = FLT_MAX;  // end at end of file
static bool  verbose       = false;
static short fromSb        = 0;        // first subband to be extracted
static short toSb          = 575;      // last subband to be extracted
static float thresh        = 0.07;     // threshold between [0;1]
static float minDur        = 0.1;      // minimum duration [sec]
static float maxInterrupt  = 0.0;      // mamixum tolerated disruption [sec]
static float releaseTime   = 0.01;     // release time for silence start [sec]
static float onsetTime     = 0.01;     // onset time for silence end [sec]
static int   bins          = 10;       // number of bins for histogram
static float starthisto    = FLT_MAX;  // start of the histogram values
static float endhisto      = DBL_MIN;  // end of the histogram values
static float tolerance     = 0.0;      // time identity tolerance for compare
static bool  wantSilence   = false;    // want silence?
static bool  wantNoise     = false;    // want noise?
static bool  wantBgLevel   = false;    // want background noise level?
static bool  wantSumSCF    = false;    // want sum of scalefactors?
static bool  wantCentroid  = false;    // want spectral centroid?
static bool  wantSSCFhisto = false;    // want histogram of SumSCF?
static bool  wantSBnrj     = false;    // want normalised subband energies?
static bool  wantSBmean    = false;    // want subband mean?
static bool  wantSBrms     = false;    // want subband rms?
static bool  wantSBv       = false;    // want subband value?
static bool  wantSBvmean   = false;    // want subband value mean?
static bool  wantSBvrms    = false;    // want subband value rms?
static bool  wantSBvnrj    = false;    // want subband value normalised nrj?
static bool  wantSBscf     = false;    // want subband scalefactor?
static bool  wantsignrj    = false;    // want signal energy?
static bool  wantsigsb     = false;    // want significant subbands?
static bool  wantbandnrj   = false;    // want band signal energy?
static bool  wantsigmag    = false;    // want signal magnitude?
static bool  wantpauserate = false;    // want pause rate?
static bool  wantbandratio = false;    // want band nrj ratio?
static bool  wantrolloff   = false;    // want roll off?
static bool  wantbw        = false;    // want bandwidth?
static bool  wantbdw       = false;    // want signal bandwidth?
static bool  wantspflux    = false;    // want spectral flux?
static bool  wantcm        = false;    // want central moment?
static bool  wantlownrj    = false;    // want lownrj?
static bool  wantvariance  = false;    // want variance?
static bool	 wantmatoutput = false;	   // want output in MATLAB mat-file
static char *matfilename   = NULL;	   // MATLAB mat-file
static char *filename      = NULL;     // mpeg-audio-file

Plugins * plugins;

//----------------------------------
void
PrintUsage(char *prog) {
    cerr << "\nUsage:" << endl;
    cerr << prog << " [Options] <mpeg-audio-file>" << endl;
    cerr << "Default Options: -S 0.0 -E eof" << endl;
    cerr << "                 (i.e. analyse whole file)" << endl;
    cerr << endl;
    cerr << "Possible Options:" << endl;
    cerr << "-h            help = prints this information" << endl;
    cerr << "-S <from>     the second (float) to start analysing from" << endl;
    cerr << "              Default: 0.0" << endl;
    cerr << "-E <to>       the second (float) to stop analysing at" << endl;
    cerr << "              Default: end of file" << endl;
    cerr << "-V            verbose" << endl;
    cerr << "-B <from> <to> subband range (ints) that gets extracted" << endl;
    cerr << "              (only required with -X sscf)" << endl;
    cerr << "              Default: 0 31 (Layers I, II)" << endl;
    cerr << "                       0 575 (Layer III)" << endl;
    cerr << "-T <thresh>   threshold value between [0;1]" << endl;
    cerr << "              Default: 0.07" << endl;
    cerr << "-M <minDur>   minimum duration [sec] of specified content" << endl;
    cerr << "              Default: 0.1" << endl;
    cerr << "-N <maxInter> mamimum tolerated disruption [sec]" << endl;
    cerr << "              Default: 0.0" << endl;
    cerr << "-R <releaseT> release time [sec] delaying start of silence" << endl;
    cerr << "-O <onsetT>   onset time [sec] earlying end of silence" << endl;
    cerr << "              Default: 0.01 (must be < minDur)" << endl;
    cerr << "-Z <bins>     number of bins for histogram" << endl;
    cerr << "              Default: 10" << endl;
    cerr << "-Y <from> <to> start and end of histogram values (float)" << endl;
    cerr << "              Default: -1 -1 (meaning: minvalue maxvalue)" << endl;
	cerr << "-F <filename> filename and path of outputfile in MATLAB binary mat-format" <<endl;
	cerr << "              Default: default.mat" <<endl;
    cerr << "-X [content]  specifies content to extract from audio file" <<endl;
    cerr << "              Possible content types:" << endl;
    cerr << "    silence   -T <thresh> -M <minDur> -N <maxInter> -O <onsetT> -R <releaseT>" << endl;
    cerr << "              Temporal Segmentation: silence segments based on sumscf" << endl;
    cerr << "    noise     -T <thresh> -M <minDur> -N <maxInter> -O <onsetT> -R <releaseT>" << endl;
    cerr << "              Temporal Segmentation: noise segments based on sumscf" << endl;
    cerr << "    bglevel   -M <minDur> -N <maxInter> -O <onsetT> -R <releaseT>" << endl;
    cerr << "    sumscf    -B <fromsb> <tosb>" << endl;
    cerr << "              1D Curve: sum of scalefactors as loudness approximation" << endl;
    cerr << "    centroid  -B <fromsb> <tosb>" << endl;
    cerr << "              1D Curve: spectral centroid" << endl;
    cerr << "    histo     -Z <bins> -Y <fromVal> <toVal>" << endl;
    cerr << "              Vector: loudness distribution for whole duration based on sumscf" << endl;
    cerr << "    sbnrj     -B <fromsb> <tosb>" << endl;
    cerr << "              Vector: normalised subband energies in each selected subbands" << endl;
    cerr << "    sbmean    -B <fromsb> <tosb>" << endl;
    cerr << "              Vector: subband mean over one window in each selected subbands" << endl;
    cerr << "    sbrms     -B <fromsb> <tosb>" << endl;
    cerr << "              Vector: subband rms over one window in each selected subbands" << endl;
    cerr << "    sbscf     -B <fromsb> <tosb>" << endl;
    cerr << "              Vector: subband scalefactor over one window in each selected subbands" << endl;
    cerr << "    sbv       -B <fromsb> <tosb>" << endl;
    cerr << "              Vector: subband values in each selected subbands" << endl;
    cerr << "    sbvmean   -B <fromsb> <tosb>" << endl;
    cerr << "              Vector: subband values mean in each selected subbands" << endl;
    cerr << "    sbvrms    -B <fromsb> <tosb>" << endl;
    cerr << "              Vector: subband values mean in each selected subbands" << endl;
    cerr << "    sbvnrj    -B <fromsb> <tosb>" << endl;
    cerr << "              Vector: subband values normalised energies in each selected subbands" << endl;
    cerr << "    signrj    -Z <window_number>" << endl;
    cerr << "              Curve: short time energy function" << endl;
    cerr << "              Window: 0: square, 1: Hamming, 2: Welch, 3: Bartlett(default)" << endl;
    cerr << "    bandnrj   -B <fromsb> <tosb> -Z <window_number>" << endl;
    cerr << "              Curve: short time band energy function" << endl;
    cerr << "              Window: 0: square, 1: Hamming, 2: Welch, 3: Bartlett(default)" << endl;
    cerr << "    sigsb     -T <thresh> " << endl;
    cerr << "              Curve: number of significant subbands function" << endl;
    cerr << "              Window: 0: square, 1: Hamming, 2: Welch, 3: Bartlett(default)" << endl;
    cerr << "    sigmag    -Z <window_number>" << endl;
    cerr << "              Curve: short time magnitude function" << endl;
    cerr << "              Window: 0: square, 1: Hamming, 2: Welch, 3: Bartlett(default)" << endl;
    cerr << "    pauser    -T <thresh> -M <duration>" << endl;
    cerr << "              Curve: values of pause rate for each segment of duration seconds" << endl;
    cerr << "    bandnrj   -B <subband boundary> <window_number>" << endl;
    cerr << "              Curve: values of band nrj ratio" << endl;
    cerr << "    rolloff  " << endl;
    cerr << "              Curve: values of roll of, resolution: granule" << endl;
    cerr << "    bdwidth   -T <thresh>" << endl;
    cerr << "              Curves: values of bandwidth, fcmin, fcmax  resolution granule" << endl;
    cerr << "    sbdw      -T <thresh>" << endl;
    cerr << "              Curves: values of bandwidth, fcmin, fcmax  resolution granule" << endl;
    cerr << "    spflux   " << endl;
    cerr << "              Curve: spectral flux between 2 successive frames" << endl;
    cerr << "    cmoment   -B <fromsb> <tosb> -Z <k_pow> -M <duration>" << endl;
    cerr << "              Vector: central moments for each selected subbands" << endl;
    cerr << "    lownrj    -M <duration>" << endl;
    cerr << "              Curve: number of low energy windows" << endl;
    cerr << "    variance  -M <duration>" << endl;
    cerr << "              Curve: variance of signal nrj" << endl;
    exit (1);
}

//----------------------------------
void
ParseArguments (int argc, char **argv) {
    // go through parameters and check values
    for (int i=1; i<argc; i++) {
        if (argv[i][0] == '-' && argv[i][1]) {
            switch (argv[i][1]) {
            case 'H': case 'h': // help
                PrintUsage(argv[0]);
                break;
            case 'S': case 's': // second to start from
                if ((i+1<argc) && !isalpha(argv[i+1][0])) {
                    from = atof(argv[++i]);
                }
                break;
            case 'E': case 'e': // second to end at
                if ((i+1<argc) && !isalpha(argv[i+1][0])) {
                    to = atof(argv[++i]);
                }
                break;
            case 'V': case 'v':// verbose
                verbose = true;
                break;
            case 'B': case 'b': // subband range
                if ((i+2<argc) && !isalpha(argv[i+1][0]) 
                    && !isalpha(argv[i+2][0])) {
                    fromSb = atoi(argv[++i]);
                    toSb   = atoi(argv[++i]);
                } else {
		  cerr << argv[0] << ": -B option needs two integers to follow"
                         << endl;
                    exit(1);
                }
                break;
            case 'T': case 't': // threshold
	      if ((i+1<argc) && !isalpha(argv[i+1][0])) {
		thresh = atof(argv[++i]);
	      }
	      break;
            case 'M': case 'm': // minimum duration
                if ((i+1<argc) && !isalpha(argv[i+1][0])) {
                    minDur = atof(argv[++i]);
                }
                break;
            case 'N': case 'n': // maximum interrupt
                if ((i+1<argc) && !isalpha(argv[i+1][0])) {
                    maxInterrupt = atof(argv[++i]);
                }
                break;
	    case 'R': case 'r': // release time
	      if ((i+1<argc) && !isalpha(argv[i+1][0])) {
		releaseTime = atof (argv[++i]);
	      }
	      break;
	    case 'O': case 'o': // onset time
	      if ((i+1<argc) && !isalpha(argv[i+1][0])) {
		onsetTime = atof (argv[++i]);
	      }
	      break;
	    case 'Z': case 'z': // no bins
	      if ((i+1<argc) && !isalpha(argv[i+1][0])) {
			bins = atoi(argv[++i]);
	      }
	      break;
        case 'Y': case 'y': // histogram range
          if ((i+2<argc) && !isalpha(argv[i+1][0]) && !isalpha(argv[i+2][0])) {
            starthisto = atof(argv[++i]);
            endhisto   = atof(argv[++i]);
          } 
		  else {
		    cerr << argv[0] << ": -Y option needs two floats to follow" << endl;
            exit(1);
          }
          break;
	    case 'F': case 'f': // MATLAB output
          if ((i+1 == argc) || !isalpha(argv[i+1][0])) {
		    cerr << argv[0] << ": -X option needs string-argument" << endl;
		    exit(1);
          } 
		  else {
			wantmatoutput = true;
			matfilename = argv[i+1];
			i++;
	      }
			break;
	    case 'X': case 'x': // extract content
                if ((i+1 == argc) || !isalpha(argv[i+1][0])) {
		  cerr << argv[0] << ": -X option needs string-argument"
		       << endl;
		  exit(1);
                } else {
		  if (!strncmp(argv[i+1],"silence",3)) {
		    wantSilence = true; i++;
		  } else if (!strncmp(argv[i+1],"noise",3)) {
		    wantNoise = true; i++;
		  } else if (!strncmp(argv[i+1],"centroid",3)) {
		    wantCentroid = true; i++;
		  } else if (!strncmp(argv[i+1],"bglevel",3)) {
		    wantBgLevel = true; i++;
		  } else if (!strncmp(argv[i+1],"sumscf",3)) {
		    wantSumSCF = true; i++;
		  } else if (!strncmp(argv[i+1],"histo",3)) {
		    wantSSCFhisto = true; i++;
		  } else if (!strncmp(argv[i+1],"sbnrj",3)) {
		    wantSBnrj = true; i++;
		  } else if (!strncmp(argv[i+1],"sbmean",3)) {
		    wantSBmean = true; i++;
		  } else if (!strncmp(argv[i+1],"sbrms",3)) {
		    wantSBrms = true; i++;
		  } else if (!strncmp(argv[i+1],"sbscf",3)) {
		    wantSBscf = true; i++;
		  } else if (!strncmp(argv[i+1],"sbvmean",4)) {
		    wantSBvmean = true; i++;
		  } else if (!strncmp(argv[i+1],"sbvrms",4)) {
		    wantSBvrms = true; i++;
		  } else if (!strncmp(argv[i+1],"sbvnrj",4)) {
		    wantSBvnrj = true; i++;
		  } else if (!strncmp(argv[i+1],"sbv",3)) {
		    wantSBv = true; i++;
		  } else if (!strncmp(argv[i+1],"signrj",4)) {
		    wantsignrj = true; i++;
		  } else if (!strncmp(argv[i+1],"sigsb",4)) {
		    wantsigsb = true; i++;
		  } else if (!strncmp(argv[i+1],"bandnrj",5)) {
		    wantbandnrj = true; i++;
		  } else if (!strncmp(argv[i+1],"sigmag",4)) {
		    wantsigmag = true; i++;
		  } else if (!strncmp(argv[i+1],"pauser",3)) {
		    wantpauserate = true; i++;
		  } else if (!strncmp(argv[i+1],"bandnrj",3)) {
		    wantbandratio = true; i++;
		  } else if (!strncmp(argv[i+1],"rolloff",3)) {
		    wantrolloff = true; i++;
		  } else if (!strncmp(argv[i+1],"bdwidth",3)) {
		    wantbw = true; i++;
		  } else if (!strncmp(argv[i+1],"sbdw",3)) {
		    wantbdw = true; i++;
		  } else if (!strncmp(argv[i+1],"spflux",3)) {
		    wantspflux = true; i++;
		  } else if (!strncmp(argv[i+1],"cmoment",3)) {
		    wantcm = true; i++;
		  } else if (!strncmp(argv[i+1],"lownrj",3)) {
		    wantlownrj = true; i++;
		  } else if (!strncmp(argv[i+1],"variance",3)) {
		    wantvariance = true; i++;
		  } else if (argv[i+1][0] != '-') {
		    cerr << argv[0] << ": -X unknown string-argument "
			 << argv[i+1] << endl;
		    exit(1);
		  }
                }
                break;
            } // switch
        } else {
            filename = argv[i];
        }
    }
}
		
//----------------------------------
void
CheckArguments(char *prgname) {
    if (filename == NULL) {
      cerr << prgname << ": Mpeg audio filename missing" << endl;
      exit(1);
    }
    if (from < 0.0)    from = 0.0;
    if (to   < from)   to   = from;
    if (fromSb   < 0)  fromSb   = 0;
    if (toSb < fromSb && !(wantbandratio)) toSb     = fromSb;
    if (thresh < 0.0 && !(wantbdw))  thresh = 0.0;
    if (thresh > 1.0)  thresh = 1.0;
    if (releaseTime > minDur) releaseTime = 0.0;
    if (onsetTime > minDur) onsetTime = 0.0;
    if (bins <= 0 && !(wantsignrj || wantcm || wantbandnrj || wantvariance))     bins = 10;
    if (tolerance < 0) tolerance = -tolerance;

#ifdef DEBUG
    cout << "Parameter Settings:"             << endl;
    cout << "From="          << from          << endl;
    cout << "To="            << to            << endl;
    cout << "Verbose="       << verbose       << endl;
    cout << "fromSb="        << fromSb        << endl;
    cout << "toSb="          << toSb          << endl;
    cout << "thresh="        << thresh        << endl;
    cout << "minDur="        << minDur        << endl;
    cout << "maxInterrupt="  << maxInterrupt  << endl;
    cout << "releaseTime="   << releaseTime   << endl;
    cout << "onsetTime="     << onsetTime     << endl;
    cout << "bins="          << bins          << endl;
    cout << "starthisto="    << starthisto    << endl;
    cout << "endhisto="      << endhisto      << endl;
    cout << "tolerance="     << tolerance     << endl;
    cout << "wantSilence="   << wantSilence   << endl;
    cout << "wantNoise="     << wantNoise     << endl;
    cout << "wantBgLevel="   << wantBgLevel   << endl;
    cout << "wantSumSCF="    << wantSumSCF    << endl;
    cout << "wantCentroid="  << wantCentroid  << endl;
    cout << "wantSSCFhisto=" << wantSSCFhisto << endl;
    cout << "wantSBnrj="     << wantSBnrj     << endl;
    cout << "wantSBmean="    << wantSBmean    << endl;
    cout << "wantSBrms="     << wantSBrms     << endl;
    cout << "wantSBv="       << wantSBv       << endl;
    cout << "wantSBvnrj="    << wantSBvnrj    << endl;
    cout << "wantSBvmean="   << wantSBvmean   << endl;
    cout << "wantSBscf="     << wantSBscf     << endl;
    cout << "wantSBvrms="    << wantSBvrms    << endl;
    cout << "wantsigsb="     << wantsigsb     << endl;
    cout << "wantbandnrj="   << wantbandnrj   << endl;
    cout << "wantsignrj="    << wantsignrj    << endl;
    cout << "wantsigmag="    << wantsigmag    << endl;
    cout << "wantpauserate=" << wantpauserate << endl;
    cout << "wantbandratio=" << wantbandratio << endl;
    cout << "wantrolloff="   << wantrolloff   << endl;
    cout << "wantbw="        << wantbw        << endl;
    cout << "wantbdw="       << wantbdw       << endl;
    cout << "wantspflux="    << wantspflux    << endl;
    cout << "wantcm="        << wantcm        << endl;
    cout << "wantlownrj="    << wantlownrj    << endl;
    cout << "wantvariance="  << wantvariance  << endl;
    cout << "wantmatoutput=" << wantmatoutput << endl;
    cout << "matfilename="   << matfilename   << endl;
    cout << "filename="      << filename      << endl;

#endif // debug
}    

//----------------------------------
int 
main(int argc, char **argv) {

    // check no. of args
    if (argc < 2) {
	PrintUsage(argv[0]);
    }

    // parse arguments
    ParseArguments (argc, argv);

    // check arguments and correct if appropriate
    CheckArguments (argv[0]);
 
    // open SOUNDfile
    SOUNDfile * mpfile = new SOUNDfile(filename);

    if (mpfile->file_type() == NOFILE) {
      cerr << "analyseSDaudio: Sorry can`t analysis this file:" << filename << endl;
      exit(1);
    }

    if (verbose) {
        cout << "SOUND file opened: " << filename << endl;
    }

    // open known plugin libraries
    plugins = new Plugins();
	#if (defined (WIN32) || (_WIN32))
		#ifdef _DEBUG
			string libName = "libMaaateMD.dll";
			if (!plugins->AddLibrary(libName)) {
			  cerr << "Error loading library libMaaateMD.dll" << endl;
			  delete plugins;
			  exit (1);
			}
		#else
			string libName = "libMaaateM.dll";
			if (!plugins->AddLibrary(libName)) {
			  cerr << "Error loading library libMaaateM.dll" << endl;
			  delete plugins;
			  exit (1);
			}
		#endif
	#else
		if (!plugins->AddLibrary(string("libMaaateM.so"))) {
		  cerr << "Error loading library libMaaateM.so" << endl;
		  delete plugins;
		  exit (1);
		}
	#endif

//----------------- sum of scalefactors ------------------------------

    SegmentData *sscf = NULL;
    if (wantSilence || wantNoise || wantBgLevel || wantSumSCF || 
	wantSSCFhisto) {
	// calculate sum of scalefactors between from and to
	if (verbose) {
	    cout << "\nCalculating sum of scalefactors ..." << endl;
	}
	sscf = sumScalefactors(plugins, mpfile, from, to, fromSb, toSb);
	if (sscf == NULL) {
	  cout << "NULL sscf" << endl;
	}
    }

    // if sum of scalefactors requested, print these
    if (wantSumSCF) {
	  if (wantmatoutput)
		  sscf->matOut(matfilename);
	  else
	      cout << *sscf;
    }
    
//----------------- spectral centroid ------------------------------

    SegmentData *cntroid;
    if (wantCentroid) {
	// calculate spectral centroid between from and to
	if (verbose) {
	    cout << "\nCalculating spectral centroid ..." << endl;
	}
	cntroid = spectralCentroid(plugins, mpfile, from, to, fromSb, toSb);
	if (cntroid == NULL) {
	  cout << "NULL centroid" << endl;
	}
    }

    // if spectral centroid requested, print these
    if (wantCentroid) {
	  if (wantmatoutput)
		  cntroid->matOut(matfilename);
	  else
		cout << *cntroid;
    }

//----------------- background noise level ------------------------------

    double bgLevel;
    if (wantBgLevel) {
	if (verbose) {
	cout << "\nCalculating background noise level ..." << endl;
	}
	bgLevel = backgroundNoiseLevel(plugins, sscf, from, to, 
				       minDur, maxInterrupt, 
				       onsetTime, releaseTime);
	cout << "BackgroundNoiseLevel=" << bgLevel << endl;
    }

    // if silence segments requested, calculate
    SegmentTable *sil;
    if (wantSilence) {
	if (verbose) {
	    cout << "\nCalculating silence segments ..." << endl;
	}
	sil = silenceSegmentation(plugins, sscf, from, to, 
				  thresh, minDur, maxInterrupt, 
				  onsetTime, releaseTime);
	if (sil == NULL) {
	  cerr << argv[0] << ": silence segmentation failed" << endl;
	} else {
	    if (verbose) {
		cout << "\nCalculated " << sil->size()
		     << " silence segments ..." << endl;
		cout << "Start\t\tEnd\t\tDuration\tConfidence" << endl;
	    }
	    // easier: sil->printPlain(false);
	    // this is only to demonstrate how to process one by one
	    int i=0;
	    while(i<sil->size()) {
		cout << (*sil)[i].start()    << "\t"
		     << (*sil)[i].end()      << "\t"
		     << (*sil)[i].duration() << "\t"
		     << (*sil)[i].confidence() << endl;
		i++;
	    }
	}
    }

//----------------- noise segments ------------------------------

    SegmentTable *noise;
    if (wantNoise) {
	if (verbose) {
	    cout << "\nCalculating noisy segments ..." << endl;
	}
	noise = noiseSegmentation(plugins, sscf, from, to, 
				  thresh, minDur, maxInterrupt, 
				  onsetTime, releaseTime);
	if (noise == NULL) {
	  cerr << argv[0] << ": noise segmentation failed" << endl;
	} else {
	  if (verbose) {
	    cout << "\nCalculated " << noise->size() 
		 << " noise segments ..." << endl;
	    cout << "Start - End - Duration" << endl;
	  }
	  noise->printPlain(false);
	}
    }
    
//----------------- histo of SSCF ------------------------------

    // if want histo of SSCF, print
    SegmentData *histo = NULL;
    if (wantSSCFhisto) {
      if (verbose) {
	cout << "\nCalculating SSCF histogram ..." << endl;
      }
      list<ModuleParam> * paramsOut = 
	histo1D (plugins, sscf, from, to, bins, starthisto, endhisto);
      list<ModuleParam>::iterator iter;
      if (paramsOut->size() == 0) {
	cout << "No paramsOut" << endl;
      }
      
      // get resulting data
      double binwidth;
      int    nrvalues;
      iter=paramsOut->begin();
      histo = (*iter).get_sd();
      ++iter; starthisto = (*iter).get_r();
      ++iter; endhisto = (*iter).get_r();
      ++iter; binwidth = (*iter).get_r();
      ++iter; nrvalues = (*iter).get_i();

      if (histo == NULL) {
	cerr << argv[0] << ": histogram calculation failed" << endl;
      } else {
	cout << "Histogram values:" << endl;
	cout << *histo;
	cout << "Values lie between " << starthisto
	     << " and " << endhisto << endl;
	cout << "Bins are " << binwidth
	     << " wide." << endl;
	cout << "There were " << nrvalues << " values." << endl;
      }
    }

//----------------- SBnrj ------------------------------
 
	SegmentData *sbnrj = NULL;
    if (wantSBnrj || wantbdw) {
	if (verbose) {
	    cout << "\nCalculating SBnrj ..." << endl;
	}
	sbnrj = SBnrj(plugins, mpfile, from, to, fromSb, toSb);
	if (sbnrj == NULL) {
	  cout << "NULL sbnrj" << endl;
	}
    }
    if (wantSBnrj) {
	  if (wantmatoutput)
		  sbnrj->matOut(matfilename);
	  else	  
		  cout << *sbnrj;
    }


//----------------- SBmean ------------------------------

    SegmentData *sbmean;
    if (wantSBmean) {
	if (verbose) {
	    cout << "\nCalculating SBmean ..." << endl;
	}
	sbmean = SBmean(plugins, mpfile, from, to, fromSb, toSb);
	if (sbmean == NULL) {
	  cout << "NULL sbmean" << endl;
	}
    }
    if (wantSBmean) {
	  if (wantmatoutput)
		  sbmean->matOut(matfilename);
	  else
		  cout << *sbmean;
    }


//----------------- SBrms ------------------------------

    SegmentData *sbrms;
    if (wantSBrms) {
	if (verbose) {
	    cout << "\nCalculating SBrms ..." << endl;
	}
	sbrms = SBrms(plugins, mpfile, from, to, fromSb, toSb);
	if (sbrms == NULL) {
	  cout << "NULL sbrms" << endl;
	}
    }
    if (wantSBrms) {
	  if (wantmatoutput)
		  sbrms->matOut(matfilename);
	  else
		  cout << *sbrms;
    }

//----------------- SBscf ------------------------------

    SegmentData *sbscf;
    if (wantSBscf) {
	if (verbose) {
	    cout << "\nCalculating SBscf ..." << endl;
	}
	sbscf = SBscf(plugins, mpfile, from, to, fromSb, toSb);
	if (sbscf == NULL) {
	  cout << "NULL sbscf" << endl;
	}
    }
    if (wantSBscf) {
	  if (wantmatoutput)
		  sbscf->matOut(matfilename);
	  else
		  cout << *sbscf;
    }

//----------------- SBv ------------------------------

    SegmentData *sbv;
    if (wantSBv) {
	if (verbose) {
	    cout << "\nCalculating SBv ..." << endl;
	}
	sbv = SBvalue(plugins, mpfile, from, to, fromSb, toSb);
	if (sbv == NULL) {
	  cout << "NULL sbv" << endl;
	}
    }
    if (wantSBv) {
	  if (wantmatoutput)
		  sbv->matOut(matfilename);
	  else
		  cout << *sbv;
    }

//----------------- SBvmean ------------------------------

    SegmentData *sbvmean;
    if (wantSBvmean) {
	if (verbose) {
	    cout << "\nCalculating SBvmean ..." << endl;
	}
	sbvmean = SBvaluemean(plugins, mpfile, from, to, fromSb, toSb);
	if (sbvmean == NULL) {
	  cout << "NULL sbvmean" << endl;
	}
    }
    if (wantSBvmean) {
	  if (wantmatoutput)
		  sbvmean->matOut(matfilename);
	  else
		  cout << *sbvmean;
    }

//----------------- SBvrms ------------------------------

    SegmentData *sbvrms;
    if (wantSBvrms) {
	if (verbose) {
	    cout << "\nCalculating SBvrms ..." << endl;
	}
	sbvrms = SBvaluerms(plugins, mpfile, from, to, fromSb, toSb);
	if (sbvrms == NULL) {
	  cout << "NULL sbvrms" << endl;
	}
    }
    if (wantSBvrms) {
	  if (wantmatoutput)
		  sbvrms->matOut(matfilename);
	  else
		  cout << *sbvrms;
    }

//----------------- SBvnrj ------------------------------

    SegmentData *sbvnrj;
    if (wantSBvnrj) {
	if (verbose) {
	    cout << "\nCalculating SBvnrj ..." << endl;
	}
	sbvnrj = SBvaluenrj(plugins, mpfile, from, to, fromSb, toSb);
	if (sbvnrj == NULL) {
	  cout << "NULL sbvnrj" << endl;
	}
    }
    if (wantSBvnrj) {
	  if (wantmatoutput)
		  sbvnrj->matOut(matfilename);
	  else
		  cout << *sbvnrj;
    }



//----------------- signrj ------------------------------

    SegmentData *signrj = NULL;
    if (wantsignrj || wantpauserate || wantlownrj || wantvariance) {
	if (verbose) {
	    cout << "\nCalculating signrj ..." << endl;
	}
	signrj = signalnrj(plugins, mpfile, from, to, bins);
	if (signrj == NULL) {
	  cout << "NULL signrj" << endl;
	}
    }
    if (wantsignrj) {
	  if (wantmatoutput)
		  signrj->matOut(matfilename);
	  else
		  cout << *signrj;
    }

//----------------- bandnrj ------------------------------
 
	SegmentData *bandnrj;
    if (wantbandnrj) {
	if (verbose) {
	    cout << "\nCalculating bandnrj ..." << endl;
	}
	bandnrj = bandNrj(plugins, mpfile, from, to, fromSb, toSb, bins);
	if (bandnrj == NULL) {
	  cout << "NULL bandnrj" << endl;
	}
    }
    if (wantbandnrj) {
	  if (wantmatoutput)
		  bandnrj->matOut(matfilename);
	  else
		  cout << *bandnrj;
    }

//----------------- sigsb ------------------------------
 
	SegmentData *sigsb;
    if (wantsigsb) {
	if (verbose) {
	    cout << "\nCalculating sigsb ..." << endl;
	}
	sigsb = sigSb(plugins, mpfile, from, to, thresh);
	if (sigsb == NULL) {
	  cout << "NULL sigsb" << endl;
	}
    }
    if (wantsigsb) {
	  if (wantmatoutput)
		  sigsb->matOut(matfilename);
	  else
		  cout << *sigsb;
    }

//----------------- sigmag ------------------------------

    SegmentData *sigmag;
    if (wantsigmag) {
	if (verbose) {
	    cout << "\nCalculating sigmag ..." << endl;
	}
	sigmag = signalMagnitude(plugins, mpfile, from, to, bins);
	if (sigmag == NULL) {
	  cout << "NULL sigmag" << endl;
	}
    }
    if (wantsigmag) {
	  if (wantmatoutput)
		  sigmag->matOut(matfilename);
	  else
		  cout << *sigmag;
    }

//----------------- pauserate ------------------------------

    SegmentData *pauserate;
    if (wantpauserate) {
	if (verbose) {
	    cout << "\nCalculating pauserate ..." << endl;
	}
	pauserate = pauseRate(plugins, signrj, from, to, thresh, minDur);
	if (pauserate == NULL) {
	  cout << "NULL pauserate" << endl;
	}
    }
    if (wantpauserate) {
	  if (wantmatoutput)
		  pauserate->matOut(matfilename);
	  else
		  cout << *pauserate;
    }	  

//----------------- bandratio ------------------------------

    SegmentData *bandratio;
    if (wantbandratio) {
	if (verbose) {
	    cout << "\nCalculating bandratio ..." << endl;
	}
	bandratio = BandNrjRatio(plugins, mpfile, from, to, fromSb, toSb);
	if (bandratio == NULL) {
	  cout << "NULL bandratio" << endl;
	}
    }
    if (wantbandratio) {
	  if (wantmatoutput)
		  bandratio->matOut(matfilename);
	  else
		  cout << *bandratio;
    }

//----------------- rolloff ------------------------------

	SegmentData *rolloff;
    if (wantrolloff) {
	if (verbose) {
	    cout << "\nCalculating rolloff ..." << endl;
	}
	rolloff = rollOff(plugins, mpfile, from, to);
	if (rolloff == NULL) {
	  cout << "NULL rolloff" << endl;
	}
    }
    if (wantrolloff) {
	  if (wantmatoutput)
		  rolloff->matOut(matfilename);
	  else
		  cout << *rolloff;
    }

//----------------- bandwidth ------------------------------

    Trio bd;
    if (wantbw) {
	if (verbose) {
	    cout << "\nCalculating bandwidth ..." << endl;
	}
	bd = bandwidth(plugins, mpfile, from, to, thresh);
	if (bd.uno == 0) {
	  cout << "NULL bd" << endl;
	}
    }
    if (wantbw) {
      cout << "bandwidth :" << endl;
      cout << *(bd.uno) << endl;
      cout << "fc min :" << endl;
      cout << *(bd.dos) << endl;
      cout << "fc max :" << endl;
      cout << *(bd.tres) << endl;

	  if (wantmatoutput)
		  // only output of bandwidth
		  bd.uno->matOut(matfilename);
    }

//----------------- signal bandwidth ------------------------------

    Trio bdw;
    if (wantbdw) {
	if (verbose) {
	    cout << "\nCalculating signal bandwidth ..." << endl;
	}
	bdw = sigBdwidth(plugins, sbnrj, from, to, thresh);
	if (bdw.uno == 0) {
	  cout << "NULL bdw" << endl;
	}
    }
    if (wantbdw) {
      cout << "bandwidth :" << endl;
      cout << *(bdw.uno) << endl;
      cout << "fc min :" << endl;
      cout << *(bdw.dos) << endl;
      cout << "fc max :" << endl;
      cout << *(bdw.tres) << endl;

	  if (wantmatoutput)
		  // only output of bandwidth
		  bdw.uno->matOut(matfilename);
    }

//----------------- spectralflux ------------------------------

    SegmentData *spflux ;
    if (wantspflux ) {
	if (verbose) {
	    cout << "\nCalculating spectralflux  ..." << endl;
	}
	spflux = spectralFlux(plugins, mpfile, from, to);
	if (spflux == NULL) {
	  cout << "NULL spflux" << endl;
	}
    }
    if (wantspflux) {
	  if (wantmatoutput)
		  spflux->matOut(matfilename);
	  else
		  cout << *spflux;
    }

//----------------- centralmoment ------------------------------

    SegmentData *cm;
    if (wantcm) {
	if (verbose) {
	    cout << "\nCalculating central moment ..." << endl;
	}
	cm = centralMoment(plugins, mpfile, from, to, fromSb, toSb, minDur, bins);
	if (cm == NULL) {
	  cout << "NULL cm" << endl;
	}
    }
    if (wantcm) {
	  if (wantmatoutput)
		  cm->matOut(matfilename);
	  else
		  cout << *cm;
    }

//----------------- lownrj ------------------------------

    SegmentData *lownrj;
    if (wantlownrj) {
	if (verbose) {
	    cout << "\nCalculating lownrj ..." << endl;
	}
	lownrj = lowNrj(plugins, signrj, from, to, minDur);
	if (lownrj == NULL) {
	  cout << "NULL lownrj" << endl;
	}
    }
    if (wantlownrj) {
	  if (wantmatoutput)
		  lownrj->matOut(matfilename);
	  else
		  cout << *lownrj;
    }

//----------------- variance ------------------------------

    SegmentData *var;
    if (wantvariance) {
	if (verbose) {
	    cout << "\nCalculating var ..." << endl;
	}
	var = variance(plugins, signrj, from, to, minDur);
	if (var == NULL) {
	  cout << "NULL var" << endl;
	}
    }
    if (wantvariance) {
	  if (wantmatoutput)
		  var->matOut(matfilename);
	  else
		  cout << *var;
    }
    
    exit (0);
	return 0;
}
