/*
    MPEG Maaate: An Australian MPEG audio analysis toolkit
    Copyright (C) 2000 Commonwealth Scientific and Industrial Research Organisation
    (CSIRO), Australia.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <float.h>
#include <ctype.h>
#include <string.h>

#include "config.h"

#include "SOUNDfile.H"
#include "plugins.H"

#ifndef true
#define true 1
#endif

#ifndef false
#define false 0
#endif

#ifndef bool
#define bool int
#endif

#ifndef NULL
#define NULL 0L
#endif

/* parameters to control what is to be extracted */
static float from          = 0.0;      /* start at second 0.0 */
static float to            = FLT_MAX;  /* end at end of file */
static bool  verbose       = false;
static short fromSb        = 0;        /* first subband to be extracted */
static short toSb          = 575;      /* last subband to be extracted */
static float thresh        = 0.07;     /* threshold between [0;1] */
static float minDur        = 0.1;      /* minimum duration [sec] */
static float maxInterrupt  = 0.0;      /* mam tolerated disruption [sec] */
static float releaseTime   = 0.01;     /* release time for sil start [sec] */
static float onsetTime     = 0.01;     /* onset time for sil end [sec] */
static int   bins          = 10;       /* number of bins for histo */
static float starthisto    = DBL_MAX;  /* start of the histogram values */
static float endhisto      = DBL_MIN;  /* end of the histogram values */
static float tolerance     = 0.0;      /* time identity tolerance for compare */
static bool  wantSilence   = false;    /* want silence? */
static bool  wantNoise     = false;    /* want noise? */
static bool  wantBgLevel   = false;    /* want background noise level? */
static bool  wantSumSCF    = false;    /* want sum of scalefactors? */
static bool  wantSSCFhisto = false;    /* want histogram of SumSCF? */
static char *filename      = NULL;     /* mpeg-audio-file */

Plugins * plugins;

/*----------------------------------*/
void
PrintUsage(char *prog) {
  printf("\nUsage:\n");
  printf("%s [Options] <mpeg-audio-file>\n", prog);
  printf("Default Options: -S 0.0 -E eof\n");
  printf("                 (i.e. analyse whole file)\n");
  printf("\n");
  printf("Possible Options:\n");
  printf("-h           help = prints this information\n");
  printf("-S <from>    the second (float) to start analysing from\n");
  printf("             Default: 0.0\n");
  printf("-E <to>      the second (float) to stop analysing at\n");
  printf("             Default: end of file\n");
  printf("-V           verbose\n");
  printf("-B <from> <to> subband range (ints) that gets extracted\n");
  printf("             (only required with -X sscf)\n");
  printf("             Default: 0 31 (Layers I, II)\n");
  printf("                      0 575 (Layer III)\n");
  printf("-T <thresh>  threshold value between [0;1]\n");
  printf("             Default: 0.07\n");
  printf("-M <minDur>  minimum duration [sec] of specified content\n");
  printf("             Default: 0.1\n");
  printf("-N <maxInter> mamimum tolerated disruption [sec]\n");
  printf("             Default: 0.0\n");
  printf("-R <releaseT> release time [sec] delaying start of segment\n");
  printf("             Default: 0.01 (must be < minDur)\n"); 
  printf("-O <onsetT>  onset time [sec] earlying end of segment\n");
  printf("             Default: 0.01 (must be < minDur)\n"); 
  printf("-Z <bins>    number of bins for histogram\n");
  printf("             Default: 10\n");
  printf("-Y <from> <to> start and end of histogram values (float)\n");
  printf("             Default: -1 -1 (meaning: minvalue maxvalue)\n");
  printf("-X [content] specifies content to extract from audio file\n");
  printf("             Possible content types:\n");
  printf("    silence  -T <thresh> -M <minDur> -N <maxInter>\n");
  printf("             Temporal Segmentation: silence segments based on sumscf\n");
  printf("    noise    -T <thresh> -M <minDur> -N <maxInter>\n");
  printf("             Temporal Segmentation: noise segments based on sumscf\n");
  printf("    bglevel  -M <minDur> -N <maxInter>\n");
  printf("             Single Value: background noise level based on sumscf\n");
  printf("    sumscf   -B <fromsb> <tosb> (only first channel)\n");
  printf("             1D Curve: sum of scalefactors as loudness approximation\n");
  printf("    histo    -Z <bins>\n");
  printf("             Vector: loudness distribution for whole duration based on sumscf\n");

  exit (1);
}

/*----------------------------------*/
void
ParseArguments (int argc, char **argv) {
  int i;

  /* go through parameters and check values */
  for (i=1; i<argc; i++) {
    if (argv[i][0] == '-' && argv[i][1]) {
      switch (argv[i][1]) {
      case 'H': case 'h': /* help */
	PrintUsage(argv[0]);
	break;
      case 'S': case 's': /* second to start from */
	if ((i+1<argc) && !isalpha(argv[i+1][0])) {
	  from = atof(argv[++i]);
	}
	break;
      case 'E': case 'e': /* second to end at */
	if ((i+1<argc) && !isalpha(argv[i+1][0])) {
	  to = atof(argv[++i]);
	}
	break;
      case 'V': case 'v': /* verbose */
	verbose = true;
	break;
      case 'B': case 'b': /* subband range */
	if ((i+2<argc) && !isalpha(argv[i+1][0]) 
	    && !isalpha(argv[i+2][0])) {
	  fromSb = atoi(argv[++i]);
	  toSb   = atoi(argv[++i]);
	} else {
	  printf("%s: -B option needs two integers to follow\n", argv[0]);
	  exit(1);
	}
	break;
      case 'T': case 't': /* threshold */
	if ((i+1<argc) && !isalpha(argv[i+1][0])) {
	  thresh = atof(argv[++i]);
	}
	break;
      case 'M': case 'm': /* minimum duration */
	if ((i+1<argc) && !isalpha(argv[i+1][0])) {
	  minDur = atof(argv[++i]);
	}
	break;
      case 'N': case 'n': /* maximum interrupt */
	if ((i+1<argc) && !isalpha(argv[i+1][0])) {
	  maxInterrupt = atof(argv[++i]);
	}
	break;
      case 'R': case 'r': /* release time */
	if ((i+1<argc) && !isalpha(argv[i+1][0])) {
	  releaseTime = atof (argv[++i]);
	}
	break;
      case 'O': case 'o': /* onset time */
	if ((i+1<argc) && !isalpha(argv[i+1][0])) {
	  onsetTime = atof (argv[++i]);
	}
	break;
      case 'Z': case 'z': /* no bins */
	if ((i+1<argc) && !isalpha(argv[i+1][0])) {
	  bins = atoi(argv[++i]);
	}
	break;
      case 'Y': case 'y': /* histogram range */
	if ((i+2<argc) && !isalpha(argv[i+1][0]) 
	    && !isalpha(argv[i+2][0])) {
	  starthisto = atof(argv[++i]);
	  endhisto   = atof(argv[++i]);
	} else {
	  printf("%s: -Y option needs two floats to follow\n", argv[0]);
	  exit(1);
	}
	break;
      case 'X': case 'x': /* extract content */
	if ((i+1 == argc) || !isalpha(argv[i+1][0])) {
	  printf("%s: -X option needs string-argument\n", argv[0]);
	  exit(1);
	} else {
	  if (!strncmp(argv[i+1],"silence",3)) {
	    wantSilence = true; i++;
	  } else if (!strncmp(argv[i+1],"noise",3)) {
	    wantNoise = true; i++;
	  } else if (!strncmp(argv[i+1],"bglevel",3)) {
	    wantBgLevel = true; i++;
	  } else if (!strncmp(argv[i+1],"sumscf",3)) {
	    wantSumSCF = true; i++;
	  } else if (!strncmp(argv[i+1],"histo",3)) {
	    wantSSCFhisto = true; i++;
 	  } else if (argv[i+1][0] != '-') {
	    printf("%s: -X unknown string-argument %s\n", argv[0],
		   argv[i+1]);
	    exit(1);
	  }
	}
	break;
      } /* switch */
    } else {
      filename = argv[i];
    }
  }
}
		
/*----------------------------------*/
void
CheckArguments(char *prgname) {
  if (filename == NULL) {
    printf("%s: Mpeg audio filename missing\n", prgname);
    exit(1);
  }
  if (from < 0.0)    from = 0.0;
  if (to   < from)   to   = from;
  if (fromSb   < 0)  fromSb   = 0;
  if (toSb < fromSb) toSb     = fromSb;
  if (thresh < 0.0)  thresh = 0.0;
  if (thresh > 1.0)  thresh = 1.0;
  if (releaseTime > minDur) releaseTime = 0.0;
  if (onsetTime > minDur)   onsetTime = 0.0;
  if (bins <= 0)     bins = 10;
  if (tolerance < 0) tolerance = -tolerance;
 
#ifdef DEBUG
  printf("Parameter Settings:\n");
  printf("From=%f\n", from);
  printf("To=%f\n", to);
  printf("Verbose=%d\n", verbose);
  printf("fromSb=%d\n", fromSb);
  printf("toSb=%d\n", toSb);
  printf("thresh=%f\n", thresh);
  printf("minDur=%f\n", minDur);
  printf("maxInterrupt=%f\n", maxInterrupt);
  printf("releaseTime=%f\n", releaseTime);
  printf("onsetTime=%f\n", onsetTime);
  printf("bins=%d\n", bins);
  printf("starthisto=%d\n", starthisto);
  printf("endhisto=%d\n", endhisto);
  printf("tolerance=%f\n", tolerance);
  printf("wantSilence=%d\n", wantSilence);
  printf("wantNoise=%d\n", wantNoise);
  printf("wantBgLevel=%d\n", wantBgLevel);
  printf("wantSumSCF=%d\n", wantSumSCF);
  printf("wantSSCFhisto=%d\n", wantSSCFhisto);
  printf("filename=%s\n", filename);
#endif
}    


/*----------------------------------*/
int 
main(int argc, char **argv) {
  SegmentData  *sscf = NULL;
  SOUNDfile     *sfile;
  double        bgLevel;
  SegmentTable *silences;
  SegmentTable *noise;
  SegmentData  *histo;
  Module       *m;
  ModuleParamList *paramsIn, *paramsOut;
  ModuleParam  *param;
  double       binwidth;
  int          nrvalues;
  int i;

  /* check no. of args */
  if (argc < 2) {
    PrintUsage(argv[0]);
  }

  /* parse arguments */
  ParseArguments (argc, argv);

  /* check arguments and correct if appropriate */
  CheckArguments (argv[0]);
 
  /* open mpegfile */
  sfile = maaateP_new_soundfile(filename);

  if (maaateP_get_filetype(sfile) == NOFILE) {
    printf("Sorry can`t analyse that file.");
    exit(1);
  }

  if (verbose) {
    printf("Sound file opened: %s\n", filename);
  }

  /* open known plugin libraries */
  plugins = maaateA_new_plugins();
  if (!maaateA_add_library(plugins, "libMaaateM.so")) {
    printf("Error loading plugin library libMaaateM.so\n");
  }

  /* calculate sum of scalefactors if required */
  if (wantSilence || wantNoise || wantSumSCF || wantBgLevel || wantSSCFhisto) {
    /* calculate sum of scalefactors between from and to */
    if (verbose) {
      printf("\nCalculating sum of scalefactors ...\n");
    }
    
    /* calculate it using the plugin module */
    m = maaateA_get_module(plugins, "sumscf");
    if (m==NULL) {
      fprintf(stderr, "module sumscf not retrieved successfully\n");
      exit(1);
    }

    /* get default values */
    paramsIn = maaateA_default_values(m);
    if (maaateA_ModuleParamList_length(paramsIn) == 0) {
      printf("No paramsIn\n");
    }

    /* set parameters */
    param = maaateA_ModuleParamList_nth (paramsIn, 0);
    maaateA_ModuleParam_setSOUNDfile (param, sfile);
    param = maaateA_ModuleParamList_nth (paramsIn, 1);
    maaateA_ModuleParam_setReal (param, from);
    param = maaateA_ModuleParamList_nth (paramsIn, 2);
    maaateA_ModuleParam_setReal (param, to);
    param = maaateA_ModuleParamList_nth (paramsIn, 3);
    maaateA_ModuleParam_setInt  (param, fromSb);
    param = maaateA_ModuleParamList_nth (paramsIn, 4);
    maaateA_ModuleParam_setInt  (param, toSb);

    /* suggest and check parameters */
    maaateA_suggest_values (m, paramsIn);
    if (maaateA_ModuleParamList_length(paramsIn) == 0) {
      printf("no paramsIn\n");
    }

    /* execute module */
    paramsOut = maaateA_apply (m, paramsIn);
    if (maaateA_ModuleParamList_length(paramsOut) == 0) {
      printf("No paramsOut\n");
    }

    /* get resulting segmentData */
    param = maaateA_ModuleParamList_nth (paramsOut, 0);
    sscf = maaateA_ModuleParam_getSegmentData (param);
    if (sscf == NULL) {
      printf("NULL sscf");
    }

  }

  /* if sum of scalefactors requested, print these */
  if (wantSumSCF) {
    maaateA_sd_print(sscf);
  }
   
  /* if background noise level requested, calculate */
  if (wantBgLevel) {
    if (verbose) {
      printf("\nCalculating background noise level ...\n");
    }
    /* calculate it using the plugin module */
    m = maaateA_get_module(plugins, "bgnoiselevel");
    if (m==NULL) {
      fprintf(stderr, "module bgnoiselevel not retrieved successfully\n");
      exit(1);
    }

    /* get default values */
    paramsIn = maaateA_default_values(m);
    if (maaateA_ModuleParamList_length(paramsIn) == 0) {
      printf("No paramsIn\n");
    }

    /* set parameters */
    param = maaateA_ModuleParamList_nth (paramsIn, 0);
    maaateA_ModuleParam_setSegmentData (param, sscf);
    param = maaateA_ModuleParamList_nth (paramsIn, 1);
    maaateA_ModuleParam_setReal (param, from);
    param = maaateA_ModuleParamList_nth (paramsIn, 2);
    maaateA_ModuleParam_setReal (param, to);
    param = maaateA_ModuleParamList_nth (paramsIn, 3);
    maaateA_ModuleParam_setReal (param, minDur);
    param = maaateA_ModuleParamList_nth (paramsIn, 4);
    maaateA_ModuleParam_setReal (param, maxInterrupt);
    param = maaateA_ModuleParamList_nth (paramsIn, 5);
    maaateA_ModuleParam_setReal (param, onsetTime);
    param = maaateA_ModuleParamList_nth (paramsIn, 6);
    maaateA_ModuleParam_setReal (param, releaseTime);

    /* suggest and check parameters */
    maaateA_suggest_values (m, paramsIn);
    if (maaateA_ModuleParamList_length(paramsIn) == 0) {
      printf("no paramsIn\n");
    }

    /* execute module */
    paramsOut = maaateA_apply (m, paramsIn);
    if (maaateA_ModuleParamList_length(paramsOut) == 0) {
      printf("No paramsOut\n");
    }

    /* get resulting double */
    param = maaateA_ModuleParamList_nth (paramsOut, 0);
    bgLevel = maaateA_ModuleParam_getReal (param);

    printf("BackgroundNoiseLevel= %g\n", bgLevel);
  }


  /* if silence segments requested, calculate */
  if (wantSilence) {
    if (verbose) {
      printf("\nCalculating silence segments ...\n");
    }

    /* calculate it using the plugin module */
    m = maaateA_get_module(plugins, "silence");
    if (m==NULL) {
      fprintf(stderr, "module silence not retrieved successfully\n");
      exit(1);
    }

    /* get default values */
    paramsIn = maaateA_default_values(m);
    if (maaateA_ModuleParamList_length(paramsIn) == 0) {
      printf("No paramsIn\n");
    }

    /* set parameters */
    param = maaateA_ModuleParamList_nth (paramsIn, 0);
    maaateA_ModuleParam_setSegmentData (param, sscf);
    param = maaateA_ModuleParamList_nth (paramsIn, 1);
    maaateA_ModuleParam_setReal (param, from);
    param = maaateA_ModuleParamList_nth (paramsIn, 2);
    maaateA_ModuleParam_setReal (param, to);
    param = maaateA_ModuleParamList_nth (paramsIn, 3);
    maaateA_ModuleParam_setReal (param, thresh);
    param = maaateA_ModuleParamList_nth (paramsIn, 4);
    maaateA_ModuleParam_setReal (param, minDur);
    param = maaateA_ModuleParamList_nth (paramsIn, 5);
    maaateA_ModuleParam_setReal (param, maxInterrupt);
    param = maaateA_ModuleParamList_nth (paramsIn, 6);
    maaateA_ModuleParam_setReal (param, onsetTime);
    param = maaateA_ModuleParamList_nth (paramsIn, 7);
    maaateA_ModuleParam_setReal (param, releaseTime);

    /* suggest and check parameters */
    maaateA_suggest_values (m, paramsIn);
    if (maaateA_ModuleParamList_length(paramsIn) == 0) {
      printf("no paramsIn\n");
    }

    /* execute module */
    paramsOut = maaateA_apply (m, paramsIn);
    if (maaateA_ModuleParamList_length(paramsOut) == 0) {
      printf("No paramsOut\n");
    }

    /* get resulting segmentData */
    param = maaateA_ModuleParamList_nth (paramsOut, 0);
    silences = maaateA_ModuleParam_getSegmentTable (param);

    if (silences == NULL) {
      printf("%s: no silences calculated\n", argv[0]);
    } else {
      if (verbose) {
	printf("\nCalculated %d silence segments\n",
	       maaateA_st_get_size(silences));
      }
      maaateA_st_print_plain(silences, false);
    }
  }
  
  /* if noise segments requested, calculate */
  if (wantNoise) {
    if (verbose) {
      printf("\nCalculating noisy segments ...\n");
    }

    /* calculate it using the plugin module */
    m = maaateA_get_module(plugins, "segmentation");
    if (m==NULL) {
      fprintf(stderr, "module segmentation not retrieved successfully\n");
      exit(1);
    }

    /* get default values */
    paramsIn = maaateA_default_values(m);
    if (maaateA_ModuleParamList_length(paramsIn) == 0) {
      printf("No paramsIn\n");
    }

    /* set parameters */
    param = maaateA_ModuleParamList_nth (paramsIn, 0);
    maaateA_ModuleParam_setSegmentData (param, sscf);
    param = maaateA_ModuleParamList_nth (paramsIn, 1);
    maaateA_ModuleParam_setReal (param, from);
    param = maaateA_ModuleParamList_nth (paramsIn, 2);
    maaateA_ModuleParam_setReal (param, to);
    param = maaateA_ModuleParamList_nth (paramsIn, 3);
    maaateA_ModuleParam_setBool (param, false);
    param = maaateA_ModuleParamList_nth (paramsIn, 4);
    maaateA_ModuleParam_setReal (param, thresh);
    param = maaateA_ModuleParamList_nth (paramsIn, 5);
    maaateA_ModuleParam_setReal (param, minDur);
    param = maaateA_ModuleParamList_nth (paramsIn, 6);
    maaateA_ModuleParam_setReal (param, maxInterrupt);
    param = maaateA_ModuleParamList_nth (paramsIn, 7);
    maaateA_ModuleParam_setReal (param, onsetTime);
    param = maaateA_ModuleParamList_nth (paramsIn, 8);
    maaateA_ModuleParam_setReal (param, releaseTime);

    /* suggest and check parameters */
    maaateA_suggest_values (m, paramsIn);
    if (maaateA_ModuleParamList_length(paramsIn) == 0) {
      printf("no paramsIn\n");
    }

    /* execute module */
    paramsOut = maaateA_apply (m, paramsIn);
    if (maaateA_ModuleParamList_length(paramsOut) == 0) {
      printf("No paramsOut\n");
    }

    /* get resulting segmentData */
    param = maaateA_ModuleParamList_nth (paramsOut, 0);
    noise = maaateA_ModuleParam_getSegmentTable (param);

    if (noise == NULL) {
      printf("%s: no noise segments found\n", argv[0]);
    } else {
      if (verbose) {
	printf("\nCalculated %d noise segments\n",
	       maaateA_st_get_size(noise));
      }
      maaateA_st_print_plain(noise, false);
    }
  }
  
  
  /* if want histo of SSCF, print */
  if (wantSSCFhisto) {
    if (verbose) {
      printf("\nCalculating SSCF histogram ...\n");;
    }

    /* calculate it using the plugin module */
    m = maaateA_get_module(plugins, "histogram1D");
    if (m==NULL) {
      fprintf(stderr, "module histogram1D not retrieved successfully\n");
      exit(1);
    }

    /* get default values */
    paramsIn = maaateA_default_values(m);
    if (maaateA_ModuleParamList_length(paramsIn) == 0) {
      printf("No paramsIn\n");
    }

    /* set parameters */
    param = maaateA_ModuleParamList_nth (paramsIn, 0);
    maaateA_ModuleParam_setSegmentData (param, sscf);
    param = maaateA_ModuleParamList_nth (paramsIn, 1);
    maaateA_ModuleParam_setReal (param, from);
    param = maaateA_ModuleParamList_nth (paramsIn, 2);
    maaateA_ModuleParam_setReal (param, to);
    param = maaateA_ModuleParamList_nth (paramsIn, 3);
    maaateA_ModuleParam_setInt (param, bins);
    param = maaateA_ModuleParamList_nth (paramsIn, 4);
    maaateA_ModuleParam_setInt (param, starthisto);
    param = maaateA_ModuleParamList_nth (paramsIn, 5);
    maaateA_ModuleParam_setInt (param, endhisto);

    /* suggest and check parameters */
    maaateA_suggest_values (m, paramsIn);
    if (maaateA_ModuleParamList_length(paramsIn) == 0) {
      printf("no paramsIn\n");
    }

    /* execute module */
    paramsOut = maaateA_apply (m, paramsIn);
    if (maaateA_ModuleParamList_length(paramsOut) == 0) {
      printf("No paramsOut\n");
    }

    /* get resulting segmentData */
    param = maaateA_ModuleParamList_nth (paramsOut, 0);
    histo = maaateA_ModuleParam_getSegmentData (param);
    param = maaateA_ModuleParamList_nth (paramsOut, 1);
    starthisto = maaateA_ModuleParam_getReal (param);
    param = maaateA_ModuleParamList_nth (paramsOut, 2);
    endhisto = maaateA_ModuleParam_getReal (param);
    param = maaateA_ModuleParamList_nth (paramsOut, 3);
    binwidth = maaateA_ModuleParam_getReal (param);
    param = maaateA_ModuleParamList_nth (paramsOut, 4);
    nrvalues = maaateA_ModuleParam_getInt (param);

    if (histo == NULL) {
      printf("%s: histogram calculation failed\n", argv[0]);
    } else {
      printf("Histogram values:\n");
      maaateA_sd_print(histo);
      printf("Values lie between %f and %f.\n", starthisto, endhisto);
      printf("Bins are %f wide.\n", binwidth);
      printf("There were %d values.\n", nrvalues);
    }
  }

  exit (0);
}

