/*
    MPEG Maaate: An Australian MPEG audio analysis toolkit
    Copyright (C) 2000 Commonwealth Scientific and Industrial Research Organisation
    (CSIRO), Australia.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "MPEGfile.H"
#include "SOUNDfile.H"

#include <float.h>

using namespace std;

///// TODOs: /////
// - output time step/resolution for each extracted parameter
// - output in xwaves format


// parameters to control what is to be extracted
static float from          = 0.0;      // start at second 0.0
static float to            = FLT_MAX;  // end at end of file
static bool  verbose       = false;
static short printHeader   = -1;       // -1=no 0=first 1=all
static short printSideinfo = -1;       // -1=no 0=first 1=all
static bool  printDur      = false;    // total file duration
static short fromSb        = 0;        // first subband to be extracted
static short toSb          = 576;      // last subband to be extracted
static short channels      = 1;        // mono/stereo processing?
static bool  wantBitalloc  = false;    // want bitallocation?
static bool  wantScf       = false;    // want scalefactors?
static bool  wantScfsi     = false;    // want scalefactor select info?
static bool  wantValues    = false;    // want frequency values?
static bool  wantPcm       = false;    // want pcm values?
static bool  wantPcmFile   = false;    // want pcm output file?
static string outfile      = string("out.raw"); // name of pcm output file
static char *filename      = NULL;     // mpeg-audio-file

static FILE     *outpcm = NULL;
static SOUNDfile *sfile = NULL;
static MPEGfile *infile = NULL;

// prints help information about parameter usage
void 
PrintUsage(char *prog) {
    cerr << "\nUsage:" << endl;
    cerr << prog << " [Options] <mpeg-audio-file>" << endl;
    cerr << "Default Options: -S 0.0 -E eof" << endl;
    cerr << "                 (i.e. analyse whole file)" << endl;
    cerr << endl;
    cerr << "Possible Options:" << endl;
    cerr << "-h         help=prints this information" << endl;
    cerr << "-S <from>  the second (float) to start analysing from" <<endl;
    cerr << "           Default: 0.0" << endl;
    cerr << "-E <to>    the second (float) to stop analysing at" << endl;
    cerr << "           Default: end of file" << endl;
    cerr << "-V         verbose" << endl;
    cerr << "-H <0/1>   0=gives header info on first audio frame" << endl;
    cerr << "           1=gives header info on all audio frames" << endl;
    cerr << "           Default: 0" << endl;
    cerr << "-I <0/1>   0=gives Layer III side info on first frame" <<endl;
    cerr << "           1=gives Layer III side info on all frames" << endl;
    cerr << "           Default: 0" << endl;
    cerr << "-D         calculates total duration of audio file" << endl;
    cerr << "-B <int> <int> subband range that gets extracted" << endl;
    cerr << "           (only required with -X" << endl;
    cerr << "           Default: 0 31 (Layers I, II)" << endl;
    cerr << "                    0 575 (Layer III)" << endl;
    cerr << "-C <1/2>   1=extract only first channel" << endl;
    cerr << "           2=extract both channels (if existant)" << endl;
    cerr << "           Default: 1" << endl;
    cerr << "-X [field] extracts info from fields from audio file" << endl;
    cerr << "           Possible field types depend on the Layer" << endl;
    cerr << "           at which the audio file is encoded:" << endl;
    cerr << "   bitalloc (Layers I, II) -B fromsb tosb -C 1/2" << endl;
    cerr << "   scalefac (Layers I, II, III) -B fromsb tosb -C 1/2" <<endl;
    cerr << "   scfsi    (Layers II, III) -B fromsb tosb -C 1/2" << endl;
    cerr << "   freqval  (Layers I, II, III) -B fromsb tosb -C 1/2" <<endl;
    cerr << "   pcmval   (Layers I, II, III) -C 1/2" << endl;
    cerr << "-O <file>  output pcm samples to file" << endl;
    cerr << "           Default: file=out.raw" << endl;
    exit (1);
}


void 
ParseArguments (int argc, char **argv) {
    // go through parameters and check values
    for (int i=1; i<argc; i++) {
	if (argv[i][0] == '-' && argv[i][1]) {
	    switch (argv[i][1]) {
	    case 'h': // help
		PrintUsage(argv[0]);
		break;
	    case 'S': case 's':// second to start from
		if ((i+1<argc) && !isalpha(argv[i+1][0])) {
		    from = atof(argv[++i]);
		}
		break;
	    case 'E': case 'e':// second to end at
		if ((i+1<argc) && !isalpha(argv[i+1][0])) {
		    to = atof(argv[++i]);
		}
		break;
	    case 'V': case 'v':// verbose
	        verbose = true;
	        break;
	    case 'H': // header
	        printHeader = 0;
		if ((i+1<argc) && !isalpha(argv[i+1][0])) {
		    printHeader = atoi(argv[++i]);
		}
		break;
	    case 'I': case 'i':// side information
	        printSideinfo = 0;
		if ((i+1<argc) && !isalpha(argv[i+1][0])) {
		    printSideinfo = atoi(argv[++i]);
		}
		break;
	    case 'D': case 'd':// total duration
	        printDur = true;
	        break;
	    case 'B': case 'b':// subband range
		if ((i+2<argc) && !isalpha(argv[i+1][0]) 
		    && !isalpha(argv[i+2][0])) {
		    fromSb = atoi(argv[++i]);
		    toSb   = atoi(argv[++i]);
		} else {
		    cerr << "inspectMPaudio: -B option needs two integers to follow"
			 << endl;
		    exit(1);
		}
		break;
	    case 'C': case 'c':// channels
		if ((i+1<argc) && !isalpha(argv[i+1][0])) {
		    channels = atoi(argv[++i]);
		}
		break;
	    case 'X': case 'x':// fields
		if ((i+1 == argc) || !isalpha(argv[i+1][0])) {
		    cerr << "inspectMPaudio: -X option needs string-argument"
			 << endl;
		    exit(1);
		} else {
		    if (!strncmp(argv[i+1],"bitalloc",3)) {
			wantBitalloc = true; i++;
		    } else if (!strncmp(argv[i+1],"scalefac",3)) {
			wantScf = true; i++;
		    } else if (!strncmp(argv[i+1],"scfsi",3)) {
			wantScfsi = true; i++;
		    } else if (!strncmp(argv[i+1],"freqval",3)) {
			wantValues = true; i++;
		    } else if (!strncmp(argv[i+1],"pcmval",3)) {
			wantPcm = true; i++;
		    } else if (argv[i+1][0] != '-') {
			cerr << "inspectMPaudio: -X unknown string-argument "
			     << argv[i+1] << endl;
			exit(1);
		    }
		}
		break;
	    case 'O': case 'o': case '0': // output pcm samples to file
	        wantPcmFile = true;
		if ((i+1<argc) && isalpha(argv[i+1][0]) 
		    && (argv[i+1][0] != '-')) {
		    outfile = string(argv[++i]);
		}
		break;
	    } // switch
	} else {
	    filename = argv[i];
	}
    }
}


void
CheckArguments( ) {
#ifdef DEBUG
    cout << "Parameter Settings:"             << endl;
    cout << "From="          << from          << endl;
    cout << "To="            << to            << endl;
    cout << "Verbose="       << verbose       << endl;
    cout << "printHeader="   << printHeader   << endl;
    cout << "printSideinfo=" << printSideinfo << endl;
    cout << "printDur="      << printDur      << endl;
    cout << "fromSb="        << fromSb        << endl;
    cout << "toSb="          << toSb          << endl;
    cout << "channels="      << channels      << endl;
    cout << "wantBitalloc="  << wantBitalloc  << endl;
    cout << "wantScf="       << wantScf       << endl;
    cout << "wantScfsi="     << wantScfsi     << endl;
    cout << "wantValues="    << wantValues    << endl;
    cout << "wantPcm="       << wantPcm       << endl;
    cout << "wantPcmFile="   << wantPcmFile   << endl;
    cout << "outfile="       << outfile       << endl;
    cout << "filename="      << filename      << endl;
#endif
    if (filename == NULL) {
	cerr << "inspectMPaudio: Mpeg audio filename missing" << endl;
	exit(1);
    }
    if (channels < 1)  channels = 1;
    if (channels > 2)  channels = 2;
    if (fromSb   < 0)  fromSb   = 0;
    if (toSb < fromSb) toSb     = fromSb;
    if (printSideinfo > 1) printSideinfo = 1;
    if (printHeader   > 1) printHeader   = 1;
    if (from < 0.0)    from = 0.0;
    if (to   < from)   to   = from;
}


int 
main(int argc, char **argv) {

    // check no. of args
    if (argc < 2) {
	PrintUsage(argv[0]);
    }

    // parse arguments
    ParseArguments (argc, argv);

    // check arguments and correct if appropriate
    CheckArguments ();

    // open file
    sfile = new SOUNDfile(filename);

    if (sfile->file_type() == MPEG) {
    infile = (MPEGfile *) sfile->format;
    } else {
      infile = NULL;
    }
    if (verbose) {
	cout << "Audio file opened: " << sfile->file() << endl;
    }

    // open pcm output file
    if (wantPcmFile) {
	outpcm=fopen(outfile.c_str(),"wb");
    }

    // position at start second
    if (!sfile->seek_time(from)) {
	cerr << "inspectMPaudio: error positioning at starting sec (" 
	     << from << ") failed." << endl;
	exit(1);
    }
    if (verbose) cout << "starting at " << sfile->at_time() << " s" << endl;


    // go through file parsing and displaying
    int ch = 0;

    while(sfile->data_available() && (sfile->at_time() <= to)) {
      // parse next frame
      if ( wantPcm||wantPcmFile) {
	infile->goTo_nextFrame(PCM);
      } else {
	sfile->next_window(PCM);
      }
	// print header if requested
	if (printHeader >= 0 && sfile->file_type() == MPEG) {
	    infile->printheader();
	    if (printHeader == 0) printHeader = -1;
	}

	// print side info if requested
	if (printSideinfo >= 0 && sfile->file_type() == MPEG) {
	    infile->printSideinfo();
	    if (printSideinfo == 0) printSideinfo = -1;
	}

	// check requested subbands for this frame
	int fromSub = fromSb;
	int toSub   = toSb;
	if (sfile->file_type() == MPEG) {
	if (infile->layer() == III) {
	    if (fromSub > 576) fromSub = 576;
	    if (toSub   > 576) toSub   = 576;
	} else {
	    if (fromSub > SBLIMIT) fromSub = SBLIMIT;
	    if (toSub   > SBLIMIT) toSub   = SBLIMIT;
	}
	}

	// check channels
	ch = min ((int) channels, sfile->channels());

	// print requested information:

	// Bitallocation information
	if (wantBitalloc) {
	  if (sfile->file_type() == MPEG) {
	    if (infile->layer() == III) {
		cerr << "inspectMPaudio: Layer III has no bitallocation information" 
		     << endl;
	    } else {
		for (int i=fromSub; i<toSub; i++) { // subbands
		    for (int j=0; j<ch; j++) { // channels
			cout << "Allocation["<<j<<","<<i<<"]="
			     <<infile->bitallocation(j,i)<<endl;
		    }
		}
	    }
	  } else {
	    cout << "Not an MPEG file: no bitallocation available."<<endl;
	  }
	}

	// Scalefactors
	if (wantScf) {
	  if (sfile->file_type() == MPEG) {
	    // print scalefactors
	    for (int i=fromSub; i<toSub; i++) { // subbands
	      for (int j=0; j<ch; j++) { // channels
		cout << "Scalefactor["<<j<<","<<i<<"]="
		     <<infile->scalefactor(j,i)<<endl;
	      }
	    }
	  } else {
	    cout << "Not an MPEG file: no scakefactors available."<<endl;
	  }
	}   
	
	// Scalefactor selection information
	if (wantScfsi) {
	  if (sfile->file_type() == MPEG) {
	    if (infile->layer() == I) {
		cerr << "inspectMPaudio: Layer I has no scalefactor selection information" 
		     << endl;
	    } else {
		if (infile->layer() != III) {
		    for (int i=fromSub; i<toSub; i++) { // subbands
			for (int j=0; j<ch; j++) { // channels
			    cout << "Scfsi["<<j<<","<<i<<"]="
				 <<infile->scfsi(j,i)<<endl;
			}
		    }
		} else { // layer III
		    for (int i=0; i<4; i++) {
			for (int j=0; j<ch; j++) { // channels
			    cout << "Scfsi["<<j<<","<<i<<"]="
				 << infile->scfsi(j,i) <<endl;
			}
		    }
		}
	    }
	  } else {
	    cout << "Not an MPEG file: no scfis available."<<endl;
	  }

	}

	// Frequency values (spectral values)
	if (wantValues) {
	  //no. samples
	  for (unsigned int i=0; i<sfile->timeticks(HIGH); i++) {
	    for (int sb=fromSub; sb<toSub; sb++) { //subbands
	      for (int j=0; j<ch; j++) { //channels
		    cout << "FreqVal[ch="<<j<<",sb="<<sb<<",timeticks="
			 <<i<<"]="
			 <<sfile->freq_value(j,sb,i,HIGH)<<endl;
	      }
	    }
	  }
	  /*
	    if (infile->layer() != III) {
	      for (int s=0; s<12; s++) { // no. samples
		for (int i=fromSub; i<toSub; i++) { // subbands
		  for (int j=0; j<ch; j++) { // channels
		    //				cout << "Sample["<<j<<","<<i<<","
		    //				     <<s<<","<<k<<"]="
		    //				     <<infile->sample(j,i,s,k)<<endl;
		    cout << "FreqVal["<<j<<","<<i<<","
			 <<s<<"]="
			 <<infile->restored_sample(j,i,s)<<endl;
		  }
		}
	      }
	    } else { // Layer III (can it be integrated with others?)
	      for (int j=0; j<ch; j++) { // channels
		for (int sb=fromSub; sb<toSub; sb++) {
		  int subb = sb/SSLIMIT;
		  int ss   = sb%SSLIMIT;
		  //			    cout << "Sample["<<j<<","<<gr<<","
		  //				 << sb<<"]="
		  //				 << infile->sample(j,gr,subb,ss) 
		  //				 << endl;
		  cout << "Restored_Sample["<<j<<","
		       << subb<<","<<ss<<"]="
		       << infile->restored_sample(j,subb,ss)
		       << endl;
		}
	      }
	    }
	  */
	}
    
    

	// PCM samples
	if (wantPcm||wantPcmFile) {

	  for (unsigned int i=0; 
	       i<(sfile->timeticks(HIGH)*sfile->nb_subbands(HIGH)); 
	       i++) {
	    for (int j=0; j<ch; j++) { //channels
	      cout << "pcm[ch="<<j<<",i="<<i<<"]="
		   <<sfile->pcm(j,i)<<endl;
	      }
	    }
	  }

	/*
	  if (infile->layer() != III) {
	    int kmax=1; // Layer I
	    short sample;
	    if (infile->layer() == II) kmax=3; // Layer II
	    for (int s=0; s<12; s++) { // no. samples
	      for (int k=0; k<kmax; k++) {
		for (int i=fromSub; i<toSub; i++) { // subbands
		  for (int j=0; j<ch; j++) {
				if (wantPcm) {
				  cout << "PCM["<<j<<","<<i<<","
				       <<s<<","<<k<<"]="
				       <<infile->pcm_sample(j,i,s,k)<<endl;
				} else { // wantPcmFile
				  sample = infile->pcm_sample(j,i,s,k);
				  fwrite(&sample, 1, 2, outpcm);
				}
		  }
		}
	      }
	    }
	  } else { // Layer III (can it be integrated with others?)
	    short sample;
	    for (int gr=0; gr<(2-infile->version()); gr++) {
	      for (int sb=fromSub; sb<toSub; sb++) {
		int subb = sb%SBLIMIT;
		int ss   = sb/SBLIMIT;
		for (int j=0; j<ch; j++) { // channels
		  if (wantPcm) {
		    cout << "PCM_Sample["<<j<<","<<gr<<","
			 <<ss <<","<< subb<<"]="
			 <<infile->pcm_sample(j,gr,subb,ss)<<endl;
		  } else { // wantPcmFile
		    sample = infile->pcm_sample(j,gr,subb,ss);
		    fwrite(&sample, 1, 2, outpcm);
		  }
		}
	      }
	    }
	  }
	}
	

	// Layer 3
	// side information
	/*
	for (int ch=0; ch<infile->channels(); ch++) {
	    for (int i=0; i<4; i++) {
		cout << "Scfsi["<<ch<<","<<i<<"]="
		     << infile->scfsi(ch,i) <<endl;
	    }
	    for (int gr=0; gr<(2-infile->version()); gr++) {
		cout << "Part2_3_Length["<<ch<<","<<gr<<"]="
		     << infile->part2_3_length(ch,gr) << endl;
		cout << "Big_Values["<<ch<<","<<gr<<"]="
		     << infile->big_values(ch,gr) << endl;
		cout << "Global_Gain["<<ch<<","<<gr<<"]="
		     << infile->global_gain(ch,gr) << endl;
		cout << "Scalefac_Compress["<<ch<<","<<gr<<"]="
		     << infile->scalefac_compress(ch,gr) << endl;
		cout << "Window_Switching["<<ch<<","<<gr<<"]="
		     << infile->window_switching(ch,gr) << endl;
		cout << "Block_Type["<<ch<<","<<gr<<"]="
		     << infile->blocktype(ch,gr) << endl;
		cout << "Mixedblock["<<ch<<","<<gr<<"]="
		     << infile->mixedblock(ch,gr) << endl;
		for (int j=0; j<3; j++) {
		    cout << "table_select["<<ch<<","<<gr<<","<<j<<"]="
			 << infile->table_select(ch,gr,j) << endl;
		    cout << "subblock_gain["<<ch<<","<<gr<<","<<j<<"]="
			 << infile->subblock_gain(ch,gr,j) << endl;
		}
		cout << "preflag["<<ch<<","<<gr<<"]="
		     << infile->preflag(ch,gr) << endl;
		cout << "Scalefac_Scale["<<ch<<","<<gr<<"]="
		     << infile->scalefac_scale(ch,gr) << endl;
		cout << "Count1Table_select["<<ch<<","<<gr<<"]="
		     << infile->count1table_select(ch,gr) << endl;
		cout << "region0_samps["<<ch<<","<<gr<<"]="
		     << infile->region0_samps(ch,gr) << endl;
		cout << "region1_samps["<<ch<<","<<gr<<"]="
		     << infile->region1_samps(ch,gr) << endl;
		cout << "region2_samps["<<ch<<","<<gr<<"]="
		     << infile->region2_samps(ch,gr) << endl;
	    }
	}
	*/

    }

    if (printDur) {
	cout << "No. of Frames analyzed: " << sfile->at_window() << endl;
	cout << "stopped at " << sfile->at_time() << " s" << endl;
    }


    // close output sound file
    if (wantPcmFile) {
	fclose (outpcm);
	cout << "PCM file written (16 bit linear, "
	     << sfile->sampling_rate() << " kHz samplingrate, "
	     << ch << " channels): "
	     << outfile.c_str() << endl;
    }
    
    exit (0);
}
