/*
    MPEG Maaate: An Australian MPEG audio analysis toolkit
    Copyright (C) 2000 Commonwealth Scientific and Industrial Research Organisation
    (CSIRO), Australia.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include "header.H"

#define SYNC_WORD        (long) 0xfff
#define CRC16_POLYNOMIAL 0x8005
 
// 0: MPEG1 Layer1-3, 1: MPEG2 Layer1-3
static const int bitrates[2][3][16] = {
    {{0,32,64,96,128,160,192,224,256,288,320,352,384,416,448,0},
     {0,32,48,56,64,80,96,112,128,160,192,224,256,320,384,0},
     {0,32,40,48,56,64,80,96,112,128,160,192,224,256,320,0}},
    
    {{0,32,48,56,64,80,96,112,128,144,160,176,192,224,256,0},
     {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,0},
     {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,0}}
};

// 0: MPEG1, 1: MPEG2
static const double samplingrates[2][4] = { {44.1, 48, 32, 0},
					    {22.05, 24, 16, 0} 
};

// Layer 1-3 mode_ext
static const unsigned int jointstereo[3][4] =  { { 4, 8, 12, 16 }, 
					{ 4, 8, 12, 16 },
					{ 0, 4, 8, 16  } 
};


//--------------------------------------------------------------------------

// which mpeg version
CSAPI_MPEG Version 
Header::version()   {
    return (Version) (header.version^1);
}


// which audio layer 
CSAPI_MPEG Layer 
Header::layer()     {
    return (Layer) (3-header.layer);
}


// is the frame checksum protected?
CSAPI_MPEG bool 
Header::crcprotected() {
    return (bool) !header.errorprotection;
}


// which bitrate?
// gives no. of slots a frame consists of (+1 if padded)
CSAPI_MPEG int 
Header::bitrate()   {
  return c_bitrate;
}
CSAPI_MPEG int 
Header::p_bitrate()   {
  return bitrates[header.version^1][3-header.layer][header.bitrate];
}


// which bitrate per channel?
CSAPI_MPEG int 
Header::bitrateperchannel() {
  return c_bitrateperchannel;
}
CSAPI_MPEG int 
Header::p_bitrateperchannel() {
  return (bitrate()/channels());
}


// return bitrate index
CSAPI_MPEG int 
Header::bitrate_index() {
    return header.bitrate;
}


// which samplingrate?
CSAPI_MPEG double 
Header::samplingrate() {
    return samplingrates[header.version^1][header.samplingrate];
}


// which samplingrate index
CSAPI_MPEG int 
Header::samplingrate_index() {
    return header.samplingrate;
}


// is a slot added to audio frame to adjust mean bitrate to samplingrate?
// LI: slot size=4bytes
// LII/III:     =1byte
CSAPI_MPEG bool 
Header::padding() {
    return (bool) header.padding;
}


// bit that can be used for private purposes.
CSAPI_MPEG bool 
Header::extended() {
    return (bool) header.extension;
}


// channel modes
CSAPI_MPEG Mode 
Header::mode() {
    return (Mode) header.mode;
}


CSAPI_MPEG const char *
Header::mode_ext_str (void)
{
    if (header.mode == jstereo) {
	switch (header.mode_ext){
	case 0:
	    if (3-header.layer == III) {
		//  intensity stereo off, ms stere off
		return "normal stereo";
	    } else {
		return "subbands 4-31 joint";
	    }
	case 1:
	    if (3-header.layer == III) {
		//  intensity stereo on, ms stere off
		return "intensity stereo";
	    } else {
		return "subbands 8-31 joint";
	    }
	case 2:
	    if (3-header.layer == III) {
		//  intensity stereo off, ms stereo on
		return "MS stereo";
	    } else {
		return "subbands 12-31 joint";
	    }
	case 3:
	    if (3-header.layer == III) {
		//  intensity stereo on, ms stereo on
		return "intensity & MS stereo";
	    } else {
		return "subbands 16-31 joint";
	    }
	default:
	    return NULL;
	}
    } else {
	return "unused";
    }
}


CSAPI_MPEG const char *
Header::mode_str (void)
{
  switch (header.mode){
  case stereo:
      // two audio channels (left + right) encoded per bitstream
      return "stereo";
  case jstereo:
      // stereophonic irrelevance exploited
      // L I, II: intensity stereo (for some high subbands, only sum-signal
      //          is encoded; scalefactors are preserved for both channels)
      // L III:   combination of intensity stereo and MS stereo (coding sum
      //          and difference signal instead of left and right)
      return "joint stereo";
  case dualch:
      // two audio channels (e.g. 2 languages) encoded per bitstream
      return "dual channel";
  case singlech:
      // one audio channel encoded per bitstream
      return "single channel";
  case multich:
      // 
      return "multiple channels";
  default:
      return NULL;
  }
}


// returns number of channels.
CSAPI_MPEG int 
Header::channels() {
  return c_channels;
}

CSAPI_MPEG int 
Header::p_channels() {
    return ((header.mode == singlech) ? 1 : 2);
}


// for jointstereo in LI and LII indicates subbands coded in intensity_stereo
//                 in LIII indicates which type of jointstereo coding is used
CSAPI_MPEG int 
Header::mode_ext() {
    return header.mode_ext;
}


// returns the subband number from which on the upper subbands are
// coded in intensity stereo
CSAPI_MPEG unsigned int 
Header::intensitystereobound(){
  return c_intensitystereobound;
}

CSAPI_MPEG unsigned int 
Header::p_intensitystereobound(){
    if (header.mode == jstereo) {
	return min(jointstereo[3-header.layer][header.mode_ext],
		   subbands());
    } else {
	return subbands();
	}
}


// copyrighted stream?
CSAPI_MPEG bool 
Header::copyright(){
    return header.copyright;
}


// original or copied stream?
CSAPI_MPEG bool 
Header::original() {
    return header.original;
}


// emphasis? (filtering is applied to improve the SNR at high frequencies)
CSAPI_MPEG int 
Header::emphasis() {
    return header.emphasis;
}


// calculates number of samples per frame
CSAPI_MPEG unsigned int 
Header::samples_per_frame() {
    Layer lay = (Layer) (3-header.layer);

    if (lay == I) {
	return 384; // 32 subbands * 12 samples
    }
    if (lay == II) {
	return 1152; // 32 subbands * 3 * 12 samples
    }
    if (lay == III) {
	if ( (header.version^1) == MPEG1) {
	    return 1152; // 32 subbands * 18 samples * 2 granules
	} else {
	    return 576; // 32 subbands * 18 samples
	}
    }
    return 0;
}


// calculates framesize in number of bytes
CSAPI_MPEG unsigned int 
Header::framesize() 
{
  return c_framesize;
}

CSAPI_MPEG unsigned int 
Header::p_framesize() 
{
    unsigned int bytesize;
    Layer lay = (Layer) (3-header.layer);

    if (lay==I) {
	bytesize = (int) ((12 * bitrate()) / samplingrate());
    } else {
	bytesize = (int) ((144 * bitrate()) / samplingrate());
    }
    if (lay==III && (header.version^1)==MPEG2) {
	bytesize /= 2; // 1 slot = 4 bit in layer III, MPEG2
    }
    if (samplingrate()==44.1 && header.padding) {
        bytesize++;
    }
    if (lay==I) {  // 1 slot = 1 byte in layer II, III
	bytesize *= 4; // 1 slot = 4 bytes in layer I
    }
    bytesize -= 4; // deduct header size
    if ( !header.errorprotection ) { // deduct crc size
	bytesize -= 2;
    }
 
    return bytesize; // framesize as number of bytes
}


// calculate number of subbands for layer II
// or returns 32 for all other layers
CSAPI_MPEG unsigned int 
Header::subbands() 
{
  return c_subbands;
}

CSAPI_MPEG unsigned int 
Header::p_subbands() 
{
    if ((3-header.layer) == II) {
	if ( (header.version^1) == MPEG1) {
	    if (bitrateperchannel() == 32 || 
		bitrateperchannel() == 48)
		if (samplingrate() == 32)
		    return 12; // table=3
		else
		    return 8; // table=2
	    else
		if (samplingrate() == 48 || 
		    (bitrateperchannel() >= 56 && 
		     bitrateperchannel() <= 80))
		    return 27; // table=0
		else
		    return 30; // table=1
	} else { // MPEG2
	    return 30; // table=4
	}
    } else {
	return 32; // SBLIMIT Layer 1 + 3
    }
}


// return checksum
CSAPI_MPEG unsigned int 
Header::check_sum()   {
    return checksum;
}


//  checks validity of fields and caches values
CSAPI_MPEG bool 
Header::checkheader () {
    Layer lay = (Layer) (3-header.layer);

    // check validity of header fields
    if ((header.sync & SYNC_WORD) != SYNC_WORD) {
	cerr << "MaaateP: Wrong sync word:" << hex << header.sync << dec << endl;
	return false;
    }
    if (header.layer == 0) {
	cerr << "MaaateP: Unknown layer:" << header.layer << endl;
	return false;
    }
    if (header.bitrate == 0) {
	cerr << "MaaateP: Free format bitrate not implemented" << endl;
	return false;
    }
    if (header.bitrate == 15) {
	cerr << "MaaateP: Unknown bitrate index: 15" << endl;
	return false;
    }
    if (header.samplingrate == 3) {
	cerr << "MaaateP: Unknown sampling rate index: 3" << endl;
	return false;
    }
    if (lay==II && 
	((header.mode==singlech && header.bitrate>=11) ||
	 (header.mode!=singlech && 
	  ((header.bitrate >= 1 && header.bitrate <= 3) ||
	   (header.bitrate == 5))))) {
	cerr << "MaaateP: Wrong bitrate-mode-combination in layer II:\n"
	     << "bitrate=" << header.bitrate
	     << "\tmode=" << header.mode
	     << endl;
	return false;
    }
    if (lay==III && header.mode==jstereo && header.mode_ext==0) {
	header.mode=0; //stereo
    }

    // cache values
    c_bitrate              = p_bitrate();
    c_channels             = p_channels();
    c_bitrateperchannel    = p_bitrateperchannel();
    c_intensitystereobound = p_intensitystereobound();
    c_framesize            = p_framesize();
    c_subbands             = p_subbands();

    return true;
}


CSAPI_MPEG void 
Header::printheader()
{
    cout << "Decoded header content:" << endl;
    cout << "Syncword="     << hex << header.sync  << dec << endl;
    cout << "Version="      << version() 
	 << "(0=Mpeg1, 1=Mpeg2)" << endl;
    cout << "Layer="        << layer()+1 << endl;
    cout << "Protected="    << crcprotected() << endl;
    cout << "Bitrate="      << bitrate() << " kbit/s" << endl;
    cout << "Samplingrate=" << samplingrate() << " kHz" << endl;
    cout << "Padding="      << padding() << endl;
    cout << "Extended="     << extended() << endl;
    cout << "Mode="         << mode() 
	 << " (" << mode_str() << ")" << endl;
    cout << "Mode_ext="     << mode_ext()
	 << " (" << mode_ext_str() << ")" << endl;
    cout << "Copyright="    << copyright() << endl;
    cout << "Original="     << original() << endl;
    cout << "Emphasis="     << emphasis() << endl;

#ifdef DEBUG
    cout << "Intensitystereobound=" << intensitystereobound() << endl;
    cout << "Channels=" << channels() << endl;
    cout << "Subbands=" << subbands() << endl;
    cout << "Framesize=" << framesize() << " byte" << endl;
#endif

}


CSAPI_MPEG void 
Header::update_CRC (unsigned int data, 
		       unsigned int length,
		       unsigned int *crc)
{
    unsigned int  masking, carry;
    
    masking = 1 << length;
    
    while((masking >>= 1)) {
        carry = *crc & 0x8000;
        *crc <<= 1;
        if (!carry ^ !(data & masking))
            *crc ^= CRC16_POLYNOMIAL;
    }
    *crc &= 0xffff;
    return;
}
