/*
    MPEG Maaate: An Australian MPEG audio analysis toolkit
    Copyright (C) 2000 Commonwealth Scientific and Industrial Research Organisation
    (CSIRO), Australia.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include        "layer1.H"


//--------------------------------------------------------------------------

// constructor
CSAPI_MPEG
Layer1::Layer1(MPEGfile *mp) 
{
  // set reference to audio frame handler
  frame = mp;

  return;
}

// parse audio data of layer 1 frame
CSAPI_MPEG bool 
Layer1::parse_data(DecodeLevel decode) 
{
  // parse bit allocation information
  decode_bitalloc();

  // check crc checksum
  if (frame->crcprotected()) {
    if (frame->checksum != calc_CRC()) {
      cerr << "MaaateP: WARNING: wrong checksum" << endl;
      return false;
    }
  }

  // parse scalefactor information
  decode_scale();

  // parse subband samples
  buffer_samples();

  if (decode > dec_fields) {
    restore_samples();

    if (decode > dec_subbands) {
      // restore pcm samples
      int ch=frame->channels();
      for (int s=0; s<12; s++) {
	for (int j=0; j<ch; j++) {
	  subband_syn (&(restoredsamples[s][j][0]),
		       j,
		       &(pcm[s][j][0]));
	}
      }
    }
  }

  return true;
}

// calculates checksum on required fields
CSAPI_MPEG unsigned int 
Layer1::calc_CRC()
{
  unsigned int crc = 0xffff;
  frame->update_CRC(frame->header.bitrate,         4, &crc);
  frame->update_CRC(frame->header.samplingrate,    2, &crc);
  frame->update_CRC(frame->header.padding,         1, &crc);
  frame->update_CRC(frame->header.extension,       1, &crc);
  frame->update_CRC(frame->header.mode,            2, &crc);
  frame->update_CRC(frame->header.mode_ext,        2, &crc);
  frame->update_CRC(frame->header.copyright,       1, &crc);
  frame->update_CRC(frame->header.original,        1, &crc);
  frame->update_CRC(frame->header.emphasis,        2, &crc);
  
  int i, k; //index variable
  int bound = frame->intensitystereobound();
  int ch = frame->channels();
  for (i=0; i<bound; i++) {
    for (k=0; k<ch; k++) {
      frame->update_CRC(allocation[k][i], 4, &crc);
    }
  }
  for (i=bound; i<SBLIMIT; i++) {
    frame->update_CRC(allocation[0][i], 4, &crc);
  }
  
  return crc;
}


// The bit allocation information, i.e. how many bits are used to code
// a sample, is decoded. Layer I has 4 bit per subband.
CSAPI_MPEG void 
Layer1::decode_bitalloc()
{
  int i, j; // index varibale

  int bound = frame->intensitystereobound();
  int ch = frame->channels();
  for (i = 0; i<bound; i++)
    for (j=0; j<ch; j++) {
      allocation[j][i] = frame->readbitsfrombuffer(4);
    }
  for (i=bound; i<SBLIMIT; i++) {
    allocation[0][i] = frame->readbitsfrombuffer(4);
    allocation[1][i] = allocation[0][i];
  }
  return;
}

// Layer 1 format of scale factor extraction involves reading
// 6 bit per subband as scale factor. One scale factor is
// encoded per group of 12 samples encoded.
CSAPI_MPEG void 
Layer1::decode_scale()
{
  int i, j; // index varibale
  int ch = frame->channels();
  for (i=0; i<SBLIMIT; i++)
    for (j=0; j<ch; j++)
      if (!allocation[j][i]) {
	scale_index[j][i] = SCALE_RANGE-1;
      } else  {   /* 6 bit per scale factor */
	scale_index[j][i] =  frame->readbitsfrombuffer(6);
      }
  return;
}

// store compressed sample from the bit stream for layer 1:
// read the number of bits as indicated by the allocation table information
// and store the group of 12 samples in the samples buffer.
CSAPI_MPEG void 
Layer1::buffer_samples()
{
  int i,j,k;
  unsigned int samp;
  int bound = frame->intensitystereobound();
  int ch = frame->channels();
  for (int s=0; s<12; s++) {
    for (i=0; i<bound; i++) {
      for (j=0; j<ch; j++) {
	if ( (k = allocation[j][i]) == 0) {
	  samples[s][j][i]= 0;
	} else {
	  samples[s][j][i] = frame->readbitsfrombuffer(k+1);
	}
      }
    }
    for (i=bound; i<SBLIMIT; i++) {
      if ( (k = allocation[0][i]) == 0) {
	samp = 0;
      } else { 
	samp = frame->readbitsfrombuffer(k+1);
      }
      for (j=0; j<ch; j++) {
	samples[s][j][i] = samp;
      }
    }
  }
  return;
}


// restore the compressed sample to a fractional number:
// - complement the MSB of the sample (dequantize)
// Use s = (s' + 2^(-nb+1) ) * 2^nb / (2^nb-1)
// where nb=bits_per_sample()
//       s'=fractional number, i.e. samples[][] with first bit inverted
// - multiply dequantized sample by scale factor (denormalize)
CSAPI_MPEG void 
Layer1::restore_samples()
{
  unsigned int nb;
  int ch = frame->channels();
  for (int s=0; s<12; s++) {
    for (int i=0; i<SBLIMIT; i++)
      for (int k=0; k<ch; k++) {
	if (allocation[k][i]) {
	  nb=bits_per_sample(k,i);

	  // invert first bit
	  if ((((samples[s][k][i]) >> (nb-1)) & 1) == 1)
	    restoredsamples[s][k][i] = 0.0;
	  else
	    restoredsamples[s][k][i] = -1.0;
	  
	  // apply formula
	  restoredsamples[s][k][i] += 
	    (double) (samples[s][k][i] & ((1 << (nb-1)) - 1)) / 
	    (double) (1L << (nb-1));
	  
	  restoredsamples[s][k][i] =
	    (double) (restoredsamples[s][k][i]+1.0/(1L << (nb-1))) *
	    (double) (1L<<nb) / (double) ((1L<<nb)-1);
	} else {
	  restoredsamples[s][k][i] = 0.0;
	}
	
	// denormalize (rescaling)
	restoredsamples[s][k][i] *= scalefactor(k,i);
      }
  }
}

CSAPI_MPEG int 
Layer1::bitallocation (unsigned int channel, unsigned int subband) 
{
    return allocation[channel][subband];
}

CSAPI_MPEG unsigned int 
Layer1::bits_per_sample (unsigned int channel, unsigned int subband) 
{
    if (allocation[channel][subband]==0 || allocation[channel][subband]==15)
	return 0;
    else
	return (allocation[channel][subband] + 1);
}

CSAPI_MPEG float 
Layer1::scalefactor (unsigned int channel, unsigned int subband, unsigned int ss) 
{
    return maaate_scalefactors[scale_index[channel][subband]];
}

CSAPI_MPEG int 
Layer1::sample (unsigned int channel, unsigned int subb, unsigned int number, unsigned int ss) 
{
    return samples[number][channel][subb];
}

CSAPI_MPEG double 
Layer1::restored_sample (unsigned int ch, unsigned int subb, unsigned int no, unsigned int ss) 
{
    return restoredsamples[no][ch][subb];
}

CSAPI_MPEG short 
Layer1::pcm_sample (unsigned int ch, unsigned int subb, unsigned int no, unsigned int ss) 
{
    return pcm[no][ch][subb];
}
