/*
    MPEG Maaate: An Australian MPEG audio analysis toolkit
    Copyright (C) 2000 Commonwealth Scientific and Industrial Research Organisation
    (CSIRO), Australia.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include "layer3.H"

// scalefactor bands for which a scfsi is transmitted (0-5,6-10...)
static const unsigned int 
sfbtable[2][5] = {{0, 6, 11, 16, 21},  // long blocks
		  {0, 6, 12,  0,  0}}; // short blocks

// structure to store bitallocation for scale factors
static const unsigned int slen[2][16] = 
{{0, 0, 0, 0, 3, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4},
 {0, 1, 2, 3, 0, 1, 2, 3, 1, 2, 3, 1, 2, 3, 2, 3}};

// structure to take widths of each scalefactor band
struct  SFBandIndex {
  unsigned int l[23];
  unsigned int s[14];
};

// width of each scale factor band (B.8 ISO/IEC 11172-3)
static const SFBandIndex sfBandIndex[6] = {
  {{0,4,8,12,16,20,24,30,36,44,52,62,74,90,110,134,162,196,238,288,342,418,576},{0,4,8,12,16,22,30,40,52,66,84,106,136,192}},
  {{0,4,8,12,16,20,24,30,36,42,50,60,72,88,106,128,156,190,230,276,330,384,576},{0,4,8,12,16,22,28,38,50,64,80,100,126,192}},
  {{0,4,8,12,16,20,24,30,36,44,54,66,82,102,126,156,194,240,296,364,448,550,576},{0,4,8,12,16,22,30,42,58,78,104,138,180,192}},
  {{0,6,12,18,24,30,36,44,54,66,80,96,116,140,168,200,238,284,336,396,464,522,576},{0,4,8,12,18,24,32,42,56,74,100,132,174,192}},
  {{0,6,12,18,24,30,36,44,54,66,80,96,114,136,162,194,232,278,330,394,464,540,576},{0,4,8,12,18,26,36,48,62,80,104,136,180,192}},
  {{0,6,12,18,24,30,36,44,54,66,80,96,116,140,168,200,238,284,336,396,464,522,576},{0,4,8,12,18,26,36,48,62,80,104,134,174,192}}
};

// preemphasis table (B.6 ISO/IEC 11172-3)
static const unsigned int 
pretab[22] = {0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,2,2,3,3,3,2,0};

// coefficients for aliasing reduction (B.9 ISO/IEC 11172-3)
static const double 
Ci[8] = {-0.6,-0.535,-0.33,-0.185,-0.095,-0.041,-0.0142,-0.0037};

// mpeg2
static const unsigned int nr_of_sfb_block[6][3][4] = {
  {{6, 5, 5, 5},{ 9, 9, 9, 9 },{6, 9, 9, 9}},
  {{6, 5, 7, 3},{ 9, 9, 12, 6},{6, 9, 12, 6}},
  {{11, 10, 0, 0},{ 18, 18, 0, 0},{15,18,0,0 }},
  {{7, 7, 7, 0},{ 12, 12, 12, 0},{6, 15, 12, 0}},
  {{6, 6, 6, 3},{12, 9, 9, 6},{6, 12, 9, 6}},
  {{8, 8, 5, 0},{15,12,9,0},{6,18,9,0}}
};

#define HTN      34  // no of huffman tables
#define MXOFF    250 // for huffman decoding
 
// huffman code tables from B.7 ISO/IEC 11172-3
/* .table  0   0  0  0  0*/
/* table 0 has no tree data*/

/* .table  1   7  2  2  0*/
static const unsigned char treedata_1[] = {
  0x2, 0x1, 0x0, 0x0, 0x2, 0x1, 0x0, 0x10, 0x2, 0x1, 0x0, 0x1, 0x0, 0x11
};

/* .table  2  17  3  3  0*/
static const unsigned char treedata_2[] = {
  0x2, 0x1, 0x0, 0x0, 0x4, 0x1, 0x2, 0x1, 0x0, 0x10,
  0x0, 0x1, 0x2, 0x1, 0x0, 0x11, 0x4, 0x1, 0x2, 0x1, 0x0, 0x20, 0x0, 0x21,
  0x2, 0x1, 0x0, 0x12, 0x2, 0x1, 0x0, 0x2, 0x0, 0x22
};

/* .table  3  17  3  3  0*/
static const unsigned char treedata_3[] = {
  0x4, 0x1, 0x2, 0x1, 0x0, 0x0, 0x0, 0x1, 0x2, 0x1,
  0x0, 0x11, 0x2, 0x1, 0x0, 0x10, 0x4, 0x1, 0x2, 0x1, 0x0, 0x20, 0x0, 0x21,
  0x2, 0x1, 0x0, 0x12, 0x2, 0x1, 0x0, 0x2, 0x0, 0x22
};

/* .table  4   0  0  0  0*/
/* table 4 has no tree data*/

/* .table  5  31  4  4  0*/
static const unsigned char treedata_5[] = {
  0x2, 0x1, 0x0, 0x0, 0x4, 0x1, 0x2, 0x1, 0x0, 0x10,
  0x0, 0x1, 0x2, 0x1, 0x0, 0x11, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x20,
  0x0, 0x2, 0x2, 0x1, 0x0, 0x21, 0x0, 0x12, 0x8, 0x1,
  0x4, 0x1, 0x2, 0x1, 0x0, 0x22, 0x0, 0x30, 0x2, 0x1, 0x0, 0x3, 0x0, 0x13,
  0x2, 0x1, 0x0, 0x31, 0x2, 0x1, 0x0, 0x32, 0x2, 0x1, 0x0, 0x23, 0x0, 0x33
};

/* .table  6  31  4  4  0*/
static const unsigned char treedata_6[] = {
  0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x0, 0x0, 0x10,
  0x0, 0x11, 0x6, 0x1, 0x2, 0x1, 0x0, 0x1, 0x2, 0x1, 0x0, 0x20, 0x0, 0x21,
  0x6, 0x1, 0x2, 0x1, 0x0, 0x12, 0x2, 0x1, 0x0, 0x2,
  0x0, 0x22, 0x4, 0x1, 0x2, 0x1, 0x0, 0x31, 0x0, 0x13, 0x4, 0x1, 0x2, 0x1,
  0x0, 0x30, 0x0, 0x32, 0x2, 0x1, 0x0, 0x23, 0x2, 0x1, 0x0, 0x3, 0x0, 0x33
};

/* .table  7  71  6  6  0*/
static const unsigned char treedata_7[] = {
  0x2, 0x1, 0x0, 0x0, 0x4, 0x1, 0x2, 0x1, 0x0, 0x10, 0x0,
  0x1, 0x8, 0x1, 0x2, 0x1, 0x0, 0x11, 0x4, 0x1, 0x2, 0x1, 0x0, 0x20,
  0x0, 0x2, 0x0, 0x21, 0x12, 0x1, 0x6, 0x1, 0x2, 0x1, 0x0,
  0x12, 0x2, 0x1, 0x0, 0x22, 0x0, 0x30, 0x4, 0x1, 0x2, 0x1, 0x0, 0x31,
  0x0, 0x13, 0x4, 0x1, 0x2, 0x1, 0x0, 0x3, 0x0, 0x32, 0x2,
  0x1, 0x0, 0x23, 0x0, 0x4, 0xa, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x40,
  0x0, 0x41, 0x2, 0x1, 0x0, 0x14, 0x2, 0x1, 0x0, 0x42, 0x0,
  0x24, 0xc, 0x1, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x33, 0x0, 0x43,
  0x0, 0x50, 0x4, 0x1, 0x2, 0x1, 0x0, 0x34, 0x0, 0x5, 0x0,
  0x51, 0x6, 0x1, 0x2, 0x1, 0x0, 0x15, 0x2, 0x1, 0x0, 0x52, 0x0, 0x25,
  0x4, 0x1, 0x2, 0x1, 0x0, 0x44, 0x0, 0x35,  0x4,
  0x1, 0x2, 0x1, 0x0, 0x53, 0x0, 0x54, 0x2, 0x1, 0x0, 0x45, 0x0, 0x55
};

/* .table  8  71  6  6  0*/
static const unsigned char treedata_8[] = {
  0x6, 0x1, 0x2, 0x1, 0x0, 0x0, 0x2, 0x1, 0x0, 0x10, 0x0,
  0x1, 0x2, 0x1, 0x0, 0x11, 0x4, 0x1, 0x2, 0x1, 0x0, 0x21, 0x0, 0x12,
  0xe, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x20, 0x0, 0x2,
  0x2, 0x1, 0x0, 0x22, 0x4, 0x1, 0x2, 0x1, 0x0, 0x30, 0x0, 0x3, 0x2, 0x1,
  0x0, 0x31, 0x0, 0x13, 0xe, 0x1, 0x8, 0x1, 0x4, 0x1, 0x2,
  0x1, 0x0, 0x32, 0x0, 0x23, 0x2, 0x1, 0x0, 0x40, 0x0, 0x4, 0x2, 0x1, 
  0x0, 0x41, 0x2, 0x1, 0x0, 0x14, 0x0, 0x42, 0xc, 0x1,
  0x6, 0x1, 0x2, 0x1, 0x0, 0x24, 0x2, 0x1, 0x0, 0x33, 0x0, 0x50, 0x4, 0x1,
  0x2, 0x1, 0x0, 0x43, 0x0, 0x34, 0x0, 0x51, 0x6, 0x1,
  0x2, 0x1, 0x0, 0x15, 0x2, 0x1, 0x0, 0x5, 0x0, 0x52, 0x6, 0x1, 0x2, 0x1,
  0x0, 0x25, 0x2, 0x1, 0x0, 0x44, 0x0, 0x35, 0x2,
  0x1, 0x0, 0x53, 0x2, 0x1, 0x0, 0x45, 0x2, 0x1, 0x0, 0x54, 0x0, 0x55
};

/* .table  9  71  6  6  0*/
static const unsigned char treedata_9[] = {
  0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x0, 0x0, 0x10, 0x2,
  0x1, 0x0, 0x1, 0x0, 0x11, 0xa, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x20,
  0x0, 0x21, 0x2, 0x1, 0x0, 0x12, 0x2, 0x1, 0x0, 0x2,
  0x0, 0x22, 0xc, 0x1, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x30, 0x0, 0x3,
  0x0, 0x31, 0x2, 0x1, 0x0, 0x13, 0x2, 0x1, 0x0, 0x32,
  0x0, 0x23, 0xc, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x41, 0x0, 0x14, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0x40, 0x0, 0x33, 0x2, 0x1, 0x0, 0x42,
  0x0, 0x24, 0xa, 0x1, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x4, 0x0, 0x50,
  0x0, 0x43, 0x2, 0x1, 0x0, 0x34, 0x0, 0x51, 0x8, 0x1,
  0x4, 0x1, 0x2, 0x1, 0x0, 0x15, 0x0, 0x52, 0x2, 0x1, 0x0, 0x25, 0x0, 0x44,
  0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x5, 0x0,
  0x54, 0x0, 0x53, 0x2, 0x1, 0x0, 0x35, 0x2, 0x1, 0x0, 0x45, 0x0, 0x55
};

/*.table 10 127  8  8  0*/
static const unsigned char treedata_10[] = {
  0x2, 0x1, 0x0, 0x0, 0x4, 0x1, 0x2, 0x1, 0x0, 0x10,
  0x0, 0x1, 0xa, 0x1, 0x2, 0x1, 0x0, 0x11, 0x4, 0x1, 0x2, 0x1, 0x0, 0x20,
  0x0, 0x2, 0x2, 0x1, 0x0, 0x21, 0x0, 0x12, 0x1c, 0x1,
  0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x22, 0x0, 0x30, 0x2, 0x1, 0x0, 0x31,
  0x0, 0x13, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x3, 
  0x0, 0x32, 0x2, 0x1, 0x0, 0x23, 0x0, 0x40, 0x4, 0x1, 0x2, 0x1, 0x0, 0x41, 
  0x0, 0x14, 0x4, 0x1, 0x2, 0x1, 0x0, 0x4, 0x0, 0x33,
  0x2, 0x1, 0x0, 0x42, 0x0, 0x24, 0x1c, 0x1, 0xa, 0x1, 0x6, 0x1, 0x4, 0x1,
  0x2, 0x1, 0x0, 0x50, 0x0, 0x5, 0x0, 0x60, 0x2, 0x1,
  0x0, 0x61, 0x0, 0x16, 0xc, 0x1, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x43,
  0x0, 0x34, 0x0, 0x51, 0x2, 0x1, 0x0, 0x15, 0x2, 0x1,
  0x0, 0x52, 0x0, 0x25, 0x4, 0x1, 0x2, 0x1, 0x0, 0x26, 0x0, 0x36, 0x0, 0x71,
  0x14, 0x1, 0x8, 0x1, 0x2, 0x1, 0x0, 0x17, 0x4, 0x1,
  0x2, 0x1, 0x0, 0x44, 0x0, 0x53, 0x0, 0x6, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1,
  0x0, 0x35, 0x0, 0x45, 0x0, 0x62, 0x2, 0x1, 0x0, 0x70,
  0x2, 0x1, 0x0, 0x7, 0x0, 0x64, 0xe, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x72,
  0x0, 0x27, 0x6, 0x1, 0x2, 0x1, 0x0, 0x63, 0x2, 0x1,
  0x0, 0x54, 0x0, 0x55, 0x2, 0x1, 0x0, 0x46, 0x0, 0x73, 0x8, 0x1, 0x4, 0x1,
  0x2, 0x1, 0x0, 0x37, 0x0, 0x65, 0x2, 0x1, 0x0, 0x56, 
  0x0, 0x74, 0x6, 0x1, 0x2, 0x1, 0x0, 0x47, 0x2, 0x1, 0x0, 0x66, 0x0, 0x75, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x57, 0x0, 0x76, 0x2, 0x1, 0x0, 0x67, 0x0, 0x77
};

/* .table 11 127  8  8  0*/
static const unsigned char treedata_11[] = {
  0x6, 0x1, 0x2, 0x1, 0x0, 0x0, 0x2, 0x1, 0x0, 0x10,
  0x0, 0x1, 0x8, 0x1, 0x2, 0x1, 0x0, 0x11, 0x4, 0x1, 0x2, 0x1, 0x0, 0x20,
  0x0, 0x2, 0x0, 0x12, 0x18, 0x1, 0x8, 0x1, 0x2, 0x1,
  0x0, 0x21, 0x2, 0x1, 0x0, 0x22, 0x2, 0x1, 0x0, 0x30, 0x0, 0x3, 0x4, 0x1,
  0x2, 0x1, 0x0, 0x31, 0x0, 0x13, 0x4, 0x1, 0x2, 0x1, 
  0x0, 0x32, 0x0, 0x23, 0x4, 0x1, 0x2, 0x1, 0x0, 0x40, 0x0, 0x4, 0x2, 0x1, 
  0x0, 0x41, 0x0, 0x14, 0x1e, 0x1, 0x10, 0x1, 0xa, 0x1,
  0x4, 0x1, 0x2, 0x1, 0x0, 0x42, 0x0, 0x24, 0x4, 0x1, 0x2, 0x1, 0x0, 0x33,
  0x0, 0x43, 0x0, 0x50, 0x4, 0x1, 0x2, 0x1, 0x0, 0x34,
  0x0, 0x51, 0x0, 0x61, 0x6, 0x1, 0x2, 0x1, 0x0, 0x16, 0x2, 0x1, 0x0, 0x6,
  0x0, 0x26, 0x2, 0x1, 0x0, 0x62, 0x2, 0x1, 0x0, 0x15,
  0x2, 0x1, 0x0, 0x5, 0x0, 0x52, 0x10, 0x1, 0xa, 0x1, 0x6, 0x1, 0x4, 0x1,
  0x2, 0x1, 0x0, 0x25, 0x0, 0x44, 0x0, 0x60, 0x2, 0x1,
  0x0, 0x63, 0x0, 0x36, 0x4, 0x1, 0x2, 0x1, 0x0, 0x70, 0x0, 0x17, 0x0, 0x71,
  0x10, 0x1, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x7,
  0x0, 0x64, 0x0, 0x72, 0x2, 0x1, 0x0, 0x27, 0x4, 0x1, 0x2, 0x1, 0x0, 0x53,
  0x0, 0x35, 0x2, 0x1, 0x0, 0x54, 0x0, 0x45, 0xa, 0x1,
  0x4, 0x1, 0x2, 0x1, 0x0, 0x46, 0x0, 0x73, 0x2, 0x1, 0x0, 0x37, 0x2, 0x1,
  0x0, 0x65, 0x0, 0x56, 0xa, 0x1, 0x6, 0x1, 0x4, 0x1,
  0x2, 0x1, 0x0, 0x55, 0x0, 0x57, 0x0, 0x74, 0x2, 0x1, 0x0, 0x47, 0x0, 0x66,
  0x4, 0x1, 0x2, 0x1, 0x0, 0x75, 0x0, 0x76, 0x2, 0x1, 0x0, 0x67, 0x0, 0x77
};

/* .table 12 127  8  8  0*/
static const unsigned char treedata_12[] = {
  0xc, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x10, 0x0, 0x1,
  0x2, 0x1, 0x0, 0x11, 0x2, 0x1, 0x0, 0x0, 0x2, 0x1, 0x0, 0x20, 0x0, 0x2,
  0x10, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x21, 0x0, 0x12,
  0x4, 0x1, 0x2, 0x1, 0x0, 0x22, 0x0, 0x31, 0x2, 0x1, 0x0, 0x13, 0x2, 0x1,
  0x0, 0x30, 0x2, 0x1, 0x0, 0x3, 0x0, 0x40, 0x1a, 0x1,
  0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x32, 0x0, 0x23, 0x2, 0x1, 0x0, 0x41,
  0x0, 0x33, 0xa, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x14,
  0x0, 0x42, 0x2, 0x1, 0x0, 0x24, 0x2, 0x1, 0x0, 0x4, 0x0, 0x50, 0x4, 0x1,
  0x2, 0x1, 0x0, 0x43, 0x0, 0x34, 0x2, 0x1, 0x0, 0x51,
  0x0, 0x15, 0x1c, 0x1, 0xe, 0x1, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x52,
  0x0, 0x25, 0x2, 0x1, 0x0, 0x53, 0x0, 0x35, 0x4, 0x1,
  0x2, 0x1, 0x0, 0x60, 0x0, 0x16, 0x0, 0x61, 0x4, 0x1, 0x2, 0x1, 0x0, 0x62,
  0x0, 0x26, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x5,
  0x0, 0x6, 0x0, 0x44, 0x2, 0x1, 0x0, 0x54, 0x0, 0x45, 0x12, 0x1, 0xa, 0x1,
  0x4, 0x1, 0x2, 0x1, 0x0, 0x63, 0x0, 0x36, 0x4, 0x1,
  0x2, 0x1, 0x0, 0x70, 0x0, 0x7, 0x0, 0x71, 0x4, 0x1, 0x2, 0x1, 0x0, 0x17,
  0x0, 0x64, 0x2, 0x1, 0x0, 0x46, 0x0, 0x72, 0xa, 0x1,
  0x6, 0x1, 0x2, 0x1, 0x0, 0x27, 0x2, 0x1, 0x0, 0x55, 0x0, 0x73, 0x2, 0x1,
  0x0, 0x37, 0x0, 0x56, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1,
  0x0, 0x65, 0x0, 0x74, 0x2, 0x1, 0x0, 0x47, 0x0, 0x66, 0x4, 0x1, 0x2, 0x1,
  0x0, 0x75, 0x0, 0x57, 0x2, 0x1, 0x0, 0x76, 0x2, 0x1, 0x0, 0x67, 0x0, 0x77
};

/* .table 13 511 16 16  0*/
static const unsigned char treedata_13[] = {
  0x2, 0x1, 0x0, 0x0, 0x6, 0x1, 0x2, 0x1, 0x0, 0x10, 0x2, 0x1, 0x0, 0x1, 0x0,
  0x11, 0x1c, 0x1, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x20, 0x0, 0x2, 0x2, 0x1, 
  0x0, 0x21, 0x0, 0x12, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x22, 0x0, 0x30, 
  0x2, 0x1, 0x0, 0x3, 0x0, 0x31, 0x6, 0x1, 0x2, 0x1, 0x0, 0x13, 0x2, 0x1, 
  0x0, 0x32, 0x0, 0x23, 0x4, 0x1, 0x2, 0x1, 0x0, 0x40, 0x0, 0x4, 0x0, 0x41, 
  0x46, 0x1, 0x1c, 0x1, 0xe, 0x1, 0x6, 0x1, 0x2, 0x1, 0x0, 0x14, 0x2, 0x1, 
  0x0, 0x33, 0x0, 0x42, 0x4, 0x1, 0x2, 0x1, 0x0, 0x24, 0x0, 0x50, 0x2, 0x1, 
  0x0, 0x43, 0x0, 0x34, 0x4, 0x1, 0x2, 0x1, 0x0, 0x51, 0x0, 0x15, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0x5, 0x0, 0x52, 0x2, 0x1, 0x0, 0x25, 0x2, 0x1, 0x0, 0x44, 
  0x0, 0x53, 0xe, 0x1, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x60, 0x0, 0x6, 
  0x2, 0x1, 0x0, 0x61, 0x0, 0x16, 0x4, 0x1, 0x2, 0x1, 0x0, 0x80, 0x0, 0x8, 
  0x0, 0x81, 0x10, 0x1, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x35, 0x0, 0x62, 
  0x2, 0x1, 0x0, 0x26, 0x0, 0x54, 0x4, 0x1, 0x2, 0x1, 0x0, 0x45, 0x0, 0x63, 
  0x2, 0x1, 0x0, 0x36, 0x0, 0x70, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x7, 
  0x0, 0x55, 0x0, 0x71, 0x2, 0x1, 0x0, 0x17, 0x2, 0x1, 0x0, 0x27, 0x0, 0x37, 
  0x48, 0x1, 0x18, 0x1, 0xc, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x18, 0x0, 0x82, 
  0x2, 0x1, 0x0, 0x28, 0x4, 0x1, 0x2, 0x1, 0x0, 0x64, 0x0, 0x46, 0x0, 0x72, 
  0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x84, 0x0, 0x48, 0x2, 0x1, 0x0, 0x90, 
  0x0, 0x9, 0x2, 0x1, 0x0, 0x91, 0x0, 0x19, 0x18, 0x1, 0xe, 0x1, 0x8, 0x1, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x73, 0x0, 0x65, 0x2, 0x1, 0x0, 0x56, 0x0, 0x74, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x47, 0x0, 0x66, 0x0, 0x83, 0x6, 0x1, 0x2, 0x1, 
  0x0, 0x38, 0x2, 0x1, 0x0, 0x75, 0x0, 0x57, 0x2, 0x1, 0x0, 0x92, 0x0, 0x29, 
  0xe, 0x1, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x67, 0x0, 0x85, 0x2, 0x1, 
  0x0, 0x58, 0x0, 0x39, 0x2, 0x1, 0x0, 0x93, 0x2, 0x1, 0x0, 0x49, 0x0, 0x86, 
  0x6, 0x1, 0x2, 0x1, 0x0, 0xa0, 0x2, 0x1, 0x0, 0x68, 0x0, 0xa, 0x2, 0x1, 
  0x0, 0xa1, 0x0, 0x1a, 0x44, 0x1, 0x18, 0x1, 0xc, 0x1, 0x4, 0x1, 0x2, 0x1, 
  0x0, 0xa2, 0x0, 0x2a, 0x4, 0x1, 0x2, 0x1, 0x0, 0x95, 0x0, 0x59, 0x2, 0x1, 
  0x0, 0xa3, 0x0, 0x3a, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x4a, 0x0, 0x96, 
  0x2, 0x1, 0x0, 0xb0, 0x0, 0xb, 0x2, 0x1, 0x0, 0xb1, 0x0, 0x1b, 0x14, 0x1, 
  0x8, 0x1, 0x2, 0x1, 0x0, 0xb2, 0x4, 0x1, 0x2, 0x1, 0x0, 0x76, 0x0, 0x77, 
  0x0, 0x94, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x87, 0x0, 0x78, 0x0, 0xa4, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x69, 0x0, 0xa5, 0x0, 0x2b, 0xc, 0x1, 0x6, 0x1, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x5a, 0x0, 0x88, 0x0, 0xb3, 0x2, 0x1, 0x0, 0x3b, 
  0x2, 0x1, 0x0, 0x79, 0x0, 0xa6, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x6a, 
  0x0, 0xb4, 0x0, 0xc0, 0x4, 0x1, 0x2, 0x1, 0x0, 0xc, 0x0, 0x98, 0x0, 0xc1, 
  0x3c, 0x1, 0x16, 0x1, 0xa, 0x1, 0x6, 0x1, 0x2, 0x1, 0x0, 0x1c, 0x2, 0x1, 
  0x0, 0x89, 0x0, 0xb5, 0x2, 0x1, 0x0, 0x5b, 0x0, 0xc2, 0x4, 0x1, 0x2, 0x1, 
  0x0, 0x2c, 0x0, 0x3c, 0x4, 0x1, 0x2, 0x1, 0x0, 0xb6, 0x0, 0x6b, 0x2, 0x1, 
  0x0, 0xc4, 0x0, 0x4c, 0x10, 0x1, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0xa8, 
  0x0, 0x8a, 0x2, 0x1, 0x0, 0xd0, 0x0, 0xd, 0x2, 0x1, 0x0, 0xd1, 0x2, 0x1, 
  0x0, 0x4b, 0x2, 0x1, 0x0, 0x97, 0x0, 0xa7, 0xc, 0x1, 0x6, 0x1, 0x2, 0x1, 
  0x0, 0xc3, 0x2, 0x1, 0x0, 0x7a, 0x0, 0x99, 0x4, 0x1, 0x2, 0x1, 0x0, 0xc5, 
  0x0, 0x5c, 0x0, 0xb7, 0x4, 0x1, 0x2, 0x1, 0x0, 0x1d, 0x0, 0xd2, 0x2, 0x1, 
  0x0, 0x2d, 0x2, 0x1, 0x0, 0x7b, 0x0, 0xd3, 0x34, 0x1, 0x1c, 0x1, 0xc, 0x1, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x3d, 0x0, 0xc6, 0x4, 0x1, 0x2, 0x1, 0x0, 0x6c, 
  0x0, 0xa9, 0x2, 0x1, 0x0, 0x9a, 0x0, 0xd4, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 
  0x0, 0xb8, 0x0, 0x8b, 0x2, 0x1, 0x0, 0x4d, 0x0, 0xc7, 0x4, 0x1, 0x2, 0x1, 
  0x0, 0x7c, 0x0, 0xd5, 0x2, 0x1, 0x0, 0x5d, 0x0, 0xe0, 0xa, 0x1, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0xe1, 0x0, 0x1e, 0x4, 0x1, 0x2, 0x1, 0x0, 0xe, 0x0, 0x2e, 
  0x0, 0xe2, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0xe3, 0x0, 0x6d, 0x2, 0x1, 
  0x0, 0x8c, 0x0, 0xe4, 0x4, 0x1, 0x2, 0x1, 0x0, 0xe5, 0x0, 0xba, 0x0, 0xf0, 
  0x26, 0x1, 0x10, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0xf1, 0x0, 0x1f, 0x6, 0x1, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0xaa, 0x0, 0x9b, 0x0, 0xb9, 0x2, 0x1, 0x0, 0x3e, 
  0x2, 0x1, 0x0, 0xd6, 0x0, 0xc8, 0xc, 0x1, 0x6, 0x1, 0x2, 0x1, 0x0, 0x4e, 
  0x2, 0x1, 0x0, 0xd7, 0x0, 0x7d, 0x2, 0x1, 0x0, 0xab, 0x2, 0x1, 0x0, 0x5e, 
  0x0, 0xc9, 0x6, 0x1, 0x2, 0x1, 0x0, 0xf, 0x2, 0x1, 0x0, 0x9c, 0x0, 0x6e, 
  0x2, 0x1, 0x0, 0xf2, 0x0, 0x2f, 0x20, 0x1, 0x10, 0x1, 0x6, 0x1, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0xd8, 0x0, 0x8d, 0x0, 0x3f, 0x6, 0x1, 0x2, 0x1, 0x0, 0xf3, 
  0x2, 0x1, 0x0, 0xe6, 0x0, 0xca, 0x2, 0x1, 0x0, 0xf4, 0x0, 0x4f, 0x8, 0x1, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0xbb, 0x0, 0xac, 0x2, 0x1, 0x0, 0xe7, 0x0, 0xf5, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0xd9, 0x0, 0x9d, 0x2, 0x1, 0x0, 0x5f, 0x0, 0xe8, 
  0x1e, 0x1, 0xc, 0x1, 0x6, 0x1, 0x2, 0x1, 0x0, 0x6f, 0x2, 0x1, 0x0, 0xf6, 
  0x0, 0xcb, 0x4, 0x1, 0x2, 0x1, 0x0, 0xbc, 0x0, 0xad, 0x0, 0xda, 0x8, 0x1, 
  0x2, 0x1, 0x0, 0xf7, 0x4, 0x1, 0x2, 0x1, 0x0, 0x7e, 0x0, 0x7f, 0x0, 0x8e, 
  0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x9e, 0x0, 0xae, 0x0, 0xcc, 0x2, 0x1, 
  0x0, 0xf8, 0x0, 0x8f, 0x12, 0x1, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0xdb, 
  0x0, 0xbd, 0x2, 0x1, 0x0, 0xea, 0x0, 0xf9, 0x4, 0x1, 0x2, 0x1, 0x0, 0x9f, 
  0x0, 0xeb, 0x2, 0x1, 0x0, 0xbe, 0x2, 0x1, 0x0, 0xcd, 0x0, 0xfa, 0xe, 0x1, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0xdd, 0x0, 0xec, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 
  0x0, 0xe9, 0x0, 0xaf, 0x0, 0xdc, 0x2, 0x1, 0x0, 0xce, 0x0, 0xfb, 0x8, 0x1, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0xbf, 0x0, 0xde, 0x2, 0x1, 0x0, 0xcf, 0x0, 0xee, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0xdf, 0x0, 0xef, 0x2, 0x1, 0x0, 0xff, 0x2, 0x1, 
  0x0, 0xed, 0x2, 0x1, 0x0, 0xfd, 0x2, 0x1, 0x0, 0xfc, 0x0, 0xfe
};

/* .table 14   0  0  0  0*/
/* table 14 has no tree data*/

/* .table 15 511 16 16  0*/
static const unsigned char treedata_15[] = {
  0x10, 0x1, 0x6, 0x1, 0x2, 0x1, 0x0, 0x0, 0x2, 0x1, 0x0, 0x10, 0x0, 0x1, 0x2,
  0x1, 0x0, 0x11, 0x4, 0x1, 0x2, 0x1, 0x0, 0x20, 0x0, 0x2, 0x2, 0x1, 0x0, 0x21, 
  0x0, 0x12, 0x32, 0x1, 0x10, 0x1, 0x6, 0x1, 0x2, 0x1, 0x0, 0x22, 0x2, 0x1, 
  0x0, 0x30, 0x0, 0x31, 0x6, 0x1, 0x2, 0x1, 0x0, 0x13, 0x2, 0x1, 0x0, 0x3, 
  0x0, 0x40, 0x2, 0x1, 0x0, 0x32, 0x0, 0x23, 0xe, 0x1, 0x6, 0x1, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0x4, 0x0, 0x14, 0x0, 0x41, 0x4, 0x1, 0x2, 0x1, 0x0, 0x33, 
  0x0, 0x42, 0x2, 0x1, 0x0, 0x24, 0x0, 0x43, 0xa, 0x1, 0x6, 0x1, 0x2, 0x1, 
  0x0, 0x34, 0x2, 0x1, 0x0, 0x50, 0x0, 0x5, 0x2, 0x1, 0x0, 0x51, 0x0, 0x15, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x52, 0x0, 0x25, 0x4, 0x1, 0x2, 0x1, 0x0, 0x44, 
  0x0, 0x53, 0x0, 0x61, 0x5a, 0x1, 0x24, 0x1, 0x12, 0x1, 0xa, 0x1, 0x6, 0x1, 
  0x2, 0x1, 0x0, 0x35, 0x2, 0x1, 0x0, 0x60, 0x0, 0x6, 0x2, 0x1, 0x0, 0x16, 
  0x0, 0x62, 0x4, 0x1, 0x2, 0x1, 0x0, 0x26, 0x0, 0x54, 0x2, 0x1, 0x0, 0x45, 
  0x0, 0x63, 0xa, 0x1, 0x6, 0x1, 0x2, 0x1, 0x0, 0x36, 0x2, 0x1, 0x0, 0x70, 
  0x0, 0x7, 0x2, 0x1, 0x0, 0x71, 0x0, 0x55, 0x4, 0x1, 0x2, 0x1, 0x0, 0x17, 
  0x0, 0x64, 0x2, 0x1, 0x0, 0x72, 0x0, 0x27, 0x18, 0x1, 0x10, 0x1, 0x8, 0x1, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x46, 0x0, 0x73, 0x2, 0x1, 0x0, 0x37, 0x0, 0x65, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x56, 0x0, 0x80, 0x2, 0x1, 0x0, 0x8, 0x0, 0x74, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x81, 0x0, 0x18, 0x2, 0x1, 0x0, 0x82, 0x0, 0x28, 
  0x10, 0x1, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x47, 0x0, 0x66, 0x2, 0x1, 
  0x0, 0x83, 0x0, 0x38, 0x4, 0x1, 0x2, 0x1, 0x0, 0x75, 0x0, 0x57, 0x2, 0x1, 
  0x0, 0x84, 0x0, 0x48, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x90, 0x0, 0x19, 
  0x0, 0x91, 0x4, 0x1, 0x2, 0x1, 0x0, 0x92, 0x0, 0x76, 0x2, 0x1, 0x0, 0x67, 
  0x0, 0x29, 0x5c, 0x1, 0x24, 0x1, 0x12, 0x1, 0xa, 0x1, 0x4, 0x1, 0x2, 0x1, 
  0x0, 0x85, 0x0, 0x58, 0x4, 0x1, 0x2, 0x1, 0x0, 0x9, 0x0, 0x77, 0x0, 0x93, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x39, 0x0, 0x94, 0x2, 0x1, 0x0, 0x49, 0x0, 0x86, 
  0xa, 0x1, 0x6, 0x1, 0x2, 0x1, 0x0, 0x68, 0x2, 0x1, 0x0, 0xa0, 0x0, 0xa, 
  0x2, 0x1, 0x0, 0xa1, 0x0, 0x1a, 0x4, 0x1, 0x2, 0x1, 0x0, 0xa2, 0x0, 0x2a, 
  0x2, 0x1, 0x0, 0x95, 0x0, 0x59, 0x1a, 0x1, 0xe, 0x1, 0x6, 0x1, 0x2, 0x1, 
  0x0, 0xa3, 0x2, 0x1, 0x0, 0x3a, 0x0, 0x87, 0x4, 0x1, 0x2, 0x1, 0x0, 0x78, 
  0x0, 0xa4, 0x2, 0x1, 0x0, 0x4a, 0x0, 0x96, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 
  0x0, 0x69, 0x0, 0xb0, 0x0, 0xb1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x1b, 0x0, 0xa5, 
  0x0, 0xb2, 0xe, 0x1, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x5a, 0x0, 0x2b, 
  0x2, 0x1, 0x0, 0x88, 0x0, 0x97, 0x2, 0x1, 0x0, 0xb3, 0x2, 0x1, 0x0, 0x79, 
  0x0, 0x3b, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x6a, 0x0, 0xb4, 0x2, 0x1, 
  0x0, 0x4b, 0x0, 0xc1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x98, 0x0, 0x89, 0x2, 0x1, 
  0x0, 0x1c, 0x0, 0xb5, 0x50, 0x1, 0x22, 0x1, 0x10, 0x1, 0x6, 0x1, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0x5b, 0x0, 0x2c, 0x0, 0xc2, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 
  0x0, 0xb, 0x0, 0xc0, 0x0, 0xa6, 0x2, 0x1, 0x0, 0xa7, 0x0, 0x7a, 0xa, 0x1, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0xc3, 0x0, 0x3c, 0x4, 0x1, 0x2, 0x1, 0x0, 0xc, 
  0x0, 0x99, 0x0, 0xb6, 0x4, 0x1, 0x2, 0x1, 0x0, 0x6b, 0x0, 0xc4, 0x2, 0x1, 
  0x0, 0x4c, 0x0, 0xa8, 0x14, 0x1, 0xa, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x8a, 
  0x0, 0xc5, 0x4, 0x1, 0x2, 0x1, 0x0, 0xd0, 0x0, 0x5c, 0x0, 0xd1, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0xb7, 0x0, 0x7b, 0x2, 0x1, 0x0, 0x1d, 0x2, 0x1, 0x0, 0xd, 
  0x0, 0x2d, 0xc, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0xd2, 0x0, 0xd3, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0x3d, 0x0, 0xc6, 0x2, 0x1, 0x0, 0x6c, 0x0, 0xa9, 0x6, 0x1, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x9a, 0x0, 0xb8, 0x0, 0xd4, 0x4, 0x1, 0x2, 0x1, 
  0x0, 0x8b, 0x0, 0x4d, 0x2, 0x1, 0x0, 0xc7, 0x0, 0x7c, 0x44, 0x1, 0x22, 0x1, 
  0x12, 0x1, 0xa, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0xd5, 0x0, 0x5d, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0xe0, 0x0, 0xe, 0x0, 0xe1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x1e, 
  0x0, 0xe2, 0x2, 0x1, 0x0, 0xaa, 0x0, 0x2e, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 
  0x0, 0xb9, 0x0, 0x9b, 0x2, 0x1, 0x0, 0xe3, 0x0, 0xd6, 0x4, 0x1, 0x2, 0x1, 
  0x0, 0x6d, 0x0, 0x3e, 0x2, 0x1, 0x0, 0xc8, 0x0, 0x8c, 0x10, 0x1, 0x8, 0x1, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0xe4, 0x0, 0x4e, 0x2, 0x1, 0x0, 0xd7, 0x0, 0x7d, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0xe5, 0x0, 0xba, 0x2, 0x1, 0x0, 0xab, 0x0, 0x5e, 
  0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0xc9, 0x0, 0x9c, 0x2, 0x1, 0x0, 0xf1, 
  0x0, 0x1f, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0xf0, 0x0, 0x6e, 0x0, 0xf2, 
  0x2, 0x1, 0x0, 0x2f, 0x0, 0xe6, 0x26, 0x1, 0x12, 0x1, 0x8, 0x1, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0xd8, 0x0, 0xf3, 0x2, 0x1, 0x0, 0x3f, 0x0, 0xf4, 0x6, 0x1, 
  0x2, 0x1, 0x0, 0x4f, 0x2, 0x1, 0x0, 0x8d, 0x0, 0xd9, 0x2, 0x1, 0x0, 0xbb, 
  0x0, 0xca, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0xac, 0x0, 0xe7, 0x2, 0x1, 
  0x0, 0x7e, 0x0, 0xf5, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x9d, 0x0, 0x5f, 
  0x2, 0x1, 0x0, 0xe8, 0x0, 0x8e, 0x2, 0x1, 0x0, 0xf6, 0x0, 0xcb, 0x22, 0x1, 
  0x12, 0x1, 0xa, 0x1, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0xf, 0x0, 0xae, 
  0x0, 0x6f, 0x2, 0x1, 0x0, 0xbc, 0x0, 0xda, 0x4, 0x1, 0x2, 0x1, 0x0, 0xad, 
  0x0, 0xf7, 0x2, 0x1, 0x0, 0x7f, 0x0, 0xe9, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 
  0x0, 0x9e, 0x0, 0xcc, 0x2, 0x1, 0x0, 0xf8, 0x0, 0x8f, 0x4, 0x1, 0x2, 0x1, 
  0x0, 0xdb, 0x0, 0xbd, 0x2, 0x1, 0x0, 0xea, 0x0, 0xf9, 0x10, 0x1, 0x8, 0x1, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x9f, 0x0, 0xdc, 0x2, 0x1, 0x0, 0xcd, 0x0, 0xeb, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0xbe, 0x0, 0xfa, 0x2, 0x1, 0x0, 0xaf, 0x0, 0xdd, 
  0xe, 0x1, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0xec, 0x0, 0xce, 0x0, 0xfb, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0xbf, 0x0, 0xed, 0x2, 0x1, 0x0, 0xde, 0x0, 0xfc, 
  0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0xcf, 0x0, 0xfd, 0x0, 0xee, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0xdf, 0x0, 0xfe, 0x2, 0x1, 0x0, 0xef, 0x0, 0xff
};

/* .table 16 511 16 16  1*/
static const unsigned char treedata_16[] = {
  0x2, 0x1, 0x0, 0x0, 0x6, 0x1, 0x2, 0x1, 0x0, 0x10, 0x2, 0x1, 0x0, 0x1, 0x0,
  0x11, 0x2a, 0x1, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x20, 0x0, 0x2, 0x2, 0x1, 
  0x0, 0x21, 0x0, 0x12, 0xa, 0x1, 0x6, 0x1, 0x2, 0x1, 0x0, 0x22, 0x2, 0x1, 
  0x0, 0x30, 0x0, 0x3, 0x2, 0x1, 0x0, 0x31, 0x0, 0x13, 0xa, 0x1, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0x32, 0x0, 0x23, 0x4, 0x1, 0x2, 0x1, 0x0, 0x40, 0x0, 0x4, 
  0x0, 0x41, 0x6, 0x1, 0x2, 0x1, 0x0, 0x14, 0x2, 0x1, 0x0, 0x33, 0x0, 0x42, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x24, 0x0, 0x50, 0x2, 0x1, 0x0, 0x43, 0x0, 0x34, 
  0x8a, 0x1, 0x28, 0x1, 0x10, 0x1, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x5, 
  0x0, 0x15, 0x0, 0x51, 0x4, 0x1, 0x2, 0x1, 0x0, 0x52, 0x0, 0x25, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0x44, 0x0, 0x35, 0x0, 0x53, 0xa, 0x1, 0x6, 0x1, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0x60, 0x0, 0x6, 0x0, 0x61, 0x2, 0x1, 0x0, 0x16, 0x0, 0x62, 
  0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x26, 0x0, 0x54, 0x2, 0x1, 0x0, 0x45, 
  0x0, 0x63, 0x4, 0x1, 0x2, 0x1, 0x0, 0x36, 0x0, 0x70, 0x0, 0x71, 0x28, 0x1, 
  0x12, 0x1, 0x8, 0x1, 0x2, 0x1, 0x0, 0x17, 0x2, 0x1, 0x0, 0x7, 0x2, 0x1, 
  0x0, 0x55, 0x0, 0x64, 0x4, 0x1, 0x2, 0x1, 0x0, 0x72, 0x0, 0x27, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0x46, 0x0, 0x65, 0x0, 0x73, 0xa, 0x1, 0x6, 0x1, 0x2, 0x1, 
  0x0, 0x37, 0x2, 0x1, 0x0, 0x56, 0x0, 0x8, 0x2, 0x1, 0x0, 0x80, 0x0, 0x81, 
  0x6, 0x1, 0x2, 0x1, 0x0, 0x18, 0x2, 0x1, 0x0, 0x74, 0x0, 0x47, 0x2, 0x1, 
  0x0, 0x82, 0x2, 0x1, 0x0, 0x28, 0x0, 0x66, 0x18, 0x1, 0xe, 0x1, 0x8, 0x1, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x83, 0x0, 0x38, 0x2, 0x1, 0x0, 0x75, 0x0, 0x84, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x48, 0x0, 0x90, 0x0, 0x91, 0x6, 0x1, 0x2, 0x1, 
  0x0, 0x19, 0x2, 0x1, 0x0, 0x9, 0x0, 0x76, 0x2, 0x1, 0x0, 0x92, 0x0, 0x29, 
  0xe, 0x1, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x85, 0x0, 0x58, 0x2, 0x1, 
  0x0, 0x93, 0x0, 0x39, 0x4, 0x1, 0x2, 0x1, 0x0, 0xa0, 0x0, 0xa, 0x0, 0x1a, 
  0x8, 0x1, 0x2, 0x1, 0x0, 0xa2, 0x2, 0x1, 0x0, 0x67, 0x2, 0x1, 0x0, 0x57, 
  0x0, 0x49, 0x6, 0x1, 0x2, 0x1, 0x0, 0x94, 0x2, 0x1, 0x0, 0x77, 0x0, 0x86, 
  0x2, 0x1, 0x0, 0xa1, 0x2, 0x1, 0x0, 0x68, 0x0, 0x95, 0xdc, 0x1, 0x7e, 0x1, 
  0x32, 0x1, 0x1a, 0x1, 0xc, 0x1, 0x6, 0x1, 0x2, 0x1, 0x0, 0x2a, 0x2, 0x1, 
  0x0, 0x59, 0x0, 0x3a, 0x2, 0x1, 0x0, 0xa3, 0x2, 0x1, 0x0, 0x87, 0x0, 0x78, 
  0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0xa4, 0x0, 0x4a, 0x2, 0x1, 0x0, 0x96, 
  0x0, 0x69, 0x4, 0x1, 0x2, 0x1, 0x0, 0xb0, 0x0, 0xb, 0x0, 0xb1, 0xa, 0x1, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x1b, 0x0, 0xb2, 0x2, 0x1, 0x0, 0x2b, 0x2, 0x1, 
  0x0, 0xa5, 0x0, 0x5a, 0x6, 0x1, 0x2, 0x1, 0x0, 0xb3, 0x2, 0x1, 0x0, 0xa6, 
  0x0, 0x6a, 0x4, 0x1, 0x2, 0x1, 0x0, 0xb4, 0x0, 0x4b, 0x2, 0x1, 0x0, 0xc, 
  0x0, 0xc1, 0x1e, 0x1, 0xe, 0x1, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0xb5, 
  0x0, 0xc2, 0x0, 0x2c, 0x4, 0x1, 0x2, 0x1, 0x0, 0xa7, 0x0, 0xc3, 0x2, 0x1, 
  0x0, 0x6b, 0x0, 0xc4, 0x8, 0x1, 0x2, 0x1, 0x0, 0x1d, 0x4, 0x1, 0x2, 0x1, 
  0x0, 0x88, 0x0, 0x97, 0x0, 0x3b, 0x4, 0x1, 0x2, 0x1, 0x0, 0xd1, 0x0, 0xd2, 
  0x2, 0x1, 0x0, 0x2d, 0x0, 0xd3, 0x12, 0x1, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 
  0x0, 0x1e, 0x0, 0x2e, 0x0, 0xe2, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x79, 
  0x0, 0x98, 0x0, 0xc0, 0x2, 0x1, 0x0, 0x1c, 0x2, 0x1, 0x0, 0x89, 0x0, 0x5b, 
  0xe, 0x1, 0x6, 0x1, 0x2, 0x1, 0x0, 0x3c, 0x2, 0x1, 0x0, 0x7a, 0x0, 0xb6, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x4c, 0x0, 0x99, 0x2, 0x1, 0x0, 0xa8, 0x0, 0x8a, 
  0x6, 0x1, 0x2, 0x1, 0x0, 0xd, 0x2, 0x1, 0x0, 0xc5, 0x0, 0x5c, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0x3d, 0x0, 0xc6, 0x2, 0x1, 0x0, 0x6c, 0x0, 0x9a, 0x58, 0x1, 
  0x56, 0x1, 0x24, 0x1, 0x10, 0x1, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x8b, 
  0x0, 0x4d, 0x2, 0x1, 0x0, 0xc7, 0x0, 0x7c, 0x4, 0x1, 0x2, 0x1, 0x0, 0xd5, 
  0x0, 0x5d, 0x2, 0x1, 0x0, 0xe0, 0x0, 0xe, 0x8, 0x1, 0x2, 0x1, 0x0, 0xe3, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0xd0, 0x0, 0xb7, 0x0, 0x7b, 0x6, 0x1, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0xa9, 0x0, 0xb8, 0x0, 0xd4, 0x2, 0x1, 0x0, 0xe1, 0x2, 0x1, 
  0x0, 0xaa, 0x0, 0xb9, 0x18, 0x1, 0xa, 0x1, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 
  0x0, 0x9b, 0x0, 0xd6, 0x0, 0x6d, 0x2, 0x1, 0x0, 0x3e, 0x0, 0xc8, 0x6, 0x1, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x8c, 0x0, 0xe4, 0x0, 0x4e, 0x4, 0x1, 0x2, 0x1, 
  0x0, 0xd7, 0x0, 0xe5, 0x2, 0x1, 0x0, 0xba, 0x0, 0xab, 0xc, 0x1, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0x9c, 0x0, 0xe6, 0x4, 0x1, 0x2, 0x1, 0x0, 0x6e, 0x0, 0xd8, 
  0x2, 0x1, 0x0, 0x8d, 0x0, 0xbb, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0xe7, 
  0x0, 0x9d, 0x2, 0x1, 0x0, 0xe8, 0x0, 0x8e, 0x4, 0x1, 0x2, 0x1, 0x0, 0xcb, 
  0x0, 0xbc, 0x0, 0x9e, 0x0, 0xf1, 0x2, 0x1, 0x0, 0x1f, 0x2, 0x1, 0x0, 0xf, 
  0x0, 0x2f, 0x42, 0x1, 0x38, 0x1, 0x2, 0x1, 0x0, 0xf2, 0x34, 0x1, 0x32, 0x1, 
  0x14, 0x1, 0x8, 0x1, 0x2, 0x1, 0x0, 0xbd, 0x2, 0x1, 0x0, 0x5e, 0x2, 0x1, 
  0x0, 0x7d, 0x0, 0xc9, 0x6, 0x1, 0x2, 0x1, 0x0, 0xca, 0x2, 0x1, 0x0, 0xac, 
  0x0, 0x7e, 0x4, 0x1, 0x2, 0x1, 0x0, 0xda, 0x0, 0xad, 0x0, 0xcc, 0xa, 0x1, 
  0x6, 0x1, 0x2, 0x1, 0x0, 0xae, 0x2, 0x1, 0x0, 0xdb, 0x0, 0xdc, 0x2, 0x1, 
  0x0, 0xcd, 0x0, 0xbe, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0xeb, 0x0, 0xed, 
  0x0, 0xee, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0xd9, 0x0, 0xea, 0x0, 0xe9, 
  0x2, 0x1, 0x0, 0xde, 0x4, 0x1, 0x2, 0x1, 0x0, 0xdd, 0x0, 0xec, 0x0, 0xce, 
  0x0, 0x3f, 0x0, 0xf0, 0x4, 0x1, 0x2, 0x1, 0x0, 0xf3, 0x0, 0xf4, 0x2, 0x1, 
  0x0, 0x4f, 0x2, 0x1, 0x0, 0xf5, 0x0, 0x5f, 0xa, 0x1, 0x2, 0x1, 0x0, 0xff, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0xf6, 0x0, 0x6f, 0x2, 0x1, 0x0, 0xf7, 0x0, 0x7f, 
  0xc, 0x1, 0x6, 0x1, 0x2, 0x1, 0x0, 0x8f, 0x2, 0x1, 0x0, 0xf8, 0x0, 0xf9, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x9f, 0x0, 0xfa, 0x0, 0xaf, 0x8, 0x1, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0xfb, 0x0, 0xbf, 0x2, 0x1, 0x0, 0xfc, 0x0, 0xcf, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0xfd, 0x0, 0xdf, 0x2, 0x1, 0x0, 0xfe, 0x0, 0xef
};

/* .table 17 511 16 16  2*/
/* .reference 16*/

/* .table 18 511 16 16  3*/
/* .reference 16*/

/* .table 19 511 16 16  4*/
/* .reference 16*/

/* .table 20 511 16 16  6*/
/* .reference 16*/

/* .table 21 511 16 16  8*/
/* .reference 16*/

/* .table 22 511 16 16 10*/
/* .reference 16*/

/* .table 23 511 16 16 13*/
/* .reference 16*/

/* .table 24 512 16 16  4*/
static const unsigned char treedata_24[] = {
  0x3c, 0x1, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x0, 0x0, 0x10, 0x2, 0x1, 0x0,
  0x1, 0x0, 0x11, 0xe, 0x1, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x20, 0x0, 0x2, 
  0x0, 0x21, 0x2, 0x1, 0x0, 0x12, 0x2, 0x1, 0x0, 0x22, 0x2, 0x1, 0x0, 0x30, 
  0x0, 0x3, 0xe, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x31, 0x0, 0x13, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0x32, 0x0, 0x23, 0x4, 0x1, 0x2, 0x1, 0x0, 0x40, 0x0, 0x4, 
  0x0, 0x41, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x14, 0x0, 0x33, 0x2, 0x1, 
  0x0, 0x42, 0x0, 0x24, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x43, 0x0, 0x34, 
  0x0, 0x51, 0x6, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x50, 0x0, 0x5, 0x0, 0x15, 
  0x2, 0x1, 0x0, 0x52, 0x0, 0x25, 0xfa, 0x1, 0x62, 0x1, 0x22, 0x1, 0x12, 0x1, 
  0xa, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x44, 0x0, 0x53, 0x2, 0x1, 0x0, 0x35, 
  0x2, 0x1, 0x0, 0x60, 0x0, 0x6, 0x4, 0x1, 0x2, 0x1, 0x0, 0x61, 0x0, 0x16, 
  0x2, 0x1, 0x0, 0x62, 0x0, 0x26, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x54, 
  0x0, 0x45, 0x2, 0x1, 0x0, 0x63, 0x0, 0x36, 0x4, 0x1, 0x2, 0x1, 0x0, 0x71, 
  0x0, 0x55, 0x2, 0x1, 0x0, 0x64, 0x0, 0x46, 0x20, 0x1, 0xe, 0x1, 0x6, 0x1, 
  0x2, 0x1, 0x0, 0x72, 0x2, 0x1, 0x0, 0x27, 0x0, 0x37, 0x2, 0x1, 0x0, 0x73, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x70, 0x0, 0x7, 0x0, 0x17, 0xa, 0x1, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0x65, 0x0, 0x56, 0x4, 0x1, 0x2, 0x1, 0x0, 0x80, 0x0, 0x8, 
  0x0, 0x81, 0x4, 0x1, 0x2, 0x1, 0x0, 0x74, 0x0, 0x47, 0x2, 0x1, 0x0, 0x18, 
  0x0, 0x82, 0x10, 0x1, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x28, 0x0, 0x66, 
  0x2, 0x1, 0x0, 0x83, 0x0, 0x38, 0x4, 0x1, 0x2, 0x1, 0x0, 0x75, 0x0, 0x57, 
  0x2, 0x1, 0x0, 0x84, 0x0, 0x48, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x91, 
  0x0, 0x19, 0x2, 0x1, 0x0, 0x92, 0x0, 0x76, 0x4, 0x1, 0x2, 0x1, 0x0, 0x67, 
  0x0, 0x29, 0x2, 0x1, 0x0, 0x85, 0x0, 0x58, 0x5c, 0x1, 0x22, 0x1, 0x10, 0x1, 
  0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x93, 0x0, 0x39, 0x2, 0x1, 0x0, 0x94, 
  0x0, 0x49, 0x4, 0x1, 0x2, 0x1, 0x0, 0x77, 0x0, 0x86, 0x2, 0x1, 0x0, 0x68, 
  0x0, 0xa1, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0xa2, 0x0, 0x2a, 0x2, 0x1, 
  0x0, 0x95, 0x0, 0x59, 0x4, 0x1, 0x2, 0x1, 0x0, 0xa3, 0x0, 0x3a, 0x2, 0x1, 
  0x0, 0x87, 0x2, 0x1, 0x0, 0x78, 0x0, 0x4a, 0x16, 0x1, 0xc, 0x1, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0xa4, 0x0, 0x96, 0x4, 0x1, 0x2, 0x1, 0x0, 0x69, 0x0, 0xb1, 
  0x2, 0x1, 0x0, 0x1b, 0x0, 0xa5, 0x6, 0x1, 0x2, 0x1, 0x0, 0xb2, 0x2, 0x1, 
  0x0, 0x5a, 0x0, 0x2b, 0x2, 0x1, 0x0, 0x88, 0x0, 0xb3, 0x10, 0x1, 0xa, 0x1, 
  0x6, 0x1, 0x2, 0x1, 0x0, 0x90, 0x2, 0x1, 0x0, 0x9, 0x0, 0xa0, 0x2, 0x1, 
  0x0, 0x97, 0x0, 0x79, 0x4, 0x1, 0x2, 0x1, 0x0, 0xa6, 0x0, 0x6a, 0x0, 0xb4, 
  0xc, 0x1, 0x6, 0x1, 0x2, 0x1, 0x0, 0x1a, 0x2, 0x1, 0x0, 0xa, 0x0, 0xb0, 
  0x2, 0x1, 0x0, 0x3b, 0x2, 0x1, 0x0, 0xb, 0x0, 0xc0, 0x4, 0x1, 0x2, 0x1, 
  0x0, 0x4b, 0x0, 0xc1, 0x2, 0x1, 0x0, 0x98, 0x0, 0x89, 0x43, 0x1, 0x22, 0x1, 
  0x10, 0x1, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x1c, 0x0, 0xb5, 0x2, 0x1, 
  0x0, 0x5b, 0x0, 0xc2, 0x4, 0x1, 0x2, 0x1, 0x0, 0x2c, 0x0, 0xa7, 0x2, 0x1, 
  0x0, 0x7a, 0x0, 0xc3, 0xa, 0x1, 0x6, 0x1, 0x2, 0x1, 0x0, 0x3c, 0x2, 0x1, 
  0x0, 0xc, 0x0, 0xd0, 0x2, 0x1, 0x0, 0xb6, 0x0, 0x6b, 0x4, 0x1, 0x2, 0x1, 
  0x0, 0xc4, 0x0, 0x4c, 0x2, 0x1, 0x0, 0x99, 0x0, 0xa8, 0x10, 0x1, 0x8, 0x1, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x8a, 0x0, 0xc5, 0x2, 0x1, 0x0, 0x5c, 0x0, 0xd1, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0xb7, 0x0, 0x7b, 0x2, 0x1, 0x0, 0x1d, 0x0, 0xd2, 
  0x9, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x2d, 0x0, 0xd3, 0x2, 0x1, 0x0, 0x3d, 
  0x0, 0xc6, 0x55, 0xfa, 0x4, 0x1, 0x2, 0x1, 0x0, 0x6c, 0x0, 0xa9, 0x2, 0x1, 
  0x0, 0x9a, 0x0, 0xd4, 0x20, 0x1, 0x10, 0x1, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 
  0x0, 0xb8, 0x0, 0x8b, 0x2, 0x1, 0x0, 0x4d, 0x0, 0xc7, 0x4, 0x1, 0x2, 0x1, 
  0x0, 0x7c, 0x0, 0xd5, 0x2, 0x1, 0x0, 0x5d, 0x0, 0xe1, 0x8, 0x1, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0x1e, 0x0, 0xe2, 0x2, 0x1, 0x0, 0xaa, 0x0, 0xb9, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0x9b, 0x0, 0xe3, 0x2, 0x1, 0x0, 0xd6, 0x0, 0x6d, 0x14, 0x1, 
  0xa, 0x1, 0x6, 0x1, 0x2, 0x1, 0x0, 0x3e, 0x2, 0x1, 0x0, 0x2e, 0x0, 0x4e, 
  0x2, 0x1, 0x0, 0xc8, 0x0, 0x8c, 0x4, 0x1, 0x2, 0x1, 0x0, 0xe4, 0x0, 0xd7, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0x7d, 0x0, 0xab, 0x0, 0xe5, 0xa, 0x1, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0xba, 0x0, 0x5e, 0x2, 0x1, 0x0, 0xc9, 0x2, 0x1, 0x0, 0x9c, 
  0x0, 0x6e, 0x8, 0x1, 0x2, 0x1, 0x0, 0xe6, 0x2, 0x1, 0x0, 0xd, 0x2, 0x1, 
  0x0, 0xe0, 0x0, 0xe, 0x4, 0x1, 0x2, 0x1, 0x0, 0xd8, 0x0, 0x8d, 0x2, 0x1, 
  0x0, 0xbb, 0x0, 0xca, 0x4a, 0x1, 0x2, 0x1, 0x0, 0xff, 0x40, 0x1, 0x3a, 0x1, 
  0x20, 0x1, 0x10, 0x1, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0xac, 0x0, 0xe7, 
  0x2, 0x1, 0x0, 0x7e, 0x0, 0xd9, 0x4, 0x1, 0x2, 0x1, 0x0, 0x9d, 0x0, 0xe8, 
  0x2, 0x1, 0x0, 0x8e, 0x0, 0xcb, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0xbc, 
  0x0, 0xda, 0x2, 0x1, 0x0, 0xad, 0x0, 0xe9, 0x4, 0x1, 0x2, 0x1, 0x0, 0x9e, 
  0x0, 0xcc, 0x2, 0x1, 0x0, 0xdb, 0x0, 0xbd, 0x10, 0x1, 0x8, 0x1, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0xea, 0x0, 0xae, 0x2, 0x1, 0x0, 0xdc, 0x0, 0xcd, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0xeb, 0x0, 0xbe, 0x2, 0x1, 0x0, 0xdd, 0x0, 0xec, 0x8, 0x1, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0xce, 0x0, 0xed, 0x2, 0x1, 0x0, 0xde, 0x0, 0xee, 
  0x0, 0xf, 0x4, 0x1, 0x2, 0x1, 0x0, 0xf0, 0x0, 0x1f, 0x0, 0xf1, 0x4, 0x1, 
  0x2, 0x1, 0x0, 0xf2, 0x0, 0x2f, 0x2, 0x1, 0x0, 0xf3, 0x0, 0x3f, 0x12, 0x1, 
  0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0xf4, 0x0, 0x4f, 0x2, 0x1, 0x0, 0xf5, 
  0x0, 0x5f, 0x4, 0x1, 0x2, 0x1, 0x0, 0xf6, 0x0, 0x6f, 0x2, 0x1, 0x0, 0xf7, 
  0x2, 0x1, 0x0, 0x7f, 0x0, 0x8f, 0xa, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0xf8, 
  0x0, 0xf9, 0x4, 0x1, 0x2, 0x1, 0x0, 0x9f, 0x0, 0xaf, 0x0, 0xfa, 0x8, 0x1, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0xfb, 0x0, 0xbf, 0x2, 0x1, 0x0, 0xfc, 0x0, 0xcf, 
  0x4, 0x1, 0x2, 0x1, 0x0, 0xfd, 0x0, 0xdf, 0x2, 0x1, 0x0, 0xfe, 0x0, 0xef
};

/* .table 25 512 16 16  5*/
/* .reference 24*/

/* .table 26 512 16 16  6*/
/* .reference 24*/

/* .table 27 512 16 16  7*/
/* .reference 24*/

/* .table 28 512 16 16  8*/
/* .reference 24*/

/* .table 29 512 16 16  9*/
/* .reference 24*/

/* .table 30 512 16 16 11*/
/* .reference 24*/

/* .table 31 512 16 16 13*/
/* .reference 24*/

/* .table 32  31  1 16  0*/
static const unsigned char treedata_32[] = {
  0x2, 0x1, 0x0, 0x0, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1,
  0x0, 0x8, 0x0, 0x4, 0x2, 0x1, 0x0, 0x1, 0x0, 0x2, 0x8, 0x1, 0x4, 0x1,
  0x2, 0x1, 0x0, 0xc, 0x0, 0xa, 0x2, 0x1, 0x0, 0x3,
  0x0, 0x6, 0x6, 0x1, 0x2, 0x1, 0x0, 0x9, 0x2, 0x1, 0x0, 0x5, 0x0, 0x7,
  0x4, 0x1, 0x2, 0x1, 0x0, 0xe, 0x0, 0xd, 0x2, 0x1, 0x0, 0xf, 0x0, 0xb,
};

/* .table 33  31  1 16  0*/
static const unsigned char treedata_33[] = {
  0x10, 0x1, 0x8, 0x1, 0x4, 0x1, 0x2, 0x1, 0x0, 0x0,
  0x0, 0x1, 0x2, 0x1, 0x0, 0x2, 0x0, 0x3, 0x4, 0x1, 0x2, 0x1, 0x0, 0x4,
  0x0, 0x5, 0x2, 0x1, 0x0, 0x6, 0x0, 0x7, 0x8, 0x1,
  0x4, 0x1, 0x2, 0x1, 0x0, 0x8, 0x0, 0x9, 0x2, 0x1, 0x0, 0xa, 0x0, 0xb,
  0x4, 0x1, 0x2, 0x1, 0x0, 0xc, 0x0, 0xd, 0x2, 0x1, 0x0, 0xe, 0x0, 0xf
};

// Information needed to recover .table lines
struct huffinfo {
  int            treelen;     // 3rd col
  int            xlen;        // 4th col
  int            ylen;        // 5th col
  int            linbits;     // 6th col
  // Information for .treedata or .reference lines
  int            ref;         // .reference table, or .treedata if -1
  const unsigned char *data;  // ptr to treedata array
};
 
static const huffinfo hufftab[HTN] = {
  {  0,  0,  0,  0, -1, NULL},		/* 0*/
  {  7,  2,  2,  0, -1, treedata_1},		/* 1*/
  { 17,  3,  3,  0, -1, treedata_2},		/* 2*/
  { 17,  3,  3,  0, -1, treedata_3},		/* 3*/
  {  0,  0,  0,  0, -1, NULL},		/* 4*/
  { 31,  4,  4,  0, -1, treedata_5},		/* 5*/
  { 31,  4,  4,  0, -1, treedata_6},		/* 6*/
  { 71,  6,  6,  0, -1, treedata_7},		/* 7*/
  { 71,  6,  6,  0, -1, treedata_8},		/* 8*/
  { 71,  6,  6,  0, -1, treedata_9},		/* 9*/
  {127,  8,  8,  0, -1, treedata_10},		/* 10*/
  {127,  8,  8,  0, -1, treedata_11},		/* 11*/
  {127,  8,  8,  0, -1, treedata_12},		/* 12*/
  {511, 16, 16,  0, -1, treedata_13},		/* 13*/
  {  0,  0,  0,  0, -1, NULL},		/* 14*/
  {511, 16, 16,  0, -1, treedata_15},		/* 15*/
  {511, 16, 16,  1, -1, treedata_16},		/* 16*/
  {511, 16, 16,  2, 16, NULL},		/* 17*/
  {511, 16, 16,  3, 16, NULL},		/* 18*/
  {511, 16, 16,  4, 16, NULL},		/* 19*/
  {511, 16, 16,  6, 16, NULL},		/* 20*/
  {511, 16, 16,  8, 16, NULL},		/* 21*/
  {511, 16, 16, 10, 16, NULL},		/* 22*/
  {511, 16, 16, 13, 16, NULL},		/* 23*/
  {512, 16, 16,  4, -1, treedata_24},		/* 24*/
  {512, 16, 16,  5, 24, NULL},		/* 25*/
  {512, 16, 16,  6, 24, NULL},		/* 26*/
  {512, 16, 16,  7, 24, NULL},		/* 27*/
  {512, 16, 16,  8, 24, NULL},		/* 28*/
  {512, 16, 16,  9, 24, NULL},		/* 29*/
  {512, 16, 16, 11, 24, NULL},		/* 30*/
  {512, 16, 16, 13, 24, NULL},		/* 31*/
  { 31,  1, 16,  0, -1, treedata_32},		/* 32*/
  { 31,  1, 16,  0, -1, treedata_33}		/* 33*/
};

struct CSAPI_MPEG huffcodetab {
  char tablename[3];      // string, containing table_description
  unsigned int xlen;      // max. x-index+
  unsigned int ylen;      // max. y-index+
  unsigned int linbits;   // number of linbits
  unsigned int linmax;    // max number to be stored in linbits
  int ref;                // a positive value indicates a reference
  unsigned char(*val)[2]; // decoder tree
  unsigned int treelen;   // length of decoder tree
};
 
// huffman encode table
static huffcodetab ht[HTN];
 
//--------------------------------------------------------------------------

// constructor
CSAPI_MPEG
Layer3::Layer3 (MPEGfile *mp)
{
  // set reference to audio frame handler
  frame = mp;
    
  // initialize buffer indexes
  bufstart=bufend=whichbit=0;
    
  // initialize huffman decoding
  if (!read_decoder_table()) {
    cerr << "MaaateP: Error reading huffman decoder table" << endl;
    exit(1);
  }

#ifdef DEBUG
  noReadBits=0;
#endif

  return;
}

// initialize the huffman decoder table
CSAPI_MPEG bool 
Layer3::read_decoder_table()
{
  const unsigned char *p;
  int nn, t;
  unsigned int i;
  int n;
  for (n=0; n<HTN; n++) {
    sprintf(ht[n].tablename, "%d", n);
    ht[n].treelen = hufftab[n].treelen;
    ht[n].xlen	  = hufftab[n].xlen;
    ht[n].ylen	  = hufftab[n].ylen;
    ht[n].linbits = hufftab[n].linbits;
    ht[n].linmax  = (1 << ht[n].linbits) - 1;

    // Check table name
    //    nn;
    sscanf(ht[n].tablename, "%u", &nn);
    if (nn != n) {
      cerr << "MaaateP: wrong table number " << n << endl;
      return false;
    } 

    t = hufftab[n].ref;
    if (t >= 0) {
      ht[n].ref		= t;
      ht[n].val		= ht[t].val;
      ht[n].treelen	= ht[t].treelen;
      if ((ht[n].xlen != ht[t].xlen) || (ht[n].ylen != ht[t].ylen)) {
	cerr << "MaaateP: wrong table reference " << n << endl;
	return false;
      };
    } else if (t == -1) {
      ht[n].ref	= -1;
      ht[n].val	= (unsigned char (*)[2]) 
	calloc(2 * (ht[n].treelen)+1, sizeof(unsigned char));
      if (ht[n].val == NULL) {
	cerr << "MaaateP: heap error while loading table " << n << endl;
	cerr << "attempting calloc " 
	     << 2 * (ht[n].treelen) << "\t"
	     << sizeof(unsigned char) << endl;
	return false;
      }
      p = hufftab[n].data;
      for (i = 0; i < ht[n].treelen; i++) {
	ht[n].val[i][0] = *p++;
	ht[n].val[i][1] = *p++;
      }
    } else {
      cerr << "MaaateP: huffman decodertable error at table " << n << endl;
      return false;
    }
  }
  return (n==HTN);
}

// parse audio data of layer 3 frame
CSAPI_MPEG bool 
Layer3::parse_data(DecodeLevel decode) 
{
  // parse side information
  decode_sideinfo();

  // initialize/reset counters
  count1_vals[0][0] = count1_vals[0][1] = 0; // counts no count1 values
  count1_vals[1][0] = count1_vals[1][1] = 0; // within frame
  spectral_bitsread=0; // counts no huffman bits read in frame
  main_data_bitsread=0; // counts no of main_data bits read in frame
    
  // check crc checksum
  if (frame->crcprotected()) {
    if (frame->checksum != calc_CRC()) {
      cerr << "MaaateP: WARNING: wrong checksum" << endl;
      //            return false;
    }
  }
    
  // set start reading of buffer to main_data_begin
  if (!setbufstart()) {
    // if not enough main_data available, skip frame
    savetointerbuffer();
    return false;
  }
    
  const int s_gr = granules();
  const int s_ch = frame->channels();
  const int s_ver = frame->version();
  int gr, ch;
  for (gr=0; gr<s_gr; gr++) {
    for (ch=0; ch<s_ch; ch++) {

      // parse scalefactor information
      if (s_ver == MPEG1) {
	decode_scale_I(gr, ch);
      } else { // MPEG2
	decode_scale_II(gr,ch);
      }
	    
      // parse huffman encoded sample data
      decode_huffmanbits(gr, ch);
    
      if (decode > dec_fields) {
	// restore the subband samples
	restore_samples(gr, ch);
      }

    }
    if (decode > dec_fields) {
      // joint stereo processing
      decode_jstereo(gr);

      if (decode > dec_subsubbands) {
	for (ch=0; ch<s_ch; ch++) {
	  // reorder spectrum
	  reorder(gr, ch);
	  
	  // alias reduction
	  antialias(gr, ch);
	  
	  // hybrid filter synthesis
	  hybrid(gr, ch);
	 
	  if (decode > dec_subbands) {
	    // polyphase filterband synthesis: creates pcm samples
	    polyphase(gr, ch);
	  }
	}
      }
    }
  }

#ifdef DEBUG
  noReadBits += main_data_bitsread;
  //    cerr << "noReadBits=" << noReadBits << endl;
#endif

  // save remaining bytes from frame->buffer into intermediate buffer
  savetointerbuffer();
   return true;
}

// parses layerIII-specific side information header
CSAPI_MPEG void 
Layer3::decode_sideinfo()
{
  const int s_ch = frame->channels();
  int ch, gr, i;
  unsigned int scfc;
  if (frame->version() == MPEG1) {
    // start of main_data (offset before header and stored in inter_buffer)
    // max value of main_data: MAX_INTER_BUFSIZE 
    si.main_data_begin = frame->readbitsfrombuffer(9);
	
    // for private use
    if (s_ch == 1) {
      si.private_bits = frame->readbitsfrombuffer(5);
    } else {
      si.private_bits = frame->readbitsfrombuffer(3);
    }
	
    // scale factor selection information
    for (ch = 0; ch < s_ch; ch++) {
      for (scfc = 0; scfc < 4; scfc++) { // scfsi-classes
	si.ch[ch].scfsi[scfc] = frame->readbitsfrombuffer(1);
      }
    }
	
    for (gr = 0; gr < 2 ; gr++) {
      for (ch = 0; ch < s_ch; ch++) {
	// no. of main_data bits used for scalefactors and huffman data
	si.ch[ch].gr[gr].part2_3_length 
	  = frame->readbitsfrombuffer(12);
		
	// no. of value pairs at low frequencies (<8191)
	si.ch[ch].gr[gr].big_values = frame->readbitsfrombuffer(9);
		
	// quantizer step size information
	si.ch[ch].gr[gr].global_gain = frame->readbitsfrombuffer(8);
		
	// no. of bits used for scalefactors
	si.ch[ch].gr[gr].scalefac_compress 
	  = frame->readbitsfrombuffer(4);
		
	// window switching flag
	si.ch[ch].gr[gr].window_switching_flag
	  = frame->readbitsfrombuffer(1);
		
	if (si.ch[ch].gr[gr].window_switching_flag) {
	  // window type for granule
	  si.ch[ch].gr[gr].block_type 
	    = frame->readbitsfrombuffer(2);
		    
	  // different windows for low and high frequencies?
	  si.ch[ch].gr[gr].mixed_block_flag
	    = frame->readbitsfrombuffer(1);
		    
	  // which huffman table to use in area i of big_values
	  for (i = 0; i < 2; i++) {
	    si.ch[ch].gr[gr].table_select[i]
	      = frame->readbitsfrombuffer(5);
	  }
		    
	  // gain offset from global gain with short windows
	  for (i = 0; i < 3; i++) {
	    si.ch[ch].gr[gr].subblock_gain[i]
	      = frame->readbitsfrombuffer(3);
	  }
		    
	  // Set region_count parameters (spectrum partitioning)
	  // since they are implicit in this case.
	  if (blocktype(ch,gr) == 0) {
	    cerr << "MaaateP: Bad side info: blocktype 0 in split block"
		 << endl;
	    exit(0);
	  } else if (blocktype(ch,gr) == 2 && mixedblock(ch,gr)) {
	    si.ch[ch].gr[gr].region0_count = 8; /* MI 9; */
	  } else {
	    si.ch[ch].gr[gr].region0_count = 7; /* MI 8; */
	  }
	  si.ch[ch].gr[gr].region1_count
	    =20 - si.ch[ch].gr[gr].region0_count;

	} else { // no window switching
	  si.ch[ch].gr[gr].block_type = 0;
	  si.ch[ch].gr[gr].mixed_block_flag = 0;
		    
	  // which huffman table to use in area i of big_values
	  for (int i = 0; i < 3; i++) {
	    si.ch[ch].gr[gr].table_select[i]
	      = frame->readbitsfrombuffer(5);
	  }

	  // spectrum partitioning
	  si.ch[ch].gr[gr].region0_count
	    = frame->readbitsfrombuffer(4);
	  si.ch[ch].gr[gr].region1_count
	    = frame->readbitsfrombuffer(3);
	}

	// additional high frequency amplification?
	si.ch[ch].gr[gr].preflag = frame->readbitsfrombuffer(1);

	// scale factor scaling
	si.ch[ch].gr[gr].scalefac_scale 
	  = frame->readbitsfrombuffer(1);

	// which huffman table for quadruples in are count1
	si.ch[ch].gr[gr].count1table_select
	  = frame->readbitsfrombuffer(1);

      }
    }

  } else {	// MPEG2

    si.main_data_begin = frame->readbitsfrombuffer(8);
    if (s_ch == 1) {
      si.private_bits = frame->readbitsfrombuffer(1);
    } else {
      si.private_bits = frame->readbitsfrombuffer(2);
    }

    for (gr = 0; gr < 1 ; gr++) {
      for (ch = 0; ch < s_ch; ch++) {
	si.ch[ch].gr[gr].part2_3_length  
	  = frame->readbitsfrombuffer(12);
	si.ch[ch].gr[gr].big_values
	  = frame->readbitsfrombuffer(9);
	si.ch[ch].gr[gr].global_gain
	  = frame->readbitsfrombuffer(8);
	si.ch[ch].gr[gr].scalefac_compress
	  = frame->readbitsfrombuffer(9);
	si.ch[ch].gr[gr].window_switching_flag
	  = frame->readbitsfrombuffer(1);
	if (si.ch[ch].gr[gr].window_switching_flag) {
	  si.ch[ch].gr[gr].block_type
	    = frame->readbitsfrombuffer(2);
	  si.ch[ch].gr[gr].mixed_block_flag
	    = frame->readbitsfrombuffer(1);
	  for (i = 0; i < 2; i++) {
	    si.ch[ch].gr[gr].table_select[i]
	      = frame->readbitsfrombuffer(5);
	  }
	  for (i = 0; i < 3; i++) {
	    si.ch[ch].gr[gr].subblock_gain[i]
	      = frame->readbitsfrombuffer(3);
	  }

	  // Set region_count parameters since they are
	  // implicit in this case.
	  if (blocktype(ch,gr) == 0) {
	    cerr << "MaaateP: Bad side info: blocktype 0 in split block"
		 << endl;
	    exit(0);
	  } else if (blocktype(ch,gr) == 2 && mixedblock(ch,gr)) {
	    si.ch[ch].gr[gr].region0_count = 8; /* MI 9; */
	  } else {
	    si.ch[ch].gr[gr].region0_count = 7; /* MI 8; */
	  }
	  si.ch[ch].gr[gr].region1_count =
	    20 - si.ch[ch].gr[gr].region0_count;
		    
	} else { // no window switching
	  si.ch[ch].gr[gr].block_type = 0;
	  si.ch[ch].gr[gr].mixed_block_flag = 0;

	  for (i = 0; i < 3; i++) {
	    si.ch[ch].gr[gr].table_select[i]
	      = frame->readbitsfrombuffer(5);
	  }
	  si.ch[ch].gr[gr].region0_count
	    = frame->readbitsfrombuffer(4);
	  si.ch[ch].gr[gr].region1_count
	    = frame->readbitsfrombuffer(3);
	}      
		
	si.ch[ch].gr[gr].scalefac_scale
	  = frame->readbitsfrombuffer(1);
	si.ch[ch].gr[gr].count1table_select
	  = frame->readbitsfrombuffer(1);
      }
    }
  }
}

CSAPI_MPEG void 
Layer3::printSideinfo()
{
  cout << "\tLayer 3 Side information:" << endl;
  cout << "\tMain_data_begin=" << main_data_begin() << endl;
  cout << "\tPrivate_bits="    << private_bits()    << endl;
  const int s_ch = frame->channels();
  int ch, area;
  unsigned int scfc;
#ifdef DEBUG
  int gr, sfb;
#endif
  for (ch=0; ch<s_ch; ch++) {
    for (scfc=0; scfc<4; scfc++) {
      cout << "\tscfsi[" << ch << "][" << scfc << "]="
	   << scfsi(ch,scfc) << endl;
    }
    for (unsigned gr=0; gr<granules(); gr++) {
      cout << "\tpart2_3_length[" << ch << "][" << gr << "]="
	   << part2_3_length(ch,gr) << endl;
      cout << "\tbig_values[" << ch << "][" << gr << "]="
	   << big_values(ch,gr) << endl;
      cout << "\tglobal_gain[" << ch << "][" << gr << "]="
	   << global_gain(ch,gr) << endl;
      cout << "\tscalefac_compress[" << ch << "][" << gr << "]="
	   << scalefac_compress(ch,gr) << endl;
      cout << "\twindow_switching[" << ch << "][" << gr << "]="
	   << window_switching(ch,gr) << endl;
      cout << "\tblocktype[" << ch << "][" << gr << "]="
	   << blocktype(ch,gr) 
	   << " (" << blocktype_str(ch,gr) << ")" << endl;
      cout << "\tmixedblock[" << ch << "][" << gr << "]="
	   << mixedblock(ch,gr) << endl;
      for (area=0; area<3; area++) {
	cout << "\ttable_select[" << ch << "][" << gr << "][" 
	     << area << "]=" << table_select(ch,gr,area) << endl;
	cout << "\tsubblock_gain[" << ch << "][" << gr << "][" 
	     << area << "]=" << subblock_gain(ch,gr,area) << endl;
      }
      cout << "\tregion0_count[" << ch << "][" << gr << "]="
	   << region0_count(ch,gr) << endl;
      cout << "\tregion1_count[" << ch << "][" << gr << "]="
	   << region1_count(ch,gr) << endl;
      cout << "\tpreflag[" << ch << "][" << gr << "]="
	   << preflag(ch,gr) << endl;
      cout << "\tscalefac_scale[" << ch << "][" << gr << "]="
	   << scalefac_scale(ch,gr) << endl;
      cout << "\tcount1table_select[" << ch << "][" << gr << "]="
	   << count1table_select(ch,gr) << endl;
#ifdef DEBUG
      cout << "\tregion0_samps[" << ch << "][" << gr << "]="
	   << region0_samps(ch,gr) << endl;
      cout << "\tregion1_samps[" << ch << "][" << gr << "]="
	   << region1_samps(ch,gr) << endl;
      cout << "\tregion2_samps[" << ch << "][" << gr << "]="
	   << region2_samps(ch,gr) << endl;
      cout << "\tcount1_values[" << ch << "][" << gr << "]="
	   << count1_values(ch,gr) << endl;
      cout << "\tslen1[" << ch << "][" << gr << "]="
	   << slen1(ch,gr) << endl;
      cout << "\tslen2[" << ch << "][" << gr << "]="
	   << slen2(ch,gr) << endl;
      cout << "\tscalefacs[" << ch << "][" << gr << "]=";
      if (window_switching(ch,gr) && (blocktype(ch,gr)==2)) {
	if (mixedblock(ch,gr)) { // mixed
	  for (sfb=0; sfb<8; sfb++) { // long
	    cout << " " << scalefac[ch][gr].l[sfb];
	  }
	  for (sfb=3; sfb<13; sfb++) { // short
	    cout << " " << scalefac[gr][ch].s[0][sfb];
	    cout << " " << scalefac[gr][ch].s[1][sfb];
	    cout << " " << scalefac[gr][ch].s[2][sfb];
	  }
	} else {
	  for (sfb=0; sfb<13; sfb++) { // short only
	    cout << " " << scalefac[gr][ch].s[0][sfb];
	    cout << " " << scalefac[gr][ch].s[1][sfb];
	    cout << " " << scalefac[gr][ch].s[2][sfb];
	  }
	}
      } else {
	for (sfb=0; sfb<21; sfb++) { // long
	  cout <<  " " << scalefac[gr][ch].l[sfb];
	}
      }
      cout << endl;
#endif
    }
  }
  return;
}

// calculates checksum on required fields
CSAPI_MPEG unsigned int 
Layer3::calc_CRC()
{
  unsigned int crc = 0xffff;
  // header
  frame->update_CRC(frame->header.bitrate,         4, &crc);
  frame->update_CRC(frame->header.samplingrate,    2, &crc);
  frame->update_CRC(frame->header.padding,         1, &crc);
  frame->update_CRC(frame->header.extension,       1, &crc);
  frame->update_CRC(frame->header.mode,            2, &crc);
  frame->update_CRC(frame->header.mode_ext,        2, &crc);
  frame->update_CRC(frame->header.copyright,       1, &crc);
  frame->update_CRC(frame->header.original,        1, &crc);
  frame->update_CRC(frame->header.emphasis,        2, &crc);

  // side info
  const int s_ch = frame->channels();
  int ch, gr, i;
  unsigned int scfc;
  if (frame->version() == MPEG1) {
    frame->update_CRC(si.main_data_begin,  9, &crc);
    if (s_ch == 1) {
      frame->update_CRC(si.private_bits, 5, &crc);
    } else {
      frame->update_CRC(si.private_bits, 3, &crc);
    }
    for (ch=0; ch<s_ch; ch++) {
      for (scfc = 0; scfc < 4; scfc++) {
	frame->update_CRC(si.ch[ch].scfsi[scfc], 1, &crc);
      }
    }
    for (gr = 0; gr < 2 ; gr++) {
      for (ch = 0; ch < s_ch; ch++) {
	frame->update_CRC(si.ch[ch].gr[gr].part2_3_length,
			  12, &crc);
	frame->update_CRC(si.ch[ch].gr[gr].big_values,
			  9, &crc);
	frame->update_CRC(si.ch[ch].gr[gr].global_gain,
			  8, &crc);
	frame->update_CRC(si.ch[ch].gr[gr].scalefac_compress,
			  4, &crc);
	frame->update_CRC(si.ch[ch].gr[gr].window_switching_flag,
			  1, &crc);
	if (si.ch[ch].gr[gr].window_switching_flag) {
	  frame->update_CRC(si.ch[ch].gr[gr].block_type,
			    2, &crc);
	  frame->update_CRC(si.ch[ch].gr[gr].mixed_block_flag,
			    1, &crc);
	  for (i = 0; i < 2; i++) {
	    frame->update_CRC(si.ch[ch].gr[gr].table_select[i],
			      5, &crc);
	  }
	  for (i = 0; i < 3; i++) {
	    frame->update_CRC(si.ch[ch].gr[gr].subblock_gain[i],
			      3, &crc);
	  }
	} else { // no window switching
	  for (i = 0; i < 3; i++) {
	    frame->update_CRC(si.ch[ch].gr[gr].table_select[i],
			      5, &crc);
	  }
	  frame->update_CRC(si.ch[ch].gr[gr].region0_count,
			    4, &crc);
	  frame->update_CRC(si.ch[ch].gr[gr].region1_count,
			    3, &crc);
	}
	frame->update_CRC(si.ch[ch].gr[gr].preflag,
			  1, &crc);
	frame->update_CRC(si.ch[ch].gr[gr].scalefac_scale,
			  1, &crc);
	frame->update_CRC(si.ch[ch].gr[gr].count1table_select,
			  1, &crc);
      }
    }
  } else { // MPEG2
    frame->update_CRC(si.main_data_begin,  8, &crc);
    if (s_ch == 1) {
      frame->update_CRC(si.private_bits, 1, &crc);
    } else {
      frame->update_CRC(si.private_bits, 2, &crc);
    }
    for (gr = 0; gr < 1 ; gr++) {
      for (ch = 0; ch < s_ch; ch++) {
	frame->update_CRC(si.ch[ch].gr[gr].part2_3_length,
			  12, &crc);
	frame->update_CRC(si.ch[ch].gr[gr].big_values,
			  9, &crc);
	frame->update_CRC(si.ch[ch].gr[gr].global_gain,
			  8, &crc);
	frame->update_CRC(si.ch[ch].gr[gr].scalefac_compress,
			  9, &crc);
	frame->update_CRC(si.ch[ch].gr[gr].window_switching_flag,
			  1, &crc);
	if (si.ch[ch].gr[gr].window_switching_flag) {
	  frame->update_CRC(si.ch[ch].gr[gr].block_type,
			    2, &crc);
	  frame->update_CRC(si.ch[ch].gr[gr].mixed_block_flag,
			    1, &crc);
	  for (i = 0; i < 2; i++) {
	    frame->update_CRC(si.ch[ch].gr[gr].table_select[i],
			      5, &crc);
	  }
	  for (i = 0; i < 3; i++) {
	    frame->update_CRC(si.ch[ch].gr[gr].subblock_gain[i],
			      3, &crc);
	  }
	} else { // no window switching
	  for (i = 0; i < 3; i++) {
	    frame->update_CRC(si.ch[ch].gr[gr].table_select[i],
			      5, &crc);
	  }
	  frame->update_CRC(si.ch[ch].gr[gr].region0_count,
			    4, &crc);
	  frame->update_CRC(si.ch[ch].gr[gr].region1_count,
			    3, &crc);
	}      
	frame->update_CRC(si.ch[ch].gr[gr].scalefac_scale,
			  1, &crc);
	frame->update_CRC(si.ch[ch].gr[gr].count1table_select,
			  1, &crc);
      }
    }
  }

  return crc;
}


// parse the scale factors (MPEG1)
CSAPI_MPEG void 
Layer3::decode_scale_I(int gr, int ch)
{
  int window;
  unsigned int sfb, scfc;
    
  if (window_switching(ch,gr) && (blocktype(ch,gr) == 2)) { 
	
    if (mixedblock(ch,gr)) {	/// MIXED
      for (sfb=0; sfb<8; sfb++) {
	scalefac[gr][ch].l[sfb] = readbits(slen1(ch,gr));
      }
      for (sfb=3; sfb<6; sfb++) {
	for (window=0; window<3; window++) {
	  scalefac[gr][ch].s[window][sfb] 
	    = readbits(slen1(ch,gr));
	}
      }
      for (sfb = 6; sfb < 12; sfb++) {
	for (window=0; window<3; window++) {
	  scalefac[gr][ch].s[window][sfb] 
	    = readbits(slen2(ch,gr));
	}
      }
      for (window=0; window<3; window++) {
	scalefac[gr][ch].s[window][12] = 0;
      }
	    
    } else { // not mixed; SHORT
	    
      for (sfb=0; sfb<6; sfb++) {
	for (window=0; window<3; window++) {
	  scalefac[gr][ch].s[window][sfb] 
	    = readbits (slen1(ch,gr));
	}
      }
      for (sfb=6; sfb<12; sfb++) {
	for (window=0; window<3; window++) {
	  scalefac[gr][ch].s[window][sfb] 
	    = readbits (slen2(ch,gr));
	}
      }
      for (window=0; window<3; window++) {
	scalefac[gr][ch].s[window][12] = 0;
      }
    }
	
  } else {   // blocktypes 0,1,3; LONG
	
    for (scfc=0; scfc<2; scfc++) {
      if ((scfsi(ch,scfc) == 0) || (gr == 0)) {
	for (sfb=sfbtable[0][scfc]; 
	     sfb<sfbtable[0][scfc+1];
	     sfb++) {
	  scalefac[gr][ch].l[sfb] = readbits(slen1(ch,gr));
	}
      }
    }
    for (scfc=2; scfc<4; scfc++) {
      if ((scfsi(ch,scfc) == 0) || (gr == 0)) {
	for (sfb=sfbtable[0][scfc]; 
	     sfb<sfbtable[0][scfc+1];
	     sfb++) {
	  scalefac[gr][ch].l[sfb] = readbits(slen2(ch,gr));
	}
      }
    }
	
  }
  scalefac[gr][ch].l[21] = 0; 

  return;
}


// parses the scale factors (MPEG2)
CSAPI_MPEG void 
Layer3::decode_scale_II(int gr, int ch)
{
  unsigned int scalefac_buffer[54];
  unsigned int new_slen[4];
  short blocktypenumber=0, blocknumber = 0;
    
  if (blocktype(ch,gr) == 2 && mixedblock(ch,gr) == 0)
    blocktypenumber = 1;
    
  if (blocktype(ch,gr) == 2 && mixedblock(ch,gr) == 1)
    blocktypenumber = 2;
    
  if (!((frame->mode_ext() & 0x1) && ch == 1)) {
    // not (intensity stereo and second channel)
    if(scalefac_compress(ch,gr) < 400) {
      new_slen[0] = (scalefac_compress(ch,gr) >> 4) / 5 ;
      new_slen[1] = (scalefac_compress(ch,gr) >> 4) % 5 ;
      new_slen[2] = (scalefac_compress(ch,gr) % 16) >> 2 ;
      new_slen[3] = (scalefac_compress(ch,gr) % 4);
      si.ch[ch].gr[gr].preflag = 0;
      blocknumber = 0;
	    
    } else if(scalefac_compress(ch,gr) < 500) {
      new_slen[0] = ((scalefac_compress(ch,gr) - 400 )  >> 2) / 5 ;
      new_slen[1] = ((scalefac_compress(ch,gr) - 400) >> 2) % 5 ;
      new_slen[2] = (scalefac_compress(ch,gr) - 400 ) % 4 ;
      new_slen[3] = 0;
      si.ch[ch].gr[gr].preflag = 0;
      blocknumber = 1;
	    
    } else if(scalefac_compress(ch,gr) < 512) {
      new_slen[0] = (scalefac_compress(ch,gr) - 500 ) / 3 ;
      new_slen[1] = (scalefac_compress(ch,gr) - 500)  % 3 ;
      new_slen[2] = 0 ;
      new_slen[3] = 0;
      si.ch[ch].gr[gr].preflag = 1;
      blocknumber = 2;
    }
  }

  if ((frame->mode_ext() & 0x1) && ch == 1) {
    // intensity stereo and second channel
    /*   intensity_scale = scalefac_comp %2; */
    unsigned int intensity_scalefac_comp = scalefac_compress(ch,gr) >> 1;
	
    if (intensity_scalefac_comp < 180) {
      new_slen[0] = intensity_scalefac_comp  / 36 ;
      new_slen[1] = (intensity_scalefac_comp % 36 ) / 6 ;
      new_slen[2] = (intensity_scalefac_comp % 36) % 6;
      new_slen[3] = 0;
      si.ch[ch].gr[gr].preflag = 0;
      blocknumber = 3;

    } else if (intensity_scalefac_comp < 244) {
      new_slen[0] = ((intensity_scalefac_comp - 180 )  % 64 ) >> 4 ;
      new_slen[1] = ((intensity_scalefac_comp - 180) % 16) >> 2 ;
      new_slen[2] = (intensity_scalefac_comp - 180 ) % 4 ;
      new_slen[3] = 0;
      si.ch[ch].gr[gr].preflag = 0;
      blocknumber = 4;
	    
    } else if (intensity_scalefac_comp < 255) {
      new_slen[0] = (intensity_scalefac_comp - 244 ) / 3 ;
      new_slen[1] = (intensity_scalefac_comp - 244 )  % 3 ;
      new_slen[2] = 0 ;
      new_slen[3] = 0;
      si.ch[ch].gr[gr].preflag = 0;
      blocknumber = 5;
    }
  }
     
  // read scale factors into buffer
  int k = 0;
  int i, window;
  unsigned int sfb, j;
  for (i=0; i<4; i++) {
    for(j=0; j<nr_of_sfb_block[blocknumber][blocktypenumber][i]; j++) {
      if(new_slen[i] == 0) {
	scalefac_buffer[k++] = 0;
      } else {   
	scalefac_buffer[k++] =  readbits(new_slen[i]);
      }
    }
  }

  k = 0;
  if (window_switching(ch,gr) && (blocktype(ch,gr) == 2)) { 

    if (mixedblock(ch,gr)) { // MIXED
      for (sfb = 0; sfb < 8; sfb++) {
	scalefac[gr][ch].l[sfb] = scalefac_buffer[k++];
      } 
      for (sfb = 3; sfb < 12; sfb++) {
	for (window=0; window<3; window++) {
	  scalefac[gr][ch].s[window][sfb] = scalefac_buffer[k++];
	}
      }
	    
      for (window=0; window<3; window++) {
	scalefac[gr][ch].s[window][12] = 0;
      }

    } else { // not mixed; SHORT
	    
      for (sfb=0; sfb<12; sfb++) {
	for (window=0; window<3; window++) {
	  scalefac[gr][ch].s[window][sfb] = scalefac_buffer[k++];
	}
      }
      for (window=0; window<3; window++) {
	scalefac[gr][ch].s[window][12] = 0;
      }
    }
	
  } else {   // blocktypes 0,1,3; LONG
	
    for (sfb = 0; sfb < 21; sfb++) {
      scalefac[gr][ch].l[sfb] = scalefac_buffer[k++];
    }
    scalefac[gr][ch].l[21] = 0; 
  }

  return;
}


// parse the huffman encoded sample data
// result: "is" data array
CSAPI_MPEG void 
Layer3::decode_huffmanbits(int gr, int ch)
{
  unsigned int x, y, v, w; // quantized samples
  int region1Start, region2Start;
  int i;

  // find region boundaries for short block case
  region1Start = region0_samps(ch,gr);
  region2Start = region1Start + region1_samps(ch,gr);

#ifdef DEBUG
  //    cerr << "maindata bits read=" << main_data_bitsread+noReadBits << endl;
#endif

  // calculate end of this spectrum
  endSpectralBits = spectral_bitsread + part2_3_length(ch,gr);

  // decode samples within big_values area
  struct huffcodetab *h;
  int endloop = big_values(ch,gr) * 2;
  for (i = 0; i < endloop; i += 2) {
    // Select the huffman decoding table
    if (i < region1Start) {
      h = &ht[table_select(ch,gr,0)];
    } else if (i<region2Start) {
      h = &ht[table_select(ch,gr,1)];
    } else {
      h = &ht[table_select(ch,gr,2)];
    }
    huffman_decoder(h, &x, &y, &v, &w);
    is[gr][ch][i / SSLIMIT][i % SSLIMIT] = x;
    is[gr][ch][(i+1) / SSLIMIT][(i+1) % SSLIMIT] = y;
  }
	    
  // decode samples within count1 area
  h = &ht[count1table_select(ch,gr) + 32];
  while (main_data_bitsread < endSpectralBits && i < SBSSLIMIT ) {
    if (huffman_decoder(h, &x, &y, &v, &w)) {
      is[gr][ch][i / SSLIMIT][i % SSLIMIT] = v;
      is[gr][ch][(i+1) / SSLIMIT][(i+1) % SSLIMIT] = w;
      is[gr][ch][(i+2) / SSLIMIT][(i+2) % SSLIMIT] = x;
      is[gr][ch][(i+3) / SSLIMIT][(i+3) % SSLIMIT] = y;
      i += 4;
    }
    // update no of count1 values
    count1_vals[gr][ch]++;
  }
    
    
#ifdef DEBUG
  //    cerr << "2:mdb=" << main_data_bitsread+noReadBits << endl;
#endif

  // Dismiss stuffing Bits (no more than 32 at a time)
  if (main_data_bitsread < endSpectralBits) {
    unsigned int stuffed = endSpectralBits - main_data_bitsread;
    while (stuffed > 0) {
      if (stuffed < 32) {
	readbits(stuffed);
	stuffed = 0;
      } else {
	readbits(32);
	stuffed -= 32;
      }
    }
  } else if (main_data_bitsread > endSpectralBits) {
    cerr << "MaaateP: Error: can't rewind stream by " 
	 << main_data_bitsread-endSpectralBits
	 << " bits." << endl;
  }

  // update spectral_bitsread; (main_data_bitsread gets updated in readbits)
  spectral_bitsread = main_data_bitsread;

  // Zero out rest (SSLMIT*SBLIMIT=576=SBSSLIMIT)
  if (i < SBSSLIMIT)
    memset ((void*)&is[gr][ch][i/SSLIMIT][i%SSLIMIT], 0,
	    (SBSSLIMIT-i)*sizeof(long int));

  /* original code
  for (; i < SBSSLIMIT; i++) { 
    is[gr][ch][i / SSLIMIT][i % SSLIMIT] = 0;
  }
  */


  return;
}


//  Do the huffman-decoding of the sample data
//  Note! for counta,countb -the 4 bit value is returned in y, discard x
CSAPI_MPEG bool 
Layer3::huffman_decoder(struct huffcodetab *h, // huffman code record
			unsigned int *x,       // decoded x value
			unsigned int *y,       // decoded y value
			unsigned int *v,       // decoded v value
			unsigned int *w)       // decoded w value
{  
  unsigned int point = 0;
  int          error = 1;

  if (h->val == NULL)
    return true;
 
  // 1000...0000
  unsigned int level = (unsigned int)1 << (sizeof(unsigned int)*8-1);

  // Table 0 needs no bits
  if ( h->treelen == 0) {
    *x = *y = 0;
    return true;
  }
 
  // Lookup in Huffman table reading data bits
  do {
    if (h->val[point][0]==0) {      /* end of tree */
      *x = h->val[point][1] >> 4;
      *y = h->val[point][1] & 0xf;
      error = 0;
      break;
    } 
    if ((main_data_bitsread + 1) > endSpectralBits) {
      return false;
    }
    if (readbits(1)) {
      while (h->val[point][1] >= MXOFF) point += h->val[point][1]; 
      point += h->val[point][1];
    } else {
      while (h->val[point][0] >= MXOFF) point += h->val[point][0]; 
      point += h->val[point][0];
    }
    level >>= 1;
  } while (level || (point < ht->treelen) );
  
  // Check for error
  if (error) { // set x and y to a medium value as a simple concealment
    cerr << "MaaateP: Illegal Huffman code in data." << endl;
    *x = ((h->xlen-1) << 1);
    *y = ((h->ylen-1) << 1);
  }
    
  // Process sign encodings for quadruples tables
  if (h->tablename[0] == '3' && (h->tablename[1] == '2' ||
				 h->tablename[1] == '3')) {
    *v = (*y>>3) & 1;
    *w = (*y>>2) & 1;
    *x = (*y>>1) & 1;
    *y = *y & 1;
 
    if (*v) {
      if ((main_data_bitsread + 1) > endSpectralBits) {
	return false;
      }
      if (readbits(1) == 1) *v = -*v;
    }
    if (*w) {
      if ((main_data_bitsread + 1) > endSpectralBits) {
	return false;
      }
      if (readbits(1) == 1) *w = -*w;
    }
    if (*x) {
      if ((main_data_bitsread + 1) > endSpectralBits) {
	return false;
      }
      if (readbits(1) == 1) *x = -*x;
    }
    if (*y) {
      if ((main_data_bitsread + 1) > endSpectralBits) {
	return false;
      }
      if (readbits(1) == 1) *y = -*y;
    }
  } else {
    // Process sign and escape encodings for dual tables
	
    if (h->linbits)
      if ((h->xlen-1) == *x) {
	if ((main_data_bitsread + h->linbits) > endSpectralBits) {
	  *x += readbits(endSpectralBits - main_data_bitsread);
	  return false;
	}
	*x += readbits(h->linbits);
      }
    if (*x) {
      if ((main_data_bitsread + 1) > endSpectralBits) {
	return false;
      }
      if (readbits(1) == 1) *x = -*x;
    }
    if (h->linbits)   
      if ((h->ylen-1) == *y) {
	if ((main_data_bitsread + h->linbits) > endSpectralBits) {
	  *x += readbits(endSpectralBits - main_data_bitsread);
	  return false;
	}
	*y += readbits(h->linbits);
      }
    if (*y) {
      if ((main_data_bitsread + 1) > endSpectralBits) {
	return false;
      }
      if (readbits(1) == 1) *y = -*y;
    }
  }
  return true;
}


// restore the quantized value from "is"
// result: "xr" data array
CSAPI_MPEG void 
Layer3::restore_samples(int gr, int ch)
{
  int cb = 0;
  int cb_begin = 0, cb_width = 0;
  int sb, ss;
  unsigned int next_cb_boundary, boundary;

  // compute overall scaling for this channel and granule
  const double globalgain = pow(2.0, (0.25 * (global_gain(ch,gr) - 210.0)));

  // compute scalefactor scaling for this channel and granule
  const double scfscale = -0.5 * (1.0 + scalefac_scale(ch,gr));

  // Choose correct scalefactor band per block type, 
  // initalize boundary
  if (window_switching(ch,gr) && blocktype(ch,gr) == 2) {
    if (mixedblock(ch,gr)) {
      // mixed blocks
      next_cb_boundary=scf_band_bound_l(1);
    } else {
      // pure short block
      next_cb_boundary=scf_band_bound_s(1)*3;
      cb_width = scf_band_bound_s(1);
      cb_begin = 0;
    } 
  } else {
    // long blocks: 0,1,3
    next_cb_boundary=scf_band_bound_l(1);
  }

  // Apply formula per block type
  for (sb = 0 ; sb < SBLIMIT ; sb++) {
    for (ss = 0 ; ss < SSLIMIT ; ss++) {
      boundary=(sb*18)+ss;
      if (boundary == next_cb_boundary) {
	// Adjust critical band boundary
	if (window_switching(ch,gr) && blocktype(ch,gr) == 2) {
	  if (mixedblock(ch,gr))  { // MIXED BLOCK
	    if (boundary == scf_band_bound_l(8)) {
	      next_cb_boundary=scf_band_bound_s(4)*3; 
	      cb = 3;
	      cb_width = scf_band_bound_s(cb+1) - 
		scf_band_bound_s(cb);
	      cb_begin = scf_band_bound_s(cb)*3;	 
	    } else if (boundary < scf_band_bound_l(8)){
	      next_cb_boundary = 
		scf_band_bound_l((++cb)+1);
	    } else {
	      next_cb_boundary 
		= scf_band_bound_s((++cb)+1)*3;
	      cb_width = scf_band_bound_s(cb+1) - 
		scf_band_bound_s(cb);
	      cb_begin = scf_band_bound_s(cb)*3;	 
	    }   
	  } else { // PURE SHORT BLOCKS
	    next_cb_boundary =  
	      scf_band_bound_s((++cb)+1)*3;
	    cb_width = scf_band_bound_s(cb+1) - 
	      scf_band_bound_s(cb);
	    cb_begin = scf_band_bound_s(cb)*3;
	  } 
		    
	} else { // LONG BLOCKS
	  next_cb_boundary = scf_band_bound_l((++cb)+1);
	}
      }

      // Use overall (global) scaling
      xr[gr][ch][sb][ss] = globalgain;

      // Do long/short dependent scaling operations
      if (window_switching(ch,gr) && 
	  (((blocktype(ch,gr) == 2) && 
	    (!mixedblock(ch,gr))) ||
	   ((blocktype(ch,gr) == 2) &&
	    mixedblock(ch,gr) && (sb >= 2)) )) {
		
	xr[gr][ch][sb][ss] *= pow(2.0, 0.25 * -8.0 * 
				  subblock_gain(ch, gr, 
						(boundary - cb_begin)/
						cb_width));
		
	xr[gr][ch][sb][ss] *= pow(2.0, scfscale *  
				  scalefac[gr][ch].s[(boundary - cb_begin)
						    /cb_width][cb]);
		
      } else { // LONG block types 0,1,3 & 
	// 1st 2 subbands of switched blocks
		
	xr[gr][ch][sb][ss] *= pow(2.0, scfscale *
				  (scalefac[gr][ch].l[cb] + 
				   preflag(ch,gr) * pretab[cb]));
      }
	    
      // Scale quantized value
      xr[gr][ch][sb][ss] *= pow(abs((int)is[gr][ch][sb][ss]), 4.0/3.0);
      if (is[gr][ch][sb][ss]<0) {
	xr[gr][ch][sb][ss] = -xr[gr][ch][sb][ss];
      }
      //	    printf("xl[%d,%d]=%f\n", sb, ss, xr[gr][ch][sb][ss]);
    }
  }

  return;
}


// restore the stereo values from "xr"
// result: "lr" data array
// note: only called for channel 0
CSAPI_MPEG void 
Layer3::decode_jstereo(int gr)
{
  // second bit in mode_ext() indicates ms_stereo
  const bool ms_stereo = (frame->mode() == jstereo) &&
    (frame->mode_ext() & 0x2);  
  // first bit indicates intensity_stereo
  const bool i_stereo  = (frame->mode() == jstereo) &&
    (frame->mode_ext() & 0x1);  

  char   is_pos[SBSSLIMIT]; 
  double is_ratio[SBSSLIMIT];
  double k[2][SBSSLIMIT];
  const int ch = 0;
  int i, j, ss, sb;
  int lines;
  int sfb, sfbcnt, max_sfb;
  const static double pi12 = PI/12;

  // Intialization
  memset ((void *)&is_pos[0], 7, SBSSLIMIT*sizeof(char));
  /* original code
  for (i = 0; i < SBSSLIMIT; i++) {
    is_pos[i] = 7;
  }
  */
    
  // intensity stereo processing
  const int s_ch = frame->channels();
  const int s_ver = frame->version();
  if ((s_ch == 2) && i_stereo ) {
    if (window_switching(ch,gr) && blocktype(ch,gr) == 2) {
      if(mixedblock(ch,gr)) {
	max_sfb = 0;
		
	for (j=0; j<3; j++ ) {
	  sfbcnt = 2;
	  for (sfb=12; sfb >=3; sfb--) {
	    lines= scf_band_bound_s(sfb+1) -
	      scf_band_bound_s(sfb);
	    i = 3*scf_band_bound_s(sfb) + (j+1) * lines - 1;
	    while (lines > 0) {
	      if (xr[gr][ch+1][i/SSLIMIT][i%SSLIMIT] != 0.0) {
		sfbcnt = sfb;
		sfb = -10;
		lines = -10;
	      }
	      lines--;
	      i--;
	    }
	  }
	  sfb = sfbcnt + 1;
		    
	  if (sfb > max_sfb) max_sfb = sfb;
		    
	  while (sfb<12) {
	    sb = scf_band_bound_s(sfb+1) -
	      scf_band_bound_s(sfb);
	    i = 3*scf_band_bound_s(sfb) + j * sb;
	    for ( ; sb > 0; sb--) {
	      is_pos[i] = scalefac[gr][ch+1].s[j][sfb];
	      if (is_pos[i] != 7)
		if (s_ver==MPEG2) {
		  i_stereo_k_values(ch,
				    gr,
				    is_pos[i],
				    i,
				    k);
		} else {
		  is_ratio[i] = tan((float)is_pos[i] * pi12);
		}
	      i++;
	    }
	    sfb++;
	  }
		    
	  sb = scf_band_bound_s(12)-scf_band_bound_s(11);
	  sfb = 3*scf_band_bound_s(11) + j * sb;
	  sb = scf_band_bound_s(13)-scf_band_bound_s(12);
		    
	  i = 3*scf_band_bound_s(11) + j * sb;
	  for ( ; sb > 0; sb--) {
	    is_pos[i] = is_pos[sfb];
	    is_ratio[i] = is_ratio[sfb];
	    k[0][i] = k[0][sfb];
	    k[1][i] = k[1][sfb];
	    i++;
	  }
	}
	if (max_sfb <= 3) {
	  i = 2;
	  ss = 17;
	  sb = -1;
	  while ( i >= 0 ) {
	    if ( xr[gr][ch+1][i][ss] != 0.0 ) {
	      sb = i*18+ss;
	      i = -1;
	    } else {
	      ss--;
	      if ( ss < 0 ) {
		i--;
		ss = 17;
	      }
	    }
	  }
	  i = 0;
	  while (((int)scf_band_bound_l(i)) <= sb) i++;
	  sfb = i;
	  i = scf_band_bound_l(i);
	  for ( ; sfb<8; sfb++) {
	    sb = scf_band_bound_l(sfb+1)-scf_band_bound_l(sfb);
	    for ( ; sb > 0; sb--) {
	      is_pos[i] = scalefac[gr][ch+1].l[sfb];
	      if (is_pos[i] != 7)
		if (s_ver == MPEG2) {
		  i_stereo_k_values(ch,
				    gr,
				    is_pos[i],
				    i,
				    k);
		} else {
		  is_ratio[i] = tan((float)is_pos[i] * pi12);
		}
	      i++;
	    }
	  }
	}
		
      } else { // SHORT BLOCKS
		
	for (j=0; j<3; j++) {
	  sfbcnt = -1;
	  for(sfb=12; sfb >=0; sfb--) {
	    lines = scf_band_bound_s(sfb+1) -
	      scf_band_bound_s(sfb);
	    i = 3*scf_band_bound_s(sfb) + (j+1) * lines - 1;
	    while (lines > 0) {
	      if (xr[gr][ch+1][i/SSLIMIT][i%SSLIMIT] != 0.0 ) {
		sfbcnt = sfb;
		sfb = -10;
		lines = -10;
	      }
	      lines--;
	      i--;
	    }
	  }
	  sfb = sfbcnt + 1;
	  while (sfb<12) {
	    sb = scf_band_bound_s(sfb+1)-scf_band_bound_s(sfb);
	    i = 3*scf_band_bound_s(sfb) + j * sb;
	    for ( ; sb > 0; sb--) {
	      is_pos[i] = scalefac[gr][ch+1].s[j][sfb];
	      if (is_pos[i] != 7)
		if(s_ver == MPEG2) {
		  i_stereo_k_values(ch,
				    gr,
				    is_pos[i],
				    i,
				    k);
		} else {
		  is_ratio[i] = tan((float)is_pos[i] * pi12);
		}
	      i++;
	    }
	    sfb++;
	  }
		    
	  sb = scf_band_bound_s(12)-scf_band_bound_s(11);
	  sfb = 3*scf_band_bound_s(11) + j * sb;
	  sb =  scf_band_bound_s(13)-scf_band_bound_s(12);
		    
	  i = 3*scf_band_bound_s(11) + j * sb;
	  for ( ; sb > 0; sb--) {
	    is_pos[i] = is_pos[sfb];
	    is_ratio[i] = is_ratio[sfb];
	    k[0][i] = k[0][sfb];
	    k[1][i] = k[1][sfb];
	    i++;
	  }
	}
      }
    } else { // LONG BLOCKS
      i = 31;
      ss = 17;
      sb = 0;
      while (i >= 0) {
	if ( xr[gr][ch+1][i][ss] != 0.0 ) {
	  sb = i*18+ss;
	  i = -1;
	} else {
	  ss--;
	  if ( ss < 0 ) {
	    i--;
	    ss = 17;
	  }
	}
      }
      i = 0;
      while (((int)scf_band_bound_l(i)) <= sb) i++;
      sfb = i;
      i = scf_band_bound_l(i);
      for ( ; sfb<21; sfb++) {
	sb = scf_band_bound_l(sfb+1) - scf_band_bound_l(sfb);
	for ( ; sb > 0; sb--) {
	  is_pos[i] = scalefac[gr][ch+1].l[sfb];
	  if (is_pos[i] != 7)
	    if(s_ver == MPEG2) {
	      i_stereo_k_values(ch,
				gr,
				is_pos[i],
				i,
				k);
	    } else {
	      is_ratio[i] = tan((float)is_pos[i] * pi12);
	    }
	  i++;
	}
      }
      sfb = scf_band_bound_l(20);
      for (sb = 576 - scf_band_bound_l(21); sb > 0; sb--) {
	is_pos[i] = is_pos[sfb];
	is_ratio[i] = is_ratio[sfb];
	k[0][i] = k[0][sfb];
	k[1][i] = k[1][sfb];
	i++;
      }
    }
  }
    

  // now write stereo values to lr
  if (s_ch == 2) {
    for(sb = 0; sb < SBLIMIT; sb++)
      for(ss = 0; ss < SSLIMIT; ss++) {
	i = (sb*18)+ss;
	if (is_pos[i] == 7) {
	  if (ms_stereo) {
	    lr[gr][0][sb][ss] = (xr[gr][0][sb][ss] +
				 xr[gr][1][sb][ss]) / 1.41421356;
	    lr[gr][1][sb][ss] = (xr[gr][0][sb][ss] -
				 xr[gr][1][sb][ss]) / 1.41421356;
	  } else {
	    lr[gr][0][sb][ss] = xr[gr][0][sb][ss];
	    lr[gr][1][sb][ss] = xr[gr][1][sb][ss];
	  }
	} else if (i_stereo) {
	  if (s_ver == MPEG2) {
	    lr[gr][0][sb][ss] = xr[gr][0][sb][ss] * k[0][i];
	    lr[gr][1][sb][ss] = xr[gr][0][sb][ss] * k[1][i]; 
	  } else {
	    lr[gr][0][sb][ss] = xr[gr][0][sb][ss] * 
	      (is_ratio[i]/(1+is_ratio[i]));
	    lr[gr][1][sb][ss] = xr[gr][0][sb][ss] * 
	      (1/(1+is_ratio[i])); 
	  }
	} else {
	  cerr << "MaaateP: Error in layer III joint stereo processing" << endl;
	}
      }

  } else {	// MONO: bypass xr[0][][] to lr[0][][]
	
    memcpy((void*)&lr[gr][0][0][0], (void*)&xr[gr][0][0][0], 
	   SBSSLIMIT*sizeof(double));
    /* original code
    for (sb = 0; sb < SBLIMIT; sb++)
      for(ss = 0; ss < SSLIMIT; ss++)
	lr[gr][0][sb][ss] = xr[gr][0][sb][ss];
    */
  }

  /* 
     for (ch=0; ch < frame->channels(); ch++) {
     for (sb = 0; sb < SBLIMIT; sb++)
     for(ss = 0; ss < SSLIMIT; ss++)
     printf("lr[%d,%d,%d]=%f\n", ch, sb, ss, lr[gr][ch][sb][ss]);
     }
  */

  return;
}


// MPEG2: calculation of k-values for intensity stereo
CSAPI_MPEG void 
Layer3::i_stereo_k_values(int   ch,
			  int   gr,
			  char  is_pos,
			  int   i,
			  double k[2][SBSSLIMIT])
{
  double io;

  if ((scalefac_compress(ch,gr) % 2) == 1) {
    io = 0.707106781188;
  } else {
    io = 0.840896415256;
  }

  if (is_pos == 0) { 
    k[0][i] = 1.0;
    k[1][i] = 1.0;
  } else if ((is_pos % 2) == 1) {
    k[0][i] = pow(io, (float)(is_pos + 1)/2.0);
    k[1][i] = 1.0;
  } else {
    k[0][i] = 1.0;
    k[1][i] = pow(io, (float)is_pos/2.0);
  }
  return;
}


// reorder the data in "lr"
// result: "ro" data array
CSAPI_MPEG void 
Layer3::reorder (int gr, int ch)
{
  int		sfb, sfb_start, sfb_lines;
  int		window, freq, src_line, des_line;

  if (window_switching(ch,gr) && blocktype(ch,gr) == 2) {
    if (mixedblock(ch,gr)) {
      // No reorder for low 2 subbands
      memcpy((void*)&ro[gr][ch][0][0], (void *)&lr[gr][ch][0][0],
	     sizeof(double)*2*SSLIMIT);
      
      /* original code
      for (sb=0 ; sb < 2 ; sb++) {
	for (ss=0 ; ss < SSLIMIT ; ss++) {
	  ro[gr][ch][sb][ss] = lr[gr][ch][sb][ss];
	}
      }
      */
	    
      // Reordering for rest switched short
      sfb_start = scf_band_bound_s(3);
      sfb_lines=scf_band_bound_s(4) - sfb_start; 
      sfb = 3;

    } else { // pure short: start from sfb = 0
      sfb_start=0;
      sfb_lines=scf_band_bound_s(1);
      sfb = 0;
    }

    for (; sfb < 13; sfb++) {
      for (window=0; window<3; window++)
	for(freq=0;freq<sfb_lines;freq++) {
	  src_line = sfb_start*3 + window*sfb_lines + freq; 
	  des_line = (sfb_start*3) + window + (freq*3);
	  ro[gr][ch][des_line/SSLIMIT][des_line%SSLIMIT]
	    = lr[gr][ch][src_line/SSLIMIT][src_line%SSLIMIT];
	}
      sfb_start=scf_band_bound_s(sfb);
      sfb_lines=scf_band_bound_s(sfb+1) - sfb_start;
    }
    
  } else {	// long blocks

    // copy array lr to ro
    memcpy((void*)&ro[gr][ch][0][0], (void *)&lr[gr][ch][0][0],
	   sizeof(double)*SBSSLIMIT);

    /* original code
       for (sb=0 ; sb < SBLIMIT ; sb++)
       for (ss=0 ; ss < SSLIMIT ; ss++) 
       ro[gr][ch][sb][ss] = lr[gr][ch][sb][ss];
    */

  }

  /* 
     for (sb=0 ; sb < SBLIMIT ; sb++)
     for (ss=0 ; ss < SSLIMIT ; ss++) 
     printf("ro[%d,%d]=%f\n",sb, ss, ro[gr][ch][sb][ss]);
  */

  return;
}


// perform antialiasing on sample data in "ro"
// result: "hybridIn" data array
CSAPI_MPEG void 
Layer3::antialias(int gr, int ch)
{
  static int	   init = 1;
  static double  ca[8], cs[8];
  double	   bu, bd;	// upper and lower butterfly inputs
  int 	   sblim, sb, ss;


  // initialize once
  if (init == 1) {
    int	i;
    double	sq;
    for (i = 0; i < 8; i++) {
      sq = sqrt(1.0 + Ci[i] * Ci[i]);
      cs[i] = 1.0 / sq;
      ca[i] = Ci[i] / sq;
    }
    init = 0;
  }
   
  // copy all inputs 
  memcpy((void*)&hybridIn[gr][ch][0][0], (void *)&ro[gr][ch][0][0],
	 sizeof(double)*SBSSLIMIT);

  /* original code
     for(int sb=0; sb<SBLIMIT; sb++) {
     for(int ss=0; ss<SSLIMIT; ss++) {
     hybridIn[gr][ch][sb][ss] = ro[gr][ch][sb][ss];
     }
     }
  */


  // return if no further work necessary
  if  (window_switching(ch,gr) && blocktype(ch,gr) == 2 &&
       !mixedblock(ch,gr)) {
    return;
  }

  if (window_switching(ch,gr) && blocktype(ch,gr) == 2 && 
      mixedblock(ch,gr)) {
    sblim = 1;
  } else {
    sblim = SBLIMIT-1;
  }

  // perform 31 alias-reduction operations between each pair of sub-bands
  // with 8 butterflies between each pair
  for (sb=0; sb < sblim; sb++) {
    for (ss = 0; ss < 8; ss++) {
      bu = ro[gr][ch][sb][17-ss];
      bd = ro[gr][ch][sb+1][ss];
      hybridIn[gr][ch][sb][17-ss] = (bu * cs[ss]) - (bd * ca[ss]);
      hybridIn[gr][ch][sb+1][ss] = (bd * cs[ss]) + (bu * ca[ss]);
    }  
  }

  /*
    for(int sb=0; sb<SBLIMIT; sb++) {
    for(int ss=0; ss<SSLIMIT; ss++) {
    printf("hybridIn[%d,%d]=%f\n", sb, ss, hybridIn[gr][ch][sb][ss]);
    }
    }
  */

  return;
}


// calculate hybrid synthesis filter on hybridIn samples
// result: "hybridOut" data array
CSAPI_MPEG void 
Layer3::hybrid(int gr, int ch)
{
  int	        ss, sb;
  int           bt;
  double        rawout[36];
  static double prevblck[2][SBLIMIT][SSLIMIT];
  static int	init = 1;

  // initialize once
  if (init == 1) {
    memset((void*)&prevblck[0][0][0], 0, 2*SBSSLIMIT*sizeof(double));

    /* original code
    int i, j, k;
    for (i = 0; i < 2; i++)
      for (j = 0; j < SBLIMIT; j++)
	for (k = 0; k < SSLIMIT; k++)
	  prevblck[i][j][k] = 0.0;
    */
    init = 0;
  }

  // go over subbands
  for (sb=0; sb<SBLIMIT; sb++) {

    bt = (window_switching(ch, gr) && mixedblock(ch,gr) &&
	  (sb < 2)) ? 0 : blocktype(ch,gr); 

    inv_mdct(hybridIn[gr][ch][sb], rawout, bt);

    // Overlap addition
    for (ss = 0; ss < SSLIMIT; ss++) {
      hybridOut[gr][ch][ss][sb] = rawout[ss] + prevblck[ch][sb][ss];
      prevblck[ch][sb][ss] = rawout[ss+18];

      //	    printf("hybridOut[%d,%d]=%f\n", sb, ss, hybridOut[gr][ch][ss][sb]);
      //	    printf("prevblck[%d,%d,%d]=%f\n", ch,sb, ss, prevblck[ch][sb][ss]);

    }
  }

  return;
}


// Calculation of the inverse MDCT
// In the case of short blocks the 3 output vectors are already
// overlapped and added in this module
CSAPI_MPEG void 
Layer3::inv_mdct(double in[18], double	out[36], int block_type)
{
  int		  i, m, p;
  static double   win[4][36];
  static int	  init = 1;
  double	  sum;
  double	  tmp[12];
  static double   COS[144]; // 4*36
  const static double pi12 = PI/12;
  const static double pi36 = PI/36;
  const static double pi24 = PI/24;
  const static double pi72 = PI/72;

  // initialze once
  if (init == 1) {
    // Type 0
    for (i = 0; i < 36; i++)
      win[0][i] = sin(pi36 * (i + 0.5));

    // Type 1
    for (i = 0; i < 18; i++)
      win[1][i] = sin(pi36 * (i + 0.5));
    for (i = 18; i < 24; i++)
      win[1][i] = 1.0;
    for (i = 24; i < 30; i++)
      win[1][i] = sin(pi12 * (i + 0.5 - 18));
    for (i = 30; i < 36; i++)
      win[1][i] = 0.0;
	
    // Type 3
    for (i = 0; i < 6; i++)
      win[3][i] = 0.0;
    for (i = 6; i < 12; i++)
      win[3][i] = sin(pi12 * (i + 0.5 - 6));
    for (i = 12; i < 18; i++)
      win[3][i] = 1.0;
    for (i = 18; i < 36; i++)
      win[3][i] = sin(pi36 * (i + 0.5));
	
    // Type 2
    for (i = 0; i < 12; i++)
      win[2][i] = sin(pi12 * (i + 0.5)) ;
    for (i = 12; i < 36; i++)
      win[2][i] = 0.0 ;
	
    for (i = 0; i < 4*36; i++)
      COS[i] = cos(pi72 * i);
	
    init = 0;
  }

  // initialize output samples
  memset((void*)&out[0], 0, 36*sizeof(double));
  /* original code
  for (i = 0; i < 36; i++) out[i]=0;
  */

  if (block_type == 2){
    for (i=0; i<3; i++){
      for (p = 0; p<12; p++){
	sum = 0.0;
	for(m=0; m<6; m++)
	  sum += in[i+3*m] * cos( pi24*(2*p+7)*(2*m+1) );
	tmp[p] = sum * win[block_type][p] ;
      }
      for (p=0; p<12; p++)
	out[6*i+p+6] += tmp[p];
    }
  } else {
    for (p=0; p<36; p++){
      sum = 0.0;
      for (m=0; m < 18; m++)
	sum += in[m] * COS[((2*p+19)*(2*m+1))%(144)];
      out[p] = sum * win[block_type][p];
    }
  }

  return;
}


// polyphase filterbank synthesis on hybridOut samples
// result: pcm samples
CSAPI_MPEG void 
Layer3::polyphase(int gr, int ch)
{
  int ss, sb;
  
  for (ss = 0; ss < SSLIMIT; ss++) {
    for (sb = 0; sb < SBLIMIT; sb++) {
      // Frequency inversion for polyphase
      if ((ss%2) && (sb%2)) {
	hybridOut[gr][ch][ss][sb] *= -1;
      }
    }

    // perform subband synthesis
    subband_syn(&(hybridOut[gr][ch][ss][0]), 
		ch, 
		&(pcm[gr][ch][ss][0]));
  }
  
  return;
}
  

// returns scalefactors for a sub-subband
CSAPI_MPEG float 
Layer3::scalefactor (unsigned int ch, unsigned int ss, unsigned int granule)
{
  int scfg=0;
  int gr=granule;
  short which;
  float exponent;

  // compute scalefactor scaling for this channel and granule
  const double scfscale = -0.5 * (1.0 + scalefac_scale(ch,gr));

  if (ss >= 576) {
    cerr << "MaaateP: Error: sub-subbands go from 0-575" << endl;
    return 0.0;
  }

  // compute scalefactor group & scalefactor from sub-subband (0-576)
  // Choose correct scalefactor group dependent on window type
  if (window_switching(ch,gr) && blocktype(ch,gr) == 2) {
    if (mixedblock(ch,gr)) {
      // mixed windows
      if (ss < 36) {
	// lower two subbands (= 2 x 18 subsubbands)
	while (ss >= scf_band_bound_l(scfg+1)) {
	  scfg++;
	}
	exponent = scfscale * (scalefac[gr][ch].l[scfg] +
			       preflag(ch,granule) * pretab[scfg]);
      } else {
	scfg=3;
	// upper 30 subbands (i.e. above subsubband 3x12)
	while (ss >= scf_band_bound_s(scfg+1)*3) {
	  scfg++;
	}
	which = ss % 3;
	exponent = scfscale * scalefac[gr][ch].s[which][scfg];
      }
    } else {
      // pure short windows
      while (ss >= scf_band_bound_s(scfg+1)*3) {
	scfg++;
      }
      which = ss % 3;
      exponent = scfscale * scalefac[gr][ch].s[which][scfg];
    } 
  } else {
    // long blocks: 0,1,3
    while (ss >= scf_band_bound_l(scfg+1)) {
      scfg++;
    }
    // check scfsi
    if (scfsi_group(ch,scfg)) {
      // scale factors for second granule same as for first granule
      gr=0;
    }    
    // calculate scalefactor
    exponent = scfscale * (scalefac[gr][ch].l[scfg] +
			   preflag(ch,granule) * pretab[scfg]);
  }

  if (exponent == 0.0) {
    // it would really be 1.0, but gives unjustified weighting when summed
    // as scalefactor of Layer III us being used similar to Layers I+II
    // return 0.0;
    // return 1.0;
    return 0.0;
  } else {
    return pow (2.0f, exponent);
  }
}

// returns boundary of scale factor band for long blocks
CSAPI_MPEG unsigned int 
Layer3::scf_band_bound_l (unsigned int subbandindex)
{
  if (subbandindex < 23) {
    if (frame->version() == MPEG1) {
      return sfBandIndex[frame->samplingrate_index()].l[subbandindex];
    } else {
      return sfBandIndex[frame->samplingrate_index()+3].l[subbandindex];
    }
  } else {
    cerr << "MaaateP: ERROR: unknown subband index (l):" << subbandindex <<\
      endl;
    return 0;
  }
}
 
// returns boundary of scale factor band for short blocks
CSAPI_MPEG unsigned int 
Layer3::scf_band_bound_s (unsigned int subbandindex)
{
  if (subbandindex < 14) {
    if (frame->version() == MPEG1) {
      return sfBandIndex[frame->samplingrate_index()].s[subbandindex];
    } else {
      return sfBandIndex[frame->samplingrate_index()+3].s[subbandindex];
    }
  } else {
    cerr << "MaaateP: ERROR: unknown subband index (s):" << subbandindex <<\
      endl;
    return 0;
  }
}
 
// returns bitallocation for scalefactors for lower frequency part
CSAPI_MPEG unsigned int 
Layer3::slen1 (unsigned int channel, unsigned int granule)
{
  return slen[0][si.ch[channel].gr[granule].scalefac_compress];
}

// returns bitallocation for scalefactors for higher frequency part
CSAPI_MPEG unsigned int 
Layer3::slen2 (unsigned int channel, unsigned int granule)
{
  return slen[1][si.ch[channel].gr[granule].scalefac_compress];
}

// returns scfsi for scalefactor goup (long windows only)
// class 0: groups 0-5 for long windows
// class 1: groups 6-10 for long windows
// class 2: groups 11-15 for long windows
// class 3: groups 16-20 for long windows
CSAPI_MPEG unsigned int 
Layer3::scfsi_group (unsigned int channel, unsigned int scfg)
{
  int scf_class=0;
  while (scfg >= sfbtable[0][scf_class+1]) scf_class++;
  return si.ch[channel].scfsi[scf_class];
}
 
// return amount of scalefactor bands for the granule
CSAPI_MPEG unsigned int 
Layer3::scf_bands(unsigned int channel, unsigned int granule)
{
  if (blocktype(channel,granule)==2 && !mixedblock(channel,granule)) {
    return 36; //3*12
  } else if (blocktype(channel,granule)==2) {
    return 35; //8+9*3
  } else { // blocktype = 0/1/3
    return 21;
  }
}

// sets the starting index of the buffer to the start of main_data
// skipping any anxilliary data from the previous main_data
CSAPI_MPEG bool 
Layer3::setbufstart()
{
  unsigned int bytesinbuf=0; // no of byte in interim buffer
  if (bufend != bufstart) {
    // bytes available in interim buffer
    if (bufend < bufstart) {
      bytesinbuf = (bufend+MAX_INTER_BUFSIZE)-bufstart;
    } else {
      bytesinbuf = bufend-bufstart;
    }
  } else {
    if (si.main_data_begin > 0) {
      cerr << "MaaateP: no data available in interim buffer" << endl;
      cerr << "         Cannot rewind " << si.main_data_begin << " bits." << endl;
      return false;
    } else {
      return true;
    }
  }

  if (si.main_data_begin > bytesinbuf) {
    cerr << "MaaateP: not enough main data available in interim buffer" << endl;
    cerr << "         Cannot rewind " << si.main_data_begin - bytesinbuf << " bits." << endl;
    return false;
  }
  bufstart += bytesinbuf-si.main_data_begin;
  bufstart %= MAX_INTER_BUFSIZE;
  whichbit = 0;
  return true;
}
    
// saves the remaining data of the frame to the interim buffer
// (should get implemented more efficiently!)
CSAPI_MPEG bool 
Layer3::savetointerbuffer()
{
  // in case of buffer overflow: count how often
  unsigned int overflow=0;

  // determine the byte from which to copy from frame's buffer
  unsigned int whichbyte = frame->bitsread / 8;
  //    if (frame->bitsread%8 > 0) whichbyte++;

  // copy until framesize is reached
  const unsigned int s_fsize = frame->framesize();
  while (whichbyte < s_fsize) {

#ifndef WORDS_BIGENDIAN
    inter_buffer[bufend] = ((unsigned char*) &(frame->buffer[whichbyte/4]))
      [3-(whichbyte%4)];
#else
    inter_buffer[bufend] = ((unsigned char*)frame->buffer)[whichbyte];
#endif
    whichbyte++;
    // next place to store
    bufend = (bufend+1)%MAX_INTER_BUFSIZE;
    if (bufend == bufstart) {
      // buffer overflow
      overflow++;
      bufstart++;
    }
  }
    
  if (overflow > 0)
    cerr << "MaaateP: WARNING: buffer overflow in interim buffer "
	 << "(Threw away " << overflow << " old bytes.)"
	 << endl;

  return true;
}

// resets the interim buffer and its variables
CSAPI_MPEG bool 
Layer3::clearinterbuffer()
{
  bufstart = bufend = whichbit = 0;

#ifdef DEBUG
  noReadBits = 0;
#endif

  return true;
}

// reads bits from appropriate buffer
CSAPI_MPEG unsigned int 
Layer3::readbits(unsigned int nobits)
{
  if (nobits == 0) {
    return 0;
  }

  // read from interim buffer as long as there's data there
  static unsigned int bitmask[9] = {
    0,  // dummy
    0x01, 0x03, 0x07, 0x0F,
    0x1F, 0x3F, 0x7F, 0xFF,
  };
    
  unsigned int val = 0;
  unsigned int bitsread=0;
    
  while (bitsread < nobits) {
    unsigned int toread = nobits-bitsread;
    if (bufstart == bufend) {
      // no more bits in interim buffer -> read from frame buffer
      val |= frame->readbitsfrombuffer(toread);
    } else {
      unsigned int bitavail=8-whichbit;
      if (toread < bitavail) {
	// all bits contained in inter_buffer[bufstart]
	val |= ((inter_buffer[bufstart] >> (bitavail-toread)) 
		& bitmask[toread]);
	whichbit += toread;
      } else {
	// read rest of inter_buffer[bufstart]
	val |= (inter_buffer[bufstart] & bitmask[bitavail]) 
	  << (toread-bitavail);
	whichbit=0;
	bufstart = (bufstart+1) % MAX_INTER_BUFSIZE;
      }
      toread=min(bitavail,toread);
    }
    bitsread += toread;
  }
  main_data_bitsread += nobits;
    
  return val;
}


// returns scfsi for channel and scalefactor class
// this tells whether the scalefactors for granule 0 are also used for gr 1
CSAPI_MPEG int 
Layer3::scfsi (unsigned int channel, unsigned int scfclass)
{
    return si.ch[channel].scfsi[scfclass];
}

// returns the samples of the 576 frequency lines (decoded huffman data)
CSAPI_MPEG int 
Layer3::sample (unsigned int ch, 
		unsigned int sb, 
		unsigned int ss,
		unsigned int gr)
{
    return is[gr][ch][sb][ss];
}

// returns the decoded subband samples
CSAPI_MPEG double 
Layer3::restored_sample (unsigned int channel, unsigned int subband, 
			 unsigned int subsubband, unsigned int granule)
{
  // printf("restored_sample(ch=%d,sb=%d,ss=%d,gr=%d)\n", channel, subband, subsubband, granule);
    return hybridOut[granule][channel][subsubband][subband];
}

// returns the decoded pcm samples
CSAPI_MPEG short 
Layer3::pcm_sample (unsigned int channel, unsigned int granule, 
		    unsigned int subband, unsigned int subsubband)
{
  //  printf("channel %d, granule=%d, subband=%d, subsubband=%d\n", channel,granule,subband,subsubband);
    return pcm[granule][channel][subsubband][subband];
}

CSAPI_MPEG double
Layer3::mdct_sample (unsigned int ch, unsigned int gr, unsigned int ssb) {
  // calculate sb and ss from ssb
  unsigned int sb, ss;
  sb = ssb / SSLIMIT;
  ss = ssb % SSLIMIT;
  return lr[gr][ch][sb][ss];
}

// return main_data_begin offset
CSAPI_MPEG unsigned int 
Layer3::main_data_begin()
{
    return si.main_data_begin;
}

// return private bits
CSAPI_MPEG unsigned int 
Layer3::private_bits()
{
    return si.private_bits;
}

// return length of part2 & 3 (scalefactors+huffman data)
CSAPI_MPEG unsigned int 
Layer3::part2_3_length(unsigned int channel, unsigned int granule)
{
    return si.ch[channel].gr[granule].part2_3_length;
}

// big values region
// contains no of sample pairs encoded from lower part of spectrum
CSAPI_MPEG unsigned int 
Layer3::big_values(unsigned int channel, unsigned int granule)
{
    return si.ch[channel].gr[granule].big_values;
}

// count1 values region
// contains no of sample quadruples encoded from middle part of spectrum
CSAPI_MPEG unsigned int 
Layer3::count1_values(unsigned int channel, unsigned int granule)
{
    return count1_vals[granule][channel];
}

// return global gain (= quantizer step size)
CSAPI_MPEG unsigned int 
Layer3::global_gain(unsigned int channel, unsigned int granule)
{
    return si.ch[channel].gr[granule].global_gain;
}

// return scalefac_compress which encodes biallocation for scalefactors
CSAPI_MPEG unsigned int 
Layer3::scalefac_compress(unsigned int channel, unsigned int granule)
{
    return si.ch[channel].gr[granule].scalefac_compress;
}

// returns 0 for long windows and 1 for transition, short or mixed windows
CSAPI_MPEG bool 
Layer3::window_switching(unsigned int channel, unsigned int granule)
{
    return (bool) (si.ch[channel].gr[granule].window_switching_flag);
}

// returns 1 for start of short windows, 2 for short windows, 3 for end of
// short windows; 0 otherwise
CSAPI_MPEG unsigned int 
Layer3::blocktype(unsigned int channel, unsigned int granule)
{
    return si.ch[channel].gr[granule].block_type;
}

// returns string for description of block type
CSAPI_MPEG const char*
Layer3::blocktype_str(unsigned int channel, unsigned int granule)
{
    if (si.ch[channel].gr[granule].window_switching_flag) {
	// window switching
	switch (si.ch[channel].gr[granule].block_type) {
	case 1:
	    if (mixedblock(channel,granule)) {
		return "start block (long block), mixed";
	    } else {
		return "start block (long block)";
	    }
	case 2:
	    if (mixedblock(channel,granule)) {
		return "short blocks, mixed";
	    } else {
		return "short blocks";
	    }
	case 3:
	    if (mixedblock(channel,granule)) {
		return "stop block (long block), mixed";
	    } else {
		return "stop block (long block)";
	    }
	default:
	    return "reserved";
	}
    } else {
	return "long block";
    }
}

// returns true if encoded block is mixed (i.e. two lowest subbands are
// transformed with long windows, upper 30 subbands with short windows);
// false otherwise
CSAPI_MPEG bool 
Layer3::mixedblock(unsigned int channel, unsigned int granule)
{
    return (bool) si.ch[channel].gr[granule].mixed_block_flag;
}

// returns which huffman table to select for big_values area
CSAPI_MPEG unsigned int 
Layer3::table_select(unsigned int channel, unsigned int granule, unsigned int area)
{
    return si.ch[channel].gr[granule].table_select[area];
}

// returns gain of area
CSAPI_MPEG unsigned int 
Layer3::subblock_gain(unsigned int channel, unsigned int granule, unsigned int area)
{
    return si.ch[channel].gr[granule].subblock_gain[area];
}

// returns region 0 boundary
CSAPI_MPEG unsigned int 
Layer3::region0_count(unsigned int channel, unsigned int granule)
{
    return si.ch[channel].gr[granule].region0_count;
}

// returns region 1 boundary
CSAPI_MPEG unsigned int 
Layer3::region1_count(unsigned int channel, unsigned int granule)
{
    return si.ch[channel].gr[granule].region1_count;
}

// returns no of samples in region0
CSAPI_MPEG unsigned int 
Layer3::region0_samps(unsigned int channel, unsigned int granule)
{
    if (big_values(channel,granule) > 0) {
	if (window_switching(channel,granule) && 
	    blocktype(channel,granule)==2) {
	    // pure short windows
	    return 36; // as scf_band_bound_s[9/3]=12 for all sampling rates
	} else {
	    return min (scf_band_bound_l (region0_count(channel,granule)+1),
			big_values(channel,granule)*2);
	}
    } else {
	return 0;
    }
}

// returns no of samples in region1
CSAPI_MPEG unsigned int 
Layer3::region1_samps(unsigned int channel, unsigned int granule)
{
    if (big_values(channel,granule) > 0) {
	if (window_switching(channel,granule) && 
	    blocktype(channel,granule)==2) {
	    // pure short windows, no region 2
	    return (2*big_values(channel,granule) - 36);
	} else {
	    return (min (scf_band_bound_l (region0_count(channel,granule) +
					   region1_count(channel,granule) + 2),
			 big_values(channel,granule)*2) -
			region0_samps(channel,granule));
	}
    } else {
	return 0;
    }
}

// returns no of samples in region2
CSAPI_MPEG unsigned int 
Layer3::region2_samps(unsigned int channel, unsigned int granule)
{
    if (big_values(channel,granule) > 0) {
	if (window_switching(channel,granule) && 
	    blocktype(channel,granule)==2) {
	    // pure short windows, no region 2
	    return 0;
	} else {
	    unsigned int region0_1_samps = region1_samps(channel,granule) +
		region0_samps(channel,granule);
	    if ((2*big_values(channel,granule)) > region0_1_samps) {
		return (2*big_values(channel,granule) - 
			region1_samps(channel,granule) -
			region0_samps(channel,granule));
	    } else {
		return 0;
	    }
	} 
    } else {
	return 0;
    }
}

// returns preemphasis flag
CSAPI_MPEG unsigned int 
Layer3::preflag(unsigned int channel, unsigned int granule)
{
    return si.ch[channel].gr[granule].preflag;
}

// returns scalefactor scaling (i.e. quantized with stepsize of 2 or sqrt(2))
CSAPI_MPEG unsigned int 
Layer3::scalefac_scale(unsigned int channel, unsigned int granule)
{
    return si.ch[channel].gr[granule].scalefac_scale;
}

// returns which huffman table to select for count1 area
CSAPI_MPEG unsigned int 
Layer3::count1table_select(unsigned int channel, unsigned int granule)
{
    return si.ch[channel].gr[granule].count1table_select;
}

// returns the length of the layer3-specific header in bytes
CSAPI_MPEG unsigned int 
Layer3::sideinfo_length()
{
    if (frame->version() == MPEG1) {
	if (frame->channels() == 1) {
	    return 17; // byte
	} else {
	    return 32; // byte
	}
    } else { // MPEG2
	if (frame->channels() == 1) {
	    return 9; // byte
	} else {
	    return 17; // byte
	}
    }	
}

// returns the no of granules
CSAPI_MPEG unsigned int 
Layer3::granules()
{
  return ((frame->version() == MPEG2) ? 1 : 2);
}

// returns the no of slots of main_data
CSAPI_MPEG unsigned int 
Layer3::main_data_slots()
{
    int nSlots = (int) ((144 * frame->bitrate()) / frame->samplingrate());

    if (frame->version() == MPEG2) {
	nSlots /= 2;
    }
    nSlots -= sideinfo_length();
    if (frame->padding()) {
        nSlots++;
    }
    nSlots -= 4; // deduct header
    if (frame->crcprotected()) {
        nSlots -= 2; // deduct checksum
    }
 
    return nSlots;
}
