/*
    MPEG Maaate: An Australian MPEG audio analysis toolkit
    Copyright (C) 2000 Commonwealth Scientific and Industrial Research Organisation
    (CSIRO), Australia.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "module.H"

#include <math.h>

#define ROUND(number) floor((number)+0.5)

static void 
init_SBvalue(Module * m) {
  
  // set up member values
  m->set_name (string("SBvalue"));
  m->set_desc (string("subband values over each subband only for channel 0"));
  m->set_author (string("CSIRO-MIS AAS Thomas VINCENT"));
  m->set_copyright (string("(c) 2002 CSIRO"));
  m->set_url (string("http://www.cmis.csiro.au/Maaate/docs/modules.html"));

  // set up input parameter list
  m->inputSpecs()->clear();  
  
  // first parameter: sound file     
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("soundfile"),
		     string("the SOUND file for which the subband "
			    "values gets extracted"),
		     MAAATE_TYPE_SOUNDFILE,
		     new ModuleParam((SOUNDfile*)NULL)) // default
     ); 
  
  // second parameter: start time
  MaaateConstraint * constraint = new MaaateConstraint();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("starttime"),
		     string("time instant from which to start the "
			    "subband values extraction"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.0), // default
		     constraint));
  
  // third parameter: end time
  constraint = new MaaateConstraint();
  constraint->clear();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("endtime"),
		     string("time instant until which to calculate "
			    "the subband values"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)DBL_MAX), // default
		     constraint));
  
  // fourth parameter: start subband
  constraint = new MaaateConstraint();
  constraint->clear();
  constraint->addConstraintGreaterThan(0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("start-subband"),
		     string("subband from which to start the subband "
			    "values extraction"),
		     MAAATE_TYPE_INT,
		     new ModuleParam(0), // default
		     constraint));

  // fifth parameter: end subband
  constraint = new MaaateConstraint();
  constraint->clear();
  constraint->addConstraintGreaterThan(0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("end-subband"),
		     string("subband at which to end the subband"
			    "values extraction"),
		     MAAATE_TYPE_INT,
		     new ModuleParam(0), // default
		     constraint));
  
  // set up output parameter list
  m->outputSpecs()->clear();
  
  // first parameter: subband energies curves
  m->outputSpecs()->push_back
    (ModuleParamSpec(string("subband values curves"),
		     string("subband values (frequency samples) for the "
			    "requested time interval for each subband"),
		     MAAATE_TYPE_SEGMENTDATA,
		     new ModuleParam((SegmentData*)NULL)) // default
     );
  
};


// function to suggest parameter values given that some of the input
// parameters have already been set; also changes constraints accordingly
static void
suggest_SBvalue (Module * m, list<ModuleParam> * paramsIn) {
  list<ModuleParam>::iterator iter;

  // get module's input parameter specifications
  list<ModuleParamSpec> * inSpecs = m->inputSpecs();
  list<ModuleParamSpec>::iterator iterSpecs;
  
  iter = paramsIn->begin(); // SOUND file
  iterSpecs = inSpecs->begin();
  if (iter == paramsIn->end()) return;
  SOUNDfile * mf = (*iter).get_sf();
  if (mf == NULL) return;
  
  ++iter; // start time
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0.0, mf->file_duration());
  double startTime = (*iter).get_r();

  ++iter; // end time
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0.0, mf->file_duration());
  double endTime = (*iter).get_r();

  //check time
  if (endTime < startTime) {
    endTime = startTime;
    (*iter).set(startTime);
  }

  ++iter; // start subband
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0, (mf->nb_subbands(HIGH) - 1) );
  int startSubband = (*iter).get_i();

  ++iter; // end subband
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0, (mf->nb_subbands(HIGH) - 1) );

  // check nr of subbands
  int endSubband = (*iter).get_i();
  if (endSubband < startSubband) {
    endSubband = startSubband;
    (*iter).set(startSubband);
  }

}

// function to work on input parameters and return output parameters
static list<ModuleParam> *
apply_SBvalue (Module * m, list<ModuleParam> * paramsIn) {

  // the output parameter list
  list<ModuleParam> * mpl = new list<ModuleParam>();
  
  // iterator in input parmeter list
  list<ModuleParam>::iterator iter;

  //Getting input parameters and last checking
  //SOUND file
  iter = paramsIn->begin();
  if (iter == paramsIn->end()) return mpl;
  SOUNDfile * mf = (*iter).get_sf();
  if (mf == NULL) return mpl;
  
  //start time
  ++iter;
  double startTime = (*iter).get_r();

  //end time
  ++iter;
  double endTime = (*iter).get_r();

  //check time
  if (endTime < startTime) {
    endTime = startTime;
  }

  // start subband
  ++iter;
  int startSubband = (*iter).get_i();
  
  //end subband
  ++iter;
  int endSubband = (*iter).get_i();

  //check subband
  if (endSubband < startSubband) {
      endSubband = startSubband;
  }
  //End getting and checking


  //convert time into window number
  long start = mf->time2window( (float)startTime);

  long end = mf->time2window( (float)endTime);

  // go to start window  within soundfile
  if (!mf->seek_window(start)) {
    cerr << "MaaateM: Error when positioning" << endl;
    cerr << "         startposition = 0.0" << endl;
    mf->seek_window(0);
    start = 0;
  }
  

  // analyse first window
  if (!mf->next_window(HIGH)) {
    cerr << "MaaateM: Warning: could not analyse first window." << endl;
    return mpl;
  }

  //colunms number
  long columns = end - start;
  columns = (columns > mf->file_window_number()) ? mf->file_window_number() : columns;
  columns *= mf->timeticks(HIGH);
  //rows number
  int nbsubband = endSubband - startSubband + 1;

  SegmentData *result = new SegmentData(startTime,
					endTime,
					columns,
					nbsubband);

  //extract subband rms
  while (mf->at_window() <= end) {

    //from one sample to another
    for (unsigned int i = 0; i < mf->timeticks(HIGH); i++) {
      //from one subband to another
      for (int sb = startSubband; sb <= endSubband; sb++) { 
      result->data[result->colFilled][sb-startSubband] =  mf->freq_value (0,sb,i,HIGH);
      }
      result->colFilled++;
    }

    // analyse next window
    if (!mf->next_window(HIGH)) {
      break;
    }
    
  }
  
#ifdef DEBUG    
  cout << "Columns filled:" << result->colFilled << endl;
  cout << "Sum:" << result->sum() << endl;
  cout << "Average:" << result->avg() << endl;
  cout << *result;             
#endif
  
  mpl->push_back(ModuleParam(result));

  return mpl;
    
}; 
            
             
Module
loadSBvalueModule (void)
{
  return Module(init_SBvalue,
		NULL,
		suggest_SBvalue,
		apply_SBvalue ,
		NULL,
		NULL);
}
