/*
    MPEG Maaate: An Australian MPEG audio analysis toolkit
    Copyright (C) 2000 Commonwealth Scientific and Industrial Research Organisation
    (CSIRO), Australia.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "module.H"

#include "tools.H"
#include <math.h>

#define ROUND(number) floor((number)+0.5)

static void 
init_bandnrj(Module * m) {
  
  // set up member values
  m->set_name (string("bandnrj"));
  m->set_desc (string("signal energy on one window for the selected subbands"));
  m->set_author (string("CSIRO-MIS AAS Thomas VINCENT"));
  m->set_copyright (string("(c) 2002 CSIRO"));
  m->set_url (string("http://www.cmis.csiro.au/Maaate/docs/modules.html"));

  // set up input parameter list
  m->inputSpecs()->clear();  
  
  // first parameter:  sound file     
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("soundfile"),
		     string("the SOUND file for which the band "
			    "energy gets calculated"),
		     MAAATE_TYPE_SOUNDFILE,
		     new ModuleParam((SOUNDfile*)NULL)) // default
     ); 
  
  // second parameter: start time
  MaaateConstraint * constraint = new MaaateConstraint();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("starttime"),
		     string("time instant from which to start the "
			    "band energy calculation"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.0), // default
		     constraint));
  
  // third parameter: end time
  constraint = new MaaateConstraint();
  constraint->clear();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("endtime"),
		     string("time instant until which to calculate "
			    "the band energy"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)DBL_MAX), // default
		     constraint));

  // fourth parameter: start subband
  constraint = new MaaateConstraint();
  constraint->clear();
  constraint->addConstraintGreaterThan(0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("start-subband"),
		     string("subband from which to start the band "
			    "energy calculation"),
		     MAAATE_TYPE_INT,
		     new ModuleParam(0), // default
		     constraint));

  // fifth parameter: end subband
  constraint = new MaaateConstraint();
  constraint->clear();
  constraint->addConstraintGreaterThan(0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("end-subband"),
		     string("subband at which to end the band "
			    "energy calculation"),
		     MAAATE_TYPE_INT,
		     new ModuleParam(0), // default
		     constraint));
 

  // sixth parameter: window function
  constraint = new MaaateConstraint();
  constraint->clear();
  constraint->addConstraintRange(0, 3);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("window-numero"),
		     string("choice the window function to apply"
			    "during the calculation"),
		     MAAATE_TYPE_INT,
		     new ModuleParam(0), // default
		     constraint));
  

  // set up output parameter list
  m->outputSpecs()->clear();
  
  // first parameter: band energy curves
  m->outputSpecs()->push_back
    (ModuleParamSpec(string("band energy curve"),
		     string("an approximation of perceptual loudness"),
		     MAAATE_TYPE_SEGMENTDATA,
		     new ModuleParam((SegmentData*)NULL)) // default
     );
  
};


// function to suggest parameter values given that some of the input
// parameters have already been set; also changes constraints accordingly
static void
suggest_bandnrj (Module * m, list<ModuleParam> * paramsIn) {
  list<ModuleParam>::iterator iter;

  // get module's input parameter specifications
  list<ModuleParamSpec> * inSpecs = m->inputSpecs();
  list<ModuleParamSpec>::iterator iterSpecs;
  
  iter = paramsIn->begin(); // sound file
  iterSpecs = inSpecs->begin();
  if (iter == paramsIn->end()) return;
  SOUNDfile * mf = (*iter).get_sf();
  if (mf == NULL) return;
  
  ++iter; // start time
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0.0, mf->file_duration());
  double startTime = (*iter).get_r();

  ++iter; // end time
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0.0, mf->file_duration());
  double endTime = (*iter).get_r();

  //check time
  if (endTime < startTime) {
    endTime = startTime;
    (*iter).set(startTime);
  }

  ++iter; // start subband
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0, mf->nb_subbands(LOW)-1);

  int startSubband = (*iter).get_i();

  ++iter; // end subband
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0, mf->nb_subbands(LOW)-1);

  // check nr of subbands
  int endSubband = (*iter).get_i();
  if (endSubband < startSubband) {
    endSubband = startSubband;
    (*iter).set(startSubband);
  }

  //nothing more to check for window number

}

// function to work on input parameters and return output parameters
static list<ModuleParam> *
apply_bandnrj (Module * m, list<ModuleParam> * paramsIn) {

  // the output parameter list
  list<ModuleParam> * mpl = new list<ModuleParam>();
  
  // iterator in input parmeter list
  list<ModuleParam>::iterator iter;

  //Getting input parameters and last checking
  //sound file
  iter = paramsIn->begin();
  if (iter == paramsIn->end()) return mpl;
  SOUNDfile * mf = (*iter).get_sf();
  if (mf == NULL) return mpl;
  
  //start time
  ++iter;
  double startTime = (*iter).get_r();

  //end time
  ++iter;
  double endTime = (*iter).get_r();

  //check time
  if (endTime < startTime) {
    endTime = startTime;
  }

  ++iter; // start subband
  int startSubband = (*iter).get_i();
  
  ++iter; // end subband
  int endSubband = (*iter).get_i();
  // check nr of subbands depending on layer
  if (endSubband < startSubband) {
    endSubband = startSubband;
  }

  //getting window numero
  ++iter;
  int window = (*iter).get_i();

  //End getting and checking
  
  //convert time into window nb
  long start = mf->time2window( (float)startTime);

  long end = mf->time2window( (float)endTime);

  // go to start window within soundfile
  if (!mf->seek_window(start)) {
    cerr << "MaaateM: Error when positioning" << endl;
    cerr << "         startposition = 0.0" << endl;
    mf->seek_window(0);
    start = 0;
    }

  
  // analyse first window
  if (!mf->next_window(LOW)) {
    cerr << "MaaateM: Warning: could not analyse first window." << endl;
    return mpl;
  }

  //col nb
  long columns = end - start;
  columns = (columns > mf->file_window_number()) ? mf->file_window_number() : columns;

  SegmentData *result = new SegmentData(startTime,
					endTime,
					columns,
					1);

  //set window function:
  double (* h) (unsigned int, int) = 0;
  switch (window) {
  case 0: //Square
    h = &square_window;
    break;
  case 1: //Hamming
    h = &hamming_window;
    break;
  case 2: //Welch
    h = &welch_window;
    break;
  case 3: //Bartlett
    h = &bartlett_window;
    break;
  default: 
    h = &square_window;
    break; //never used
  }

  // value of the window at one given time
  double hvalue;
  //temporary value of the sum
  double sum = 0.0;
  double temp;
  int Mmax; 

  while (mf->at_window() <= end) {
 
    // nb of samples in one window, it is also on how many samples
    //the signal energy is calculated
    Mmax = mf->timeticks(LOW);
 
    //sum in the time domain
    for (int m = 0; m < Mmax; m++) {
      //calculate the value of h(n-m) for this value of m, with n=Mmax-1
      //hvalue = square_window ( Mmax-1, (Mmax-1)-m);
      hvalue = (*h) (Mmax-1, (Mmax -1)-m);
      
      temp = 0.0;

      //sum in the frequency domain
      for (int sb = startSubband; sb < endSubband; sb++) {
	temp += pow(mf->freqvalue_st_mean(sb,m,LOW),2);
      }
      sum += hvalue * temp;
    }
    
    //final result
    result->data[result->colFilled++][0] = sum; 
    
    //init sum
    sum=0.0;

    
    // analyse next window
    if (!mf->next_window(LOW)) {
      break;
    }
    
  }
  
#ifdef DEBUG    
  cout << "Columns filled:" << result->colFilled << endl;
  cout << "Sum:" << result->sum() << endl;
  cout << "Average:" << result->avg() << endl;
  cout << *result;             
#endif
  
  mpl->push_back(ModuleParam(result));

  return mpl;
  
}; 
            
             
Module
loadbandnrjModule (void)
{
  return Module(init_bandnrj,
		NULL,
		suggest_bandnrj,
		apply_bandnrj ,
		NULL,
		NULL);
}
