/*
    MPEG Maaate: An Australian MPEG audio analysis toolkit
    Copyright (C) 2000 Commonwealth Scientific and Industrial Research Organisation
    (CSIRO), Australia.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "module.H"

#include "plugins.H"

// initialisation of the silence module
static void 
init_bgnoiselevel(Module * m) {
  // set up member values
  m->set_name(string("bgnoiselevel"));
  m->set_desc(string("calculates the level of background noise using "
		     "silence segmentation to find that threshold at "
		     "the first silence segmentation result"));
  m->set_author(string("CSIRO-MIS DMIS Silvia Pfeiffer"));
  m->set_copyright(string("(c) 2002 CSIRO"));
  m->set_url(string("http://www.cmis.csiro.au/Maaate/docs/modules.html"));

  // set up input parameter list
  m->inputSpecs()->clear();
  
  // first parameter: segment data
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("loudness"),
		     string("the loudness contour being stored in a "
			    "segment data structure, if available"),
		     MAAATE_TYPE_SEGMENTDATA,
		     new ModuleParam((SegmentData*)NULL)) // default
     ); 
  
  // second parameter: start time
  MaaateConstraint * constraint = new MaaateConstraint();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("starttime"),
		     string("time instant from which to start the "
			    "bgnoiselevel calculation"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.0), // default
		     constraint));
  
  // third parameter: end time
  constraint->clear();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("endtime"),
		     string("time instant until which to calculate "
			    "the bgnoiselevel"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)DBL_MAX), // default
		     constraint));
  
  // fourth parameter: minimum duration
  constraint->clear();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("minduration"),
		     string("minimum amount of time the loudness contour "
			    "must stay above the given threshold "
			    "to be considered as silent"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.1), // default 100ms
		     constraint));
  
  // fifth parameter: tolerated interruption
  constraint->clear();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("tolerated_interruption"),
		     string("maximum amount of time the loudness contour "
			    "may lie below the given threshold "
			    "without affecting the bgnoiselevel"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.01), // default 10ms
		     constraint));
  
  // sixth parameter: onset time
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("onsettime"),
		     string("time by which to delay the start of a bgnoiselevel"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.0)));

  // seventh parameter: release time
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("releasetime"),
		     string("time to shorten the end of a bgnoiselevel"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.0)));
  
  
  // set up output parameter list
  m->outputSpecs()->clear();
  
  // first output parameter: bgnoiselevel output
  m->outputSpecs()->push_back
    (ModuleParamSpec(string("bgnoiselevel"),
		     string("calculated background noise level"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam(0.0)) // default
     );
  
};


// function to suggest parameter values and set constraints given that
// some of the input parameters have already been set
static void
suggest_bgnoiselevel (Module * m, list<ModuleParam> * paramsIn) {
  list<ModuleParam>::iterator iter;

  // get module's input parameter specifications
  list<ModuleParamSpec> * inSpecs = m->inputSpecs();
  list<ModuleParamSpec>::iterator iterSpecs;

  iter = paramsIn->begin(); // SegmentData
  iterSpecs = inSpecs->begin();
  if (iter == paramsIn->end()) return;
  SegmentData * sd = (*iter).get_sd();
  // there has to be a curve cause it cannot be calculated
  if (sd == NULL) return;
  
  ++iter; // start time
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(sd->start(), sd->end());
  double startTime = (*iter).get_r();

  ++iter; // end time
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(sd->start(), sd->end());
  double endTime = (*iter).get_r();
  if (endTime < startTime) {
    endTime = startTime;
    (*iter).set(startTime);
  }

  ++iter; // minimum duration
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0.0, sd->duration());
  double minDuration = (*iter).get_r();

  ++iter; // maximum tolerated interruption
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0.0, sd->duration());
  
  ++iter; // onset time
  ++iterSpecs;
  if ((*iter).get_r() > minDuration) (*iter).set(minDuration);

  ++iter; // release time
  ++iterSpecs;
  if ((*iter).get_r() > minDuration) (*iter).set(minDuration);

};

// function to work on input parameters and return output parameters
static list<ModuleParam> *
apply_bgnoiselevel (Module * m, list<ModuleParam> * paramsIn) {
  // the output parameter list
  static list<ModuleParam> mpl;
  mpl.clear();

  list<ModuleParam>::iterator iter;

  // open required plugin library
  Plugins * plugins = new Plugins();
#if (defined (WIN32) || (_WIN32))
#ifdef _DEBUG
  string libName = "libMaaateMD.dll";
  if (!plugins->AddLibrary(libName)) {
    cerr << "Error loading library libMaaateMD.dll" << endl;
    exit (1);
    delete plugins;
    return &mpl;
  }
#else
  string libName = "libMaaateM.dll";
  if (!plugins->AddLibrary(libName)) {
    cerr << "Error loading library libMaaateM.dll" << endl;
    exit (1);
    delete plugins;
    return &mpl;
  }
#endif
#else
  if (!plugins->AddLibrary(string("libMaaateM.so"))) {
    cerr << "Error loading library libMaaateM.so" << endl;
    exit (1);
    delete plugins;
    return &mpl;
  }
#endif

  iter = paramsIn->begin(); // SegmentData
  if (iter == paramsIn->end()) return &mpl;
  SegmentData * sd = (*iter).get_sd();
  // if no sd given, return as there is no information available
  // on what file to calculate loudness approximation on
  if (sd == NULL) {
    return &mpl;
  }
  
  ++iter; // start time
  double startTime = (*iter).get_r();

  ++iter; // end time
  double endTime = (*iter).get_r();
  if (endTime < startTime) {
    endTime = startTime;
  }

  ++iter; // minimum duration
  double minDuration = (*iter).get_r();

  ++iter; // maximum tolerated interruption
  double maxInterruptTolerated = (*iter).get_r();
  
  ++iter; // onset time
  double onsetTime = (*iter).get_r();
  if (onsetTime > minDuration) onsetTime = minDuration;

  ++iter; // release time
  double releaseTime = (*iter).get_r();
  if (releaseTime > minDuration) releaseTime = minDuration;

  // calculate segmentation using segmentation module
  Module * sm = plugins->GetModule(string("segmentation"));


  // ------------------------------------------------------------
  // try silence segmentation at different thresholds until
  // a segmentation is found
  int   found=0;
  double thresh=0.0;
  while (found<1 && thresh <= 1.0) {
    // XXX: this could be implemented to give an exacter resolution
    // if required
    thresh += 0.01;

    // set parameters for segmentation module
    // XXX: this isn't quite optimal as it only requires changing of
    // one parameter
    list<ModuleParam> sm_paramsIn;
    sm_paramsIn.push_back(sd);        // loudness contour
    sm_paramsIn.push_back(startTime); 
    sm_paramsIn.push_back(endTime);
    sm_paramsIn.push_back(true);      // below threshold
    sm_paramsIn.push_back(thresh);
    sm_paramsIn.push_back(minDuration);
    sm_paramsIn.push_back(maxInterruptTolerated);
    sm_paramsIn.push_back(onsetTime);
    sm_paramsIn.push_back(releaseTime);

    // suggest and check parameters
    /* xxx */
    sm->suggestValues(&sm_paramsIn);
    if (sm_paramsIn.size() == 0) {
      delete plugins;
      return &mpl;
    }
    
    // execute module
    list<ModuleParam> * paramsOut = sm->apply(&sm_paramsIn);
    if (paramsOut->size() == 0) {
      delete plugins;
      return &mpl;
    }

    // get resulting segmentation
    SegmentTable *result = paramsOut->front().get_st();

    found = result->size();
    delete result;
    delete paramsOut;
  }
    
  mpl.push_back(ModuleParam(thresh));
  delete plugins;
  return &mpl;
  
};


Module
loadBgNoiseLevelModule (void)
{
  // init_bgnoiselevel gets executed during construction of module
  return Module(init_bgnoiselevel,
		NULL,
		suggest_bgnoiselevel,
		apply_bgnoiselevel,
		NULL,
		NULL);
}
