/*
    MPEG Maaate: An Australian MPEG audio analysis toolkit
    Copyright (C) 2000 Commonwealth Scientific and Industrial Research Organisation
    (CSIRO), Australia.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "module.H"

#include <math.h>

#define ROUND(number) floor((number)+0.5)

static void 
init_centralmoment(Module * m) {
  
  // set up member values
  m->set_name (string("centralMoment"));
  m->set_desc (string("subband central moments"));
  m->set_author (string("CSIRO-MIS AAS Thomas VINCENT"));
  m->set_copyright (string("(c) 2002 CSIRO"));
  m->set_url (string("http://www.cmis.csiro.au/Maaate/docs/modules.html"));

  // set up input parameter list
  m->inputSpecs()->clear();  
  
  // first parameter: sound file     
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("soundfile"),
		     string("the SOUND file for which the central "
			    "moments gets calculated"),
		     MAAATE_TYPE_SOUNDFILE,
		     new ModuleParam((SOUNDfile*)NULL)) // default
     ); 
  
  // second parameter: start time
  MaaateConstraint * constraint = new MaaateConstraint();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("starttime"),
		     string("time instant from which to start the "
			    "central moments calculation"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.0), // default
		     constraint));
  
  // third parameter: end time
  constraint = new MaaateConstraint();
  constraint->clear();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("endtime"),
		     string("time instant until which to calculate "
			    "the central moments"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)DBL_MAX), // default
		     constraint));
  
  // fourth parameter: start subband
  constraint = new MaaateConstraint();
  constraint->clear();
  constraint->addConstraintGreaterThan(0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("start-subband"),
		     string("subband from which to start the subband "
			    "energy calculation"),
		     MAAATE_TYPE_INT,
		     new ModuleParam(0), // default
		     constraint));

  // fifth parameter: end subband
  constraint = new MaaateConstraint();
  constraint->clear();
  constraint->addConstraintGreaterThan(0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("end-subband"),
		     string("subband at which to end the subband"
			    "energy calculation"),
		     MAAATE_TYPE_INT,
		     new ModuleParam(0), // default
		     constraint));
  

  // sixth parameter: duration
  constraint = new MaaateConstraint();
  constraint->clear();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("duration"),
		     string("duration in seconds over which to compute"
			    "the central moments"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.5), // default
		     constraint));


  // seventh parameter: k
  constraint = new MaaateConstraint();
  constraint->clear();
  constraint->addConstraintRange(2,5);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("k"),
		     string("value of k, the pow to use"
			    "to compute the central moments"),
		     MAAATE_TYPE_INT,
		     new ModuleParam(2), // default
		     constraint));
  

  // set up output parameter list
  m->outputSpecs()->clear();
  
  // first parameter: central moments curves
  m->outputSpecs()->push_back
    (ModuleParamSpec(string("central moments curves"),
		     string("capture how much a subband`s energy is dispersed from its mean"),
		     MAAATE_TYPE_SEGMENTDATA,
		     new ModuleParam((SegmentData*)NULL)) // default
     );
  
};


// function to suggest parameter values given that some of the input
// parameters have already been set; also changes constraints accordingly
static void
suggest_centralmoment (Module * m, list<ModuleParam> * paramsIn) {
  list<ModuleParam>::iterator iter;

  // get module's input parameter specifications
  list<ModuleParamSpec> * inSpecs = m->inputSpecs();
  list<ModuleParamSpec>::iterator iterSpecs;
  
  iter = paramsIn->begin(); // SOUND file
  iterSpecs = inSpecs->begin();
  if (iter == paramsIn->end()) return;
  SOUNDfile * mf = (*iter).get_sf();
  if (mf == NULL) return;
  
  ++iter; // start time
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0.0, mf->file_duration());
  double startTime = (*iter).get_r();

  ++iter; // end time
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0.0, mf->file_duration());
  double endTime = (*iter).get_r();

  //check time
  if (endTime < startTime) {
    endTime = startTime;
    (*iter).set(startTime);
  }

  ++iter; // start subband
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0, mf->nb_subbands(HIGH)-1);
  int startSubband = (*iter).get_i();

  ++iter; // end subband
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0, mf->nb_subbands(HIGH)-1);

  // check nr of subbands
  int endSubband = (*iter).get_i();
  if (endSubband < startSubband) {
    endSubband = startSubband;
    (*iter).set(startSubband);
  }

  ++iter; // duration
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0.0, mf->file_duration());
  double duration = (*iter).get_r();

  //check duration
  double time = endTime - startTime;
  if ( duration > time) {
    duration = time;
    (*iter).set(time);
  }

  //nothing to do for k

}

// function to work on input parameters and return output parameters
static list<ModuleParam> *
apply_centralmoment (Module * m, list<ModuleParam> * paramsIn) {

  // the output parameter list
  list<ModuleParam> * mpl = new list<ModuleParam>();
  
  // iterator in input parmeter list
  list<ModuleParam>::iterator iter;

  //Getting input parameters and last checking
  //sound file
  iter = paramsIn->begin();
  if (iter == paramsIn->end()) return mpl;
  SOUNDfile * mf = (*iter).get_sf();
  if (mf == NULL) return mpl;
  
  //start time
  ++iter;
  double startTime = (*iter).get_r();

  //end time
  ++iter;
  double endTime = (*iter).get_r();

  //check time
  if (endTime < startTime) {
    endTime = startTime;
  }

  // start subband
  ++iter;
  int startSubband = (*iter).get_i();
  
  //end subband
  ++iter;
  int endSubband = (*iter).get_i();

  //check subband
  if (endSubband < startSubband) {
      endSubband = startSubband;
  }

  //duration
  ++iter;
  double duration = (*iter).get_r();

  //check duration
  double time = endTime - startTime;
  if ( duration > time) {
    duration = time;
  }

  //k
  ++iter;
  int k = (*iter).get_i();


  //End getting and checking

  //convert time into window number
  long start = mf->time2window( (float)startTime);

  long end = mf->time2window( (float)endTime);

  // go to start window within soundfile
  if (!mf->seek_window(start)) {
    cerr << "MaaateM: Error when positioning" << endl;
    cerr << "         startposition = 0.0" << endl;
    mf->seek_window(0);
    start = 0;
  
  }
  
  // analyse first window
  if (!mf->next_window(HIGH)) {
    cerr << "MaaateM: Warning: could not analyse first window." << endl;
    return mpl;
  }


  int M = mf->time2window(duration); //nb of window in duration (strictly, not rounded)
  int nb_gr =  end - start; //nb of window in the soundfile time-segment to analyse (strictly, not rounded)

  //check:
  if (M == 0) {
    M = 1;
  }

  int rest = nb_gr % M;

  //columns nb
  int columns = nb_gr / M + ( (rest == 0) ? 0 : 1 );

  //rows nb
  int nbsubband = endSubband - startSubband + 1;

  SegmentData *result = new SegmentData(startTime,
					endTime,
					columns,
					nbsubband);

  //temporary results
  int idx; //index variable
  double **temp;
  double *sum_avg = new double[nbsubband];
  double tempy;
  int cour = 0;

  temp = new double*[M];
  for (idx=0; idx<M; idx++)
    temp[idx] = new double[nbsubband];

  //init sum_avg
  for (int sb = 0; sb < nbsubband; sb++) {
    sum_avg[sb] = 0.0;
  }

  while (result->colFilled < columns) {
 
    // extract the subband mean for each selected subbands 
    // and compute the sum necessary to compute the mean.
    for (int sb = startSubband; sb <= endSubband; sb++) {  //from one subband to another
      temp[cour][sb-startSubband]= mf->subband_mean(sb,HIGH);
      sum_avg[sb-startSubband] +=  temp[cour][sb-startSubband];
    }
    
    cour++;
    
    
    if (result->colFilled == (columns-1) && rest != 0 && cour == rest) { //at the end of the segment to analyse
      //let`s compute the result for that duration segment
      
      for ( int sb = 0; sb < nbsubband; sb++) { // from one subband to another
	
	tempy = 0.0;
	for ( int i = 0; i < rest; i++ ) {
	  tempy += pow( (temp[i][sb] - (sum_avg[sb] / rest) ) , k);
	}
	result->data[result->colFilled][sb] = tempy / rest;
      }
      
      result->colFilled++;
      break;
    }
    

    if (cour == M) { //at the end of a duration
      //let`s compute the result for that duration segment
      
      for ( int sb = 0; sb < nbsubband; sb++) { // from one subband to another
	
	tempy = 0.0;
	for ( int i = 0; i < M; i++ ) {
	  tempy += pow( (temp[i][sb] - (sum_avg[sb] / M) ), k);
	  if ( i == (M-1)) {
	    sum_avg[sb] = 0.0;
	  }
	}
	result->data[result->colFilled][sb] = tempy / M;
      }
      
      result->colFilled++;

      //init for next duration segment
      cour = 0;
    }
    
    // analyse next window
    if (!mf->next_window(HIGH)) {
      break;
    }
	
  }

	
  
#ifdef DEBUG    
  cout << "Columns filled:" << result->colFilled << endl;
  cout << "Sum:" << result->sum() << endl;
  cout << "Average:" << result->avg() << endl;
  cout << *result;             
#endif
  
  mpl->push_back(ModuleParam(result));

  for (idx=0; idx<M; idx++)
	delete[] temp[nbsubband];
  delete[] temp;
  temp = 0;

  return mpl;
  
}; 

             
Module
loadcentralMomentModule (void)
{
  return Module(init_centralmoment,
		NULL,
		suggest_centralmoment,
		apply_centralmoment,
		NULL,
		NULL);
}
