/*
    MPEG Maaate: An Australian MPEG audio analysis toolkit
    Copyright (C) 2000 Commonwealth Scientific and Industrial Research Organisation
    (CSIRO), Australia.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "module.H"

// initialisation of the histogram1D module
static void 
init_histogram1D(Module * m) {
  // set up member values
  m->set_name(string("histogram1D"));
  m->set_desc(string("calculates a histogram from a (1D) contour "
		     "using a specific number of bins and an interval "
		     "over which the bins are evenly distributed"));
  m->set_author(string("CSIRO-MIS DMIS Silvia Pfeiffer"));
  m->set_copyright(string("(c) 1997-2000 CSIRO"));
  m->set_url(string("http://www.cmis.csiro.au/Maaate/docs/modules.html"));

  // set up input parameter list
  m->inputSpecs()->clear();
  
  // first parameter: segment data
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("segment-data"),
		     string("the contour being stored in a segment data "
			    "structure"),
		     MAAATE_TYPE_SEGMENTDATA,
		     new ModuleParam((SegmentData*)NULL)) // default
     ); 
  
  // second parameter: start time
  MaaateConstraint * constraint = new MaaateConstraint();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("starttime"),
		     string("time instant from which to start the "
			    "histogram1D calculation"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.0), // default
		     constraint));
  
  // third parameter: end time
  constraint->clear();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("endtime"),
		     string("time instant until which to calculate "
			    "the histogram1D"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)DBL_MAX), // default
		     constraint));
  
  // fourth parameter: number of bins
  constraint->clear();
  constraint->addConstraintGreaterThan(0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("bins"),
		     string("number of bins for histogram"),
		     MAAATE_TYPE_INT,
		     new ModuleParam((int)10), // default
		     constraint));

  // fifth parameter: lower boundary of histogram interval
  // DBL_MAX implies min available value
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("minhisto"),
		     string("lower boundary of the interval covered by "
			    "bins of the histogram"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)DBL_MAX))); // default

  // sixth parameter: upper boundary of histogram interval
  // DBL_MIN implies max available value
  constraint->clear();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("maxhisto"),
		     string("upper boundary of the interval covered by "
			    "bins of the histogram"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)DBL_MIN))); // default
  
  // set up output parameter list
  m->outputSpecs()->clear();
  
  // first output parameter: histogram1D output
  m->outputSpecs()->push_back
    (ModuleParamSpec(string("histogram1D"),
		     string("calculated histogram1D"),
		     MAAATE_TYPE_SEGMENTDATA,
		     new ModuleParam((SegmentData*)NULL)) // default
     );
  
  // second output parameter: histogram1D minhisto value used
  m->outputSpecs()->push_back
    (ModuleParamSpec(string("minhisto"),
		     string("actually used minhisto value"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.0)) // default
     );
  
  // third output parameter: histogram1D maxhisto value uesd
  m->outputSpecs()->push_back
    (ModuleParamSpec(string("maxhisto"),
		     string("actually used maxhisto value"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.0)) // default
     );
  
  // fourth output parameter: histogram1D binwidth
  m->outputSpecs()->push_back
    (ModuleParamSpec(string("binwidth"),
		     string("actually used with of the bins"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.0)) // default
     );
  
  // fifth output parameter: histogram1D number of values
  m->outputSpecs()->push_back
    (ModuleParamSpec(string("nrvalues"),
		     string("number of calculated values"),
		     MAAATE_TYPE_INT,
		     new ModuleParam((int)0)) // default
     );
  
};


// function to suggest parameter values given that some of the input
// parameters have already been set
static void
suggest_histogram1D (Module * m, list<ModuleParam> * paramsIn) {
  list<ModuleParam>::iterator iter;

  // get module's input parameter specifications
  list<ModuleParamSpec> * inSpecs = m->inputSpecs();
  list<ModuleParamSpec>::iterator iterSpecs;

  iter = paramsIn->begin(); // SegmentData
  iterSpecs = inSpecs->begin();

  // if no input parameters, return
  if (iter == paramsIn->end()) return;
  SegmentData * sd = (*iter).get_sd();
  if (sd == NULL) return;
  
  ++iter; // start time
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(sd->start(), sd->end());
  double startTime = (*iter).get_r();

  ++iter; // end time
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(sd->start(), sd->end());
  double endTime = (*iter).get_r();
  if (endTime < startTime) {
    endTime = startTime;
    (*iter).set(startTime);
  }

  ++iter; // number of bins
  ++iterSpecs;

  ++iter; // lower interval boundary
  ++iterSpecs;
  double lowerBdry = (*iter).get_r();

  ++iter; // upper interval boundary
  ++iterSpecs;
  /* xxx: somehow breaks
  if (lowerBdry < DBL_MAX) {
    (*iterSpecs).constraint()->clear();
    (*iterSpecs).constraint()->addConstraintGreaterThan(lowerBdry);
  }
  */

};


// function to work on input parameters and return output parameters
static list<ModuleParam> *
apply_histogram1D (Module * m, list<ModuleParam> * paramsIn) {
  // the output parameter list
  list<ModuleParam> * mpl = new list<ModuleParam>();
  list<ModuleParam>::iterator iter;

  iter = paramsIn->begin(); // SegmentData
  if (iter == paramsIn->end()) return mpl;
  SegmentData * sd = (*iter).get_sd();
  if (sd == NULL) return mpl;
  
  ++iter; // start time
  double startTime = (*iter).get_r();

  ++iter; // end time
  double endTime = (*iter).get_r();
  if (endTime < startTime) {
    endTime = startTime;
  }

  ++iter; // number of bins
  int bins = (*iter).get_i();

  ++iter; // lower interval boundary
  double minhisto = (*iter).get_r();

  ++iter; // upper interval boundary
  double maxhisto = (*iter).get_r();

  // check that there's data there to calculate histo from
  if (sd->no_columns()<=0 || sd->no_rows()<=0) {
    return mpl;
  }

  // create result segment with 1 column and 'bins' rows
  // and initialize data values
  SegmentData *result = new SegmentData (startTime, endTime, 1, bins);
  result->initData(0.0);
  result->colFilled = 1;

  // check if there are values given for minhisto and maxhisto
  // within which to scale the bins
  // if no, calculate max and min values as boundaries
  double maxRow, minRow;
  if (!(minhisto < DBL_MAX && maxhisto > DBL_MIN)) {
    // determine max and min within the requested columns of the 
    // first row (for bin resolution calc)
    maxRow = sd->smax(startTime, endTime, 0, 0);
    minRow = sd->smin(startTime, endTime, 0, 0);
  } else {
    maxRow = maxhisto;
    minRow = minhisto;
  }

  // determine bin resolution
  double binRes;
  if (maxRow > minRow) {
    binRes = 1.0*bins / (maxRow - minRow);
  } else {
    binRes = 1.0;
  }
  
  // determine col number corresponding to start and end times
  // asserting the data field had been filled
  int startcol = sd->time2col(startTime);
  int endcol   = sd->time2col(endTime);
  if (sd->col_filled() < endcol) endcol = sd->col_filled();

  // -------------------------------------------------------------------
  // calculate histogrm from sd values 
  // -------------------------------------------------------------------
  int binNo;
  for (int i=startcol; i<endcol; i++) {
    // determine bin number
    // e.g. sd-values lie between 0-1; 2 bins
    // => values in [0;0.5[ increment result->sd[0][0] (bin 0)
    //    values in [0.5;1] increment result->sd[0][1] (bin 1)
    binNo = (int) (binRes * (sd->data[i][0] - minRow));
    if (binNo >= bins) binNo = bins-1;
    if (binNo < 0) binNo = 0; // just in case ...
    result->data[0][binNo] += 1;
  }

  // push results to output parameter list
  mpl->push_back(ModuleParam(result));
  mpl->push_back(ModuleParam(minRow));
  mpl->push_back(ModuleParam(maxRow));
  mpl->push_back(ModuleParam((maxRow-minRow)/bins));
  mpl->push_back(ModuleParam(endcol-startcol));

  return mpl;
  
};


Module
loadHistogram1DModule (void)
{
  // init_histogram1D gets executed during construction of module
  return Module(init_histogram1D,
		NULL,
		suggest_histogram1D,
		apply_histogram1D,
		NULL,
		NULL);
}
