/*
    MPEG Maaate: An Australian MPEG audio analysis toolkit
    Copyright (C) 2000 Commonwealth Scientific and Industrial Research Organisation
    (CSIRO), Australia.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "module.H"

// initialisation of the segmentation module
static void 
init_segmentation(Module * m) {
  // set up member values
  m->set_name(string("segmentation"));
  m->set_desc(string("calculates a temporal segmentation from a contour "
		     "based upon the contour staying above or below "
		     "a threshold for a minimum duration"));
  m->set_author(string("CSIRO-MIS DMIS Silvia Pfeiffer"));
  m->set_copyright(string("(c) 1997-2000 CSIRO"));
  m->set_url(string("http://www.cmis.csiro.au/Maaate/docs/modules.html"));

  // set up input parameter list
  m->inputSpecs()->clear();
  
  // first parameter: segment data
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("segment-data"),
		     string("the contour being stored in a segment data "
			    "structure"),
		     MAAATE_TYPE_SEGMENTDATA,
		     new ModuleParam((SegmentData*)NULL)) // default
     ); 
  
  // second parameter: start time
  MaaateConstraint * constraint = new MaaateConstraint();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("starttime"),
		     string("time instant from which to start the "
			    "segmentation calculation"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.0), // default
		     constraint));
  
  // third parameter: end time
  constraint->clear();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("endtime"),
		     string("time instant until which to calculate "
			    "the segmentation"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)DBL_MAX), // default
		     constraint));
  
  // fourth parameter: below [or above]
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("below_threshold"),
		     string("whether to find regions of the contour below "
			    "or above the threshold"),
		     MAAATE_TYPE_BOOL,
		     new ModuleParam(true))); // default

  // fifth parameter: threshold
  constraint->clear();
  constraint->addConstraintRange(0.0, 1.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("threshold"),
		     string("level above/below which the contour "
			    "is considered for segmentation"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.5), // default
		     constraint));
  
  // sixth parameter: minimum duration
  constraint->clear();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("minduration"),
		     string("minimum amount of time the contour "
			    "must stay above/below the given threshold "
			    "to give a segment"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.0), // default
		     constraint));
  
  // seventh parameter: tolerated interruption
  constraint->clear();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("tolerated_interruption"),
		     string("maximum amount of time the contour "
			    "may fall outside the given threshold "
			    "without affecting the segmentation"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.0), // default
		     constraint));
  
  // eighth parameter: onset time
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("onsettime"),
		     string("time by which to delay the start of a segment"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.0)));

  // ninth parameter: release time
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("releasetime"),
		     string("time to shorten the end of a segment"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.0)));
  
  
  // set up output parameter list
  m->outputSpecs()->clear();
  
  // first output parameter: segmentation output
  m->outputSpecs()->push_back
    (ModuleParamSpec(string("segmentation"),
		     string("calculated segmentation"),
		     MAAATE_TYPE_SEGMENTTABLE,
		     new ModuleParam((SegmentTable*)NULL)) // default
     );
  
};


// function to suggest parameter values given that some of the input
// parameters have already been set
static void
suggest_segmentation (Module * m, list<ModuleParam> * paramsIn) {
  list<ModuleParam>::iterator iter;

  // get module's input parameter specifications
  list<ModuleParamSpec> * inSpecs = m->inputSpecs();
  list<ModuleParamSpec>::iterator iterSpecs;

  iter = paramsIn->begin(); // SegmentData
  iterSpecs = inSpecs->begin();
  if (iter == paramsIn->end()) return;
  SegmentData * sd = (*iter).get_sd();
  if (sd == NULL) return;
  
  ++iter; // start time
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(sd->start(), sd->end());
  double startTime = (*iter).get_r();

  ++iter; // end time
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(sd->start(), sd->end());
  double endTime = (*iter).get_r();
  if (endTime < startTime) {
    endTime = startTime;
    (*iter).set(startTime);
  }

  ++iter; // below
  ++iterSpecs;

  ++iter; // threshold
  ++iterSpecs; // constraint of [0.0;1.0] already exists
  double threshold = (*iter).get_r();

  ++iter; // minimum duration
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0.0, sd->duration());
  double minDuration = (*iter).get_r();

  ++iter; // maximum tolerated interruption
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0.0, sd->duration());
  
  ++iter; // onset time
  ++iterSpecs;
  if ((*iter).get_r() > minDuration) (*iter).set(minDuration);

  ++iter; // release time
  ++iterSpecs;
  if ((*iter).get_r() > minDuration) (*iter).set(minDuration);

};

// function to work on input parameters and return output parameters
static list<ModuleParam> *
apply_segmentation (Module * m, list<ModuleParam> * paramsIn) {
  // the output parameter list
  list<ModuleParam> * mpl = new list<ModuleParam>();
  list<ModuleParam>::iterator iter;

  iter = paramsIn->begin(); // SegmentData
  if (iter == paramsIn->end()) return mpl;
  SegmentData * sd = (*iter).get_sd();
  if (sd == NULL) return mpl;
  
  ++iter; // start time
  double startTime = (*iter).get_r();

  ++iter; // end time
  double endTime = (*iter).get_r();
  if (endTime < startTime) {
    endTime = startTime;
  }

  ++iter; // below
  bool below = (*iter).get_b();

  ++iter; // threshold
  double threshold = (*iter).get_r();

  ++iter; // minimum duration
  double minDuration = (*iter).get_r();

  ++iter; // maximum tolerated interruption
  double maxInterruptTolerated = (*iter).get_r();
  
  ++iter; // onset time
  double onsetTime = (*iter).get_r();
  if (onsetTime > minDuration) onsetTime = minDuration;

  ++iter; // release time
  double releaseTime = (*iter).get_r();
  if (releaseTime > minDuration) releaseTime = minDuration;

  // calculate the threshold value from relative value
  double maxSD = sd->smax(startTime, endTime, 0, 0);
  double minSD = sd->smin(startTime, endTime, 0, 0);
  double thresh = (maxSD - minSD)*threshold + minSD;

  // calculate start and end columns to go through
  int startcol = sd->time2col(startTime);
  int endcol = sd->time2col(endTime);
  if (sd->col_filled() < endcol) endcol = sd->col_filled();

  // create result object
  SegmentTable *result = new SegmentTable();

  // -------------------------------------------------------------------
  // calculate segments from sd values considering threshold
  // -------------------------------------------------------------------

  bool   within     = false; // has a possible segment started?
  double segStart   = 0.0;   // when did it start?
  double segEnd     = 0.0;   // when does it end?
  double segDur     = 0.0;   // how long is it now?
  double interDur   = 0.0;   // how long is a interruption?
  double confidence = 0.0;   // how strong is the calculated content?
  int    numSteps   = 0;     // counts no steps of a segment (for confidence)
  double area       = 0.0;   // area covered by signal (for confidence)
  double areaUnderT = 0.0;   // area below threshold (for confidence)

  // go through sd values counting values below/above threshold
  for (int i=startcol; i<endcol; i++) {
    if ((below ? sd->data[i][0] <= thresh :
	 sd->data[i][0] >= thresh)) { // below/above thresh
      // check for start of segment
      if (within) { // continue a segment
	segDur += sd->resolution();
	numSteps++;
	area += sd->data[i][0] - minSD;
      } else { // start a segment
	segDur = sd->resolution();
	numSteps = 1;
	area = sd->data[i][0] - minSD;
	within = true;
	segStart = startTime + i*sd->resolution();
      }
      interDur = 0.0;
    } else { // above/below thresh
      interDur += sd->resolution();
      // if segment so far was too small, ignore it
      if (segDur < minDuration) {
	within = false;
      }
      // check for end of segment and if it was long enough
      if (within &&
	  (interDur > maxInterruptTolerated) &&
	  ((segDur-interDur) >= minDuration)) {
	segEnd = segStart + segDur - interDur;
	// calculate confidence value
	areaUnderT = numSteps * (thresh-minSD);
	if (thresh > 0.0) {
	  if (below) {
	    confidence = 1.0 - (area / areaUnderT);
	  } else {
	    confidence = (area - areaUnderT) / 
	      (numSteps*(maxSD-minSD) - areaUnderT);
	  }
	} else {
	  confidence = 1.0;
	}
	// adjust with onset and release time
	// but bigger than start and end times of sd
	segStart = min(max(sd->start(), segStart+onsetTime),sd->end());
	segEnd = max(min(sd->end(), segEnd-releaseTime),sd->start());
	
#ifdef DEBUG
	cout << "/----" << endl;
	cout << "insert " << segStart << "\t" << segEnd << endl;
	cout << "frame " << mf->frame(segStart) << "\t" 
	     << mf->frame(segEnd) << endl;
	cout << "confidence " << confidence << endl;
	cout << "area " << area << endl;
	cout << "numSteps " << numSteps << endl;
	cout << "----/" << endl;
#endif
	
	result->append (SegmentData(segStart, 
				    segEnd,
				    0, 0, // cols + rows
				    0,    // id
				    ' ',  // flag
				    confidence));
	within = false;
      }
      // if still within segment, adjust duration and area
      if (within) {
	segDur += sd->resolution();
	numSteps++;
	area += thresh - minSD;
      }
    }
  }

  // append the ending content if long enough
  if (within &&  ((segDur-interDur) >= minDuration)) {
    segEnd = segStart + segDur - interDur;
    // calculate confidence value
    if (thresh > 0.0) {
      areaUnderT = numSteps * (thresh-minSD);
      if (below) {
	confidence = 1.0 - (area / areaUnderT);
      } else {
	confidence = (area - areaUnderT) / 
	  (numSteps*(maxSD-minSD) - areaUnderT);
      }
    } else {
      confidence = 1.0;
    }
    // adjust with onset and release time
    segStart = min(max(sd->start(), segStart+onsetTime),sd->end());
    segEnd = max(min(sd->end(), segEnd-releaseTime),sd->start());
#ifdef DEBUG
    cout << "/----" << endl;
    cout << "insert " << segStart << "\t" << segEnd << endl;
    cout << "frame " << mf->frame(segStart) << "\t" 
	 << mf->frame(silEnd) << endl;
    cout << "confidence " << confidence << endl;
    cout << "----/" << endl;
#endif
    result->append (SegmentData(segStart, segEnd,
				0, 0, // cols + rows
				0,    // id
				' ',  // flag
				confidence));
  }

  
  mpl->push_back(ModuleParam(result));

  return mpl;
  
};


Module
loadSegmentationModule (void)
{
  // init_segmentation gets executed during construction of module
  return Module(init_segmentation,
		NULL,
		suggest_segmentation,
		apply_segmentation,
		NULL,
		NULL);
}
