/*
    MPEG Maaate: An Australian MPEG audio analysis toolkit
    Copyright (C) 2000 Commonwealth Scientific and Industrial Research Organisation
    (CSIRO), Australia.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "module.H"

#include <math.h>

#define ROUND(number) floor((number)+0.5)

static void 
init_sigbdwidth(Module * m) {
  
  // set up member values
  m->set_name (string("sigBdwidth"));
  m->set_desc (string("bandwidth calculated via a non-normailsed threshold"));
  m->set_author (string("CSIRO-MIS AAS Thomas VINCENT"));
  m->set_copyright (string("(c) 2002 CSIRO"));
  m->set_url (string("http://www.cmis.csiro.au/Maaate/docs/modules.html"));

  // set up input parameter list
  m->inputSpecs()->clear();  
  
  // first parameter: SegmentData     
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("segment-data"),
		     string("the segment-data from which the signal "
			    "bandwidth gets calculated"),
		     MAAATE_TYPE_SEGMENTDATA,
		     new ModuleParam((SegmentData*)NULL)) // default
     ); 
  
  // second parameter: start time
  MaaateConstraint * constraint = new MaaateConstraint();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("starttime"),
		     string("time instant from which to start the "
			    "signal bandwidth calculation"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.0), // default
		     constraint));
  
  // third parameter: end time
  constraint = new MaaateConstraint();
  constraint->clear();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("endtime"),
		     string("time instant until which to calculate "
			    "the signal bandwidth"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)DBL_MAX), // default
		     constraint));
  

  // fourth parameter: threshold
  constraint = new MaaateConstraint();
  constraint->clear();
  constraint->addConstraintRange(- DBL_MAX,0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("threshold"),
		     string("normalised threshold which permit to find out"
			    "the signal bandwidth"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam(0.0), // default
		     constraint));
  

  // set up output parameter list
  m->outputSpecs()->clear();
  
  // first parameter: bandwidth curve
  m->outputSpecs()->push_back
    (ModuleParamSpec(string("bandwidth curve"),
		     string("indicator of speech/music in Hz"),
		     MAAATE_TYPE_SEGMENTDATA,
		     new ModuleParam((SegmentData*)NULL)) // default
     );

  // second parameter: fcmin curve
  m->outputSpecs()->push_back
    (ModuleParamSpec(string("fcmin curve"),
		     string("min freq upon the threshold in Hz"),
		     MAAATE_TYPE_SEGMENTDATA,
		     new ModuleParam((SegmentData*)NULL)) // default
     );

  // third parameter: fcmax curve
  m->outputSpecs()->push_back
    (ModuleParamSpec(string("fcmax curve"),
		     string("max freq upon the threshold in Hz"),
		     MAAATE_TYPE_SEGMENTDATA,
		     new ModuleParam((SegmentData*)NULL)) // default
     );
  
};


// function to suggest parameter values given that some of the input
// parameters have already been set; also changes constraints accordingly
static void
suggest_sigbdwidth (Module * m, list<ModuleParam> * paramsIn) {
  list<ModuleParam>::iterator iter;

  // get module's input parameter specifications
  list<ModuleParamSpec> * inSpecs = m->inputSpecs();
  list<ModuleParamSpec>::iterator iterSpecs;
  
  iter = paramsIn->begin(); // Segmentdata
  iterSpecs = inSpecs->begin();
  if (iter == paramsIn->end()) return;
  SegmentData * sd = (*iter).get_sd();
  if (sd == NULL) return;
  
  ++iter; // start time
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0.0, sd->duration());
  double startTime = (*iter).get_r();

  ++iter; // end time
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0.0, sd->duration());
  double endTime = (*iter).get_r();

  //check time
  if (endTime < startTime) {
    endTime = startTime;
    (*iter).set(startTime);
  }

  //nothing to check for threshold
}

// function to work on input parameters and return output parameters
static list<ModuleParam> *
apply_sigbdwidth (Module * m, list<ModuleParam> * paramsIn) {

  // the output parameter list
  list<ModuleParam> * mpl = new list<ModuleParam>();
  
  // iterator in input parmeter list
  list<ModuleParam>::iterator iter;

  //Getting input parameters and last checking
  //sound file
  iter = paramsIn->begin();
  if (iter == paramsIn->end()) return mpl;
  SegmentData * sd = (*iter).get_sd();
  if (sd == NULL) return mpl;
  
  //start time
  ++iter;
  double startTime = (*iter).get_r();

  //end time
  ++iter;
  double endTime = (*iter).get_r();

  //check time
  if (endTime < startTime) {
    endTime = startTime;
  }

  //threshold
  ++iter;
  double Ts = (*iter).get_r();

  //End getting and checking


 // calculate start and end columns to go through
  int startcol = sd->time2col(startTime);
  int endcol = sd->time2col(endTime);
  if (sd->col_filled() < endcol) endcol = sd->col_filled();
  int nbcol = endcol - startcol;

   SegmentData *result = new SegmentData(startTime,
					endTime,
					nbcol,
					1);

  SegmentData *result1 = new SegmentData(startTime,
					endTime,
					nbcol,
					1);

  SegmentData *result2 = new SegmentData(startTime,
					endTime,
					nbcol,
					1);

  //total no of subbands
  int nb_SB = sd->no_rows();


  //temporary results
  int accol = startcol ;
  int min;
  int max;

  while (accol < endcol) {
   

    // extract the min
    min = 0;
    while ( min < nb_SB ) {
      if ( (sd->data[accol][min]) > Ts )  break;
      min++;
    } 

      //there is a min
      if ( min != nb_SB) {
	// extract the max
	max = nb_SB-1;
	while ( max > min ) {
	  if ( (sd->data[accol][max]) > Ts )  break;
	  max--;
	} 
	//set result
	result->data[result->colFilled][0] = max-min+1;

      } else { // there is no bandwidth
	//set result
	result->data[result->colFilled][0] = 0.0;
	max = -1;
      }

      result->colFilled++;

      //store fcmin and fcmax
      result1->data[result1->colFilled++][0] = min; //FCMIN
      result2->data[result2->colFilled++][0] = max; //FCMAX
      accol ++;
  }
  
#ifdef DEBUG    
  cout << "Columns filled:" << result->colFilled << endl;
  cout << "Sum:" << result->sum() << endl;
  cout << "Average:" << result->avg() << endl;
  cout << *result;             
#endif
  
  mpl->push_back(ModuleParam(result));
  mpl->push_back(ModuleParam(result1));
  mpl->push_back(ModuleParam(result2));
 
  return mpl;
  
 
}            
             
Module
loadsigBdwidthModule (void)
{
  return Module(init_sigbdwidth,
		NULL,
		suggest_sigbdwidth,
		apply_sigbdwidth,
		NULL,
		NULL);
}
