/*
    MPEG Maaate: An Australian MPEG audio analysis toolkit
    Copyright (C) 2000 Commonwealth Scientific and Industrial Research Organisation
    (CSIRO), Australia.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "module.H"

#include <math.h>

#define ROUND(number) floor((number)+0.5)

static void 
init_sigsb(Module * m) {
  
  // set up member values
  m->set_name (string("sigsb"));
  m->set_desc (string("number of subbands with a significant level calculated via a threshold, resolution: window"));
  m->set_author (string("CSIRO-MIS AAS Thomas VINCENT"));
  m->set_copyright (string("(c) 2002 CSIRO"));
  m->set_url (string("http://www.cmis.csiro.au/Maaate/docs/modules.html"));

  // set up input parameter list
  m->inputSpecs()->clear();  
  
  // first parameter: sound file     
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("soundfile"),
		     string("the SOUND file for which the number of "
			    "subbands gets calculated"),
		     MAAATE_TYPE_SOUNDFILE,
		     new ModuleParam((SOUNDfile*)NULL)) // default
     ); 
  
  // second parameter: start time
  MaaateConstraint * constraint = new MaaateConstraint();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("starttime"),
		     string("time instant from which to start the "
			    "calculation"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.0), // default
		     constraint));
  
  // third parameter: end time
  constraint = new MaaateConstraint();
  constraint->clear();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("endtime"),
		     string("time instant until which to calculate "
			    "the number of significant subbands"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)DBL_MAX), // default
		     constraint));
  

  // fourth parameter: threshold
  constraint = new MaaateConstraint();
  constraint->clear();
  constraint->addConstraintRange(0.0,1.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("threshold"),
		     string("normalised threshold which permit to find out"
			    "the number of significant subbands"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.1), // default                      XXX: try it
		     constraint));
  

  // set up output parameter list
  m->outputSpecs()->clear();
  
  // first parameter: significant subbands curve
  m->outputSpecs()->push_back
    (ModuleParamSpec(string("number of significant subbands curve"),
		     string("indicator of speech/music"),
		     MAAATE_TYPE_SEGMENTDATA,
		     new ModuleParam((SegmentData*)NULL)) // default
     );
  
};


// function to suggest parameter values given that some of the input
// parameters have already been set; also changes constraints accordingly
static void
suggest_sigsb (Module * m, list<ModuleParam> * paramsIn) {
  list<ModuleParam>::iterator iter;

  // get module's input parameter specifications
  list<ModuleParamSpec> * inSpecs = m->inputSpecs();
  list<ModuleParamSpec>::iterator iterSpecs;
  
  iter = paramsIn->begin(); // SOUND file
  iterSpecs = inSpecs->begin();
  if (iter == paramsIn->end()) return;
  SOUNDfile * mf = (*iter).get_sf();
  if (mf == NULL) return;
  
  ++iter; // start time
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0.0, mf->file_duration());
  double startTime = (*iter).get_r();

  ++iter; // end time
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0.0, mf->file_duration());
  double endTime = (*iter).get_r();

  //check time
  if (endTime < startTime) {
    endTime = startTime;
    (*iter).set(startTime);
  }

  //nothing to check for threshold
}

// function to work on input parameters and return output parameters
static list<ModuleParam> *
apply_sigsb (Module * m, list<ModuleParam> * paramsIn) {

  // the output parameter list
  list<ModuleParam> * mpl = new list<ModuleParam>();
  
  // iterator in input parmeter list
  list<ModuleParam>::iterator iter;

  //Getting input parameters and last checking
  //SOUND file
  iter = paramsIn->begin();
  if (iter == paramsIn->end()) return mpl;
  SOUNDfile * mf = (*iter).get_sf();
  if (mf == NULL) return mpl;
  
  //start time
  ++iter;
  double startTime = (*iter).get_r();

  //end time
  ++iter;
  double endTime = (*iter).get_r();

  //check time
  if (endTime < startTime) {
    endTime = startTime;
  }

  //threshold
  ++iter;
  double Ts = (*iter).get_r();  //normalised value

  //End getting and checking

  //convert time into window number
  long start = mf->time2window(startTime);
  
  long end = mf->time2window(endTime);
  
  // go to start window within soundfile
    if (!mf->seek_window(start)) {
      cerr << "MaaateM: Error when positioning" << endl;
      cerr << "         startposition = 0.0" << endl;
      mf->seek_window(0);
    }
  

  //analyse first window
  if (!mf->next_window(HIGH)) {
    cerr << "MaaateM: error parsing frame header" << endl;
    return mpl;
  }

  //columns number
  int columns = end - start;
  columns = (columns > mf->file_window_number()) ? mf->file_window_number() : columns;

  //bandwidth
  SegmentData *result = new SegmentData(startTime,
					endTime,
					columns,
					1);


  //total no of subbands
  int nb_SB = mf->nb_subbands(HIGH);

  //temporary results
  int sb;
  double *array = new double[nb_SB];
  double thresh;

  while (mf->at_window() <= end) {

    //find the threshold using the normalised value and the local max
    thresh = 0.0;
    //find maximum of that window
    for (sb = 0; sb < nb_SB; sb++) {
      array[sb] =  mf->subband_mean(sb,HIGH);
      if ( array[sb] > thresh ) {
	thresh = array[sb];
      }
    }
    thresh = (thresh == 0.0) ? DBL_MAX : (thresh * Ts);    

    //count the significant subbands
    int count = 0;
    for (sb = 0; sb < nb_SB; sb++) {
      if ( array[sb] >= thresh ) count++;
    }

    result->data[result->colFilled++][0] = count;
    
    // analyse next window
    if (!mf->next_window(HIGH)) {
      break;
    }
   
  }
  
#ifdef DEBUG    
  cout << "Columns filled:" << result->colFilled << endl;
  cout << "Sum:" << result->sum() << endl;
  cout << "Average:" << result->avg() << endl;
  cout << *result;             
#endif
  
  mpl->push_back(ModuleParam(result));

  delete[] array;
  array = 0;
 
  return mpl;
  
 
}            
             
Module
loadsigsbModule (void)
{
  return Module(init_sigsb,
		NULL,
		suggest_sigsb,
		apply_sigsb,
		NULL,
		NULL);
}
