/*
    MPEG Maaate: An Australian MPEG audio analysis toolkit
    Copyright (C) 2000 Commonwealth Scientific and Industrial Research Organisation
    (CSIRO), Australia.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "module.H"

#include <math.h>

#define ROUND(number) floor((number)+0.5)

static void 
init_spectralFlux(Module * m) {
  
  // set up member values
  m->set_name (string("spectralFlux"));
  m->set_desc (string("2-norm of the difference between normalised subband value vectors"
		      "at n and n+1"));
  m->set_author (string("CSIRO-MIS AAS Thomas VINCENT"));
  m->set_copyright (string("(c) 2002 CSIRO"));
  m->set_url (string("http://www.cmis.csiro.au/Maaate/docs/modules.html"));

  // set up input parameter list
  m->inputSpecs()->clear();  
  
  // first parameter:  sound file     
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("soundfile"),
		     string("the SOUND file for which the spectral "
			    "flux gets calculated"),
		     MAAATE_TYPE_SOUNDFILE,
		     new ModuleParam((SOUNDfile*)NULL)) // default
     ); 
  
  // second parameter: start time
  MaaateConstraint * constraint = new MaaateConstraint();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("starttime"),
		     string("time instant from which to start the "
			    "spectral flux calculation"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.0), // default
		     constraint));
  
  // third parameter: end time
  constraint = new MaaateConstraint();
  constraint->clear();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("endtime"),
		     string("time instant until which to calculate "
			    "the spectral flux"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)DBL_MAX), // default
		     constraint));
  
  // set up output parameter list
  m->outputSpecs()->clear();
  
  // first parameter: subband energies curves
  m->outputSpecs()->push_back
    (ModuleParamSpec(string("spectral flux curve"),
		     string("determines changes of spectral energy distribution of 2 successive frames"),
		     MAAATE_TYPE_SEGMENTDATA,
		     new ModuleParam((SegmentData*)NULL)) // default
     );
  
};


// function to suggest parameter values given that some of the input
// parameters have already been set; also changes constraints accordingly
static void
suggest_spectralFlux (Module * m, list<ModuleParam> * paramsIn) {
  list<ModuleParam>::iterator iter;

  // get module's input parameter specifications
  list<ModuleParamSpec> * inSpecs = m->inputSpecs();
  list<ModuleParamSpec>::iterator iterSpecs;
  
  iter = paramsIn->begin(); // sound file
  iterSpecs = inSpecs->begin();
  if (iter == paramsIn->end()) return;
  SOUNDfile * mf = (*iter).get_sf();
  if (mf == NULL) return;
  
  ++iter; // start time
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0.0, mf->file_duration());
  double startTime = (*iter).get_r();

  ++iter; // end time
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0.0, mf->file_duration());
  double endTime = (*iter).get_r();

  //check time
  if (endTime < startTime) {
    endTime = startTime;
    (*iter).set(startTime);
  }

}

// function to work on input parameters and return output parameters
static list<ModuleParam> *
apply_spectralFlux (Module * m, list<ModuleParam> * paramsIn) {

  // the output parameter list
  list<ModuleParam> * mpl = new list<ModuleParam>();
  
  // iterator in input parmeter list
  list<ModuleParam>::iterator iter;

  //Getting input parameters and last checking
  //sound file
  iter = paramsIn->begin();
  if (iter == paramsIn->end()) return mpl;
  SOUNDfile * mf = (*iter).get_sf();
  if (mf == NULL) return mpl;
  
  //start time
  ++iter;
  double startTime = (*iter).get_r();

  //end time
  ++iter;
  double endTime = (*iter).get_r();

  //check time
  if (endTime < startTime) {
    endTime = startTime;
  }
  //End getting and checking

  //convert time into window nb
  long start = mf->time2window( (float)startTime);

  long end = mf->time2window( (float)endTime);

  // go to start window within soundfile
    if (!mf->seek_window(start)) {
      cerr << "MaaateM: Error when positioning" << endl;
      cerr << "         startposition = 0.0" << endl;
      mf->seek_window(0);
      start = 0;
    }
  
  // analyse first window
  if (!mf->next_window(HIGH)) {
    cerr << "MaaateM: Warning: could not analyse first window." << endl;
    return mpl;
  }

  //col nb
  long columns = end - start;
  columns = (columns > mf->file_window_number()) ? mf->file_window_number() : columns;

  SegmentData *result = new SegmentData(startTime,
					endTime,
					columns,
					1);


  //temporary results
  int SBM = mf->nb_subbands(LOW);
  double *temp = new double[SBM];
  double *old_temp = new double[SBM];
  double max = 0.0;
  int sb;


  //init: full-fill old_temp
 for (sb = 0; sb < SBM; sb++) {  //from one subband to another
   old_temp[sb] = mf->subband_rms(sb,LOW);
   
   if (old_temp[sb] > max) {
     max = old_temp[sb];
   }
 }

 if (max != 0.0) {
   //division possible
   for (int sb = 0; sb < SBM; sb++) {
     old_temp[sb] /= max;
   }
 } //else =0 but that is already done!


  while (mf->at_window() <= end) {

    // extract the subband rms, compute square for relevant subbands
    // and find max of all subbands
    for (sb = 0; sb < SBM; sb++) {  //from one subband to another
      temp[sb] = mf->subband_rms(sb,LOW);
      
      if (temp[sb] > max) {
	max = temp[sb];
      }
    }
    
    if (max != 0.0) {
      //division possible
      for (int sb = 0; sb < SBM; sb++) {
		temp[sb] /= max;
      }
    } //else =0 but that is already done!

    //computing result with old frame
    for (sb = 0; sb < SBM; sb++) {
      max += pow( old_temp[sb] - temp[sb], 2 );
    }
    
    //saving result
    result->data[result->colFilled++][0] = sqrt (max);

    //transfer temp in old_temp
	memcpy (old_temp,temp,SBM*sizeof(double));
    max=0.0;

    // analyse next window
    if (!mf->next_window(LOW)) {
      break;
    }
    
  }
  
#ifdef DEBUG    
  cout << "Columns filled:" << result->colFilled << endl;
  cout << "Sum:" << result->sum() << endl;
  cout << "Average:" << result->avg() << endl;
  cout << *result;             
#endif
  
  mpl->push_back(ModuleParam(result));

  delete[] temp;
  temp = 0;
  delete[] old_temp;
  old_temp = 0;

  return mpl;
  
}; 
            
             
Module
loadspectralFluxModule (void)
{
  return Module(init_spectralFlux,
		NULL,
		suggest_spectralFlux,
		apply_spectralFlux,
		NULL,
		NULL);
}
