/*
    MPEG Maaate: An Australian MPEG audio analysis toolkit
    Copyright (C) 2000 Commonwealth Scientific and Industrial Research Organisation
    (CSIRO), Australia.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "module.H"

#define ROUND(number) floor((number)+0.5)

// initialisation of the sum of scalefactor module
static void 
init_sumscf(Module * m) {
  // set up member values
  m->set_name (string("sumscf"));
  m->set_desc (string("sums the scalefactors over each subband to calculate"
		      "a loudness approximation"));
  m->set_author (string("CSIRO-MIS DMIS"));
  m->set_copyright (string("(c) 2002 CSIRO"));
  m->set_url (string("http://www.cmis.csiro.au/Maaate/docs/modules.html"));

  // set up input parameter list
  m->inputSpecs()->clear();
  
  // first parameter: sound file
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("soundfile"),
		     string("the SOUND file for which the sum of "
			    "scalefactors gets calculated"),
		     MAAATE_TYPE_SOUNDFILE,
		     new ModuleParam((SOUNDfile*)NULL)) // default
     ); 
  
  // second parameter: start time
  MaaateConstraint * constraint = new MaaateConstraint();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("starttime"),
		     string("time instant from which to start the "
			    "sum of scalefactor calculation"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)0.0), // default
		     constraint));
  
  // third parameter: end time
  constraint = new MaaateConstraint();
  constraint->clear();
  constraint->addConstraintGreaterThan(0.0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("endtime"),
		     string("time instant until which to calculate "
			    "the sum of scalefactors"),
		     MAAATE_TYPE_REAL,
		     new ModuleParam((double)DBL_MAX), // default
		     constraint));
  
  // fourth parameter: start subband
  constraint = new MaaateConstraint();
  constraint->clear();
  constraint->addConstraintGreaterThan(0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("start-subband"),
		     string("subband from which to start the sum "
			    "of scalefactor calculation"),
		     MAAATE_TYPE_INT,
		     new ModuleParam(0), // default
		     constraint));

  // fifth parameter: end subband
  constraint = new MaaateConstraint();
  constraint->clear();
  constraint->addConstraintGreaterThan(0);
  m->inputSpecs()->push_back
    (ModuleParamSpec(string("end-subband"),
		     string("subband at which to end the sum of "
			    "scalefactor calculation"),
		     MAAATE_TYPE_INT,
		     new ModuleParam(0), // default
		     constraint));
  
  // set up output parameter list
  m->outputSpecs()->clear();
  
  // first parameter: sum of scalefactor curve
  m->outputSpecs()->push_back
    (ModuleParamSpec(string("loudness-curve"),
		     string("a loudness approximation for the "
			    "requested time interval calculated "
			    "from the sum of scalefactors"),
		     MAAATE_TYPE_SEGMENTDATA,
		     new ModuleParam((SegmentData*)NULL)) // default
     );
  
};


// function to suggest parameter values given that some of the input
// parameters have already been set; also changes constraints accordingly
static void
suggest_sumscf (Module * m, list<ModuleParam> * paramsIn) {
  list<ModuleParam>::iterator iter;

  // get module's input parameter specifications
  list<ModuleParamSpec> * inSpecs = m->inputSpecs();
  list<ModuleParamSpec>::iterator iterSpecs;

  iter = paramsIn->begin(); // sound file
  iterSpecs = inSpecs->begin();
  if (iter == paramsIn->end()) return;
  SOUNDfile * sf = (*iter).get_sf();
  if (sf == NULL) return;
  
  ++iter; // start time
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0.0, sf->file_duration());
  double startTime = (*iter).get_r();

  ++iter; // end time
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0.0, sf->file_duration());
  double endTime = (*iter).get_r();
  if (endTime < startTime) {
    endTime = startTime;
    (*iter).set(startTime);
  }

  ++iter; // start subband
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0, sf->nb_subbands(HIGH)-1);

  int startSubband = (*iter).get_i();

  ++iter; // end subband
  ++iterSpecs;
  (*iterSpecs).constraint()->clear();
  (*iterSpecs).constraint()->addConstraintRange(0, sf->nb_subbands(HIGH)-1);

  // check nr of subbands
  int endSubband = (*iter).get_i();
  if (endSubband < startSubband) {
    endSubband = startSubband;
    (*iter).set(startSubband);
  }

};

// function to work on input parameters and return output parameters
static list<ModuleParam> *
apply_sumscf (Module * m, list<ModuleParam> * paramsIn) {
  // the output parameter list
  list<ModuleParam> * mpl = new list<ModuleParam>();
  list<ModuleParam>::iterator iter;

  iter = paramsIn->begin(); // sound file
  if (iter == paramsIn->end()) return mpl;
  
  SOUNDfile * sf = (*iter).get_sf();
  if (sf == NULL) return mpl;
  
  ++iter; // start time
  double startTime = (*iter).get_r();

  ++iter; // end time
  double endTime = (*iter).get_r();
  if (endTime < startTime) {
    endTime = startTime;
  }

  ++iter; // start subband
  int startSubband = (*iter).get_i();
  
  ++iter; // end subband
  int endSubband = (*iter).get_i();
  // check nr of subbands depending on layer
  if (endSubband < startSubband) {
    endSubband = startSubband;
  }
  
  //convert time into window nb
  long start = sf->time2window( (float)startTime);

  long end = sf->time2window( (float)endTime);

  // go to start window within soundfile
  if (!sf->seek_window(start)) {
    cerr << "MaaateM: Error when positioning" << endl;
    cerr << "         startposition = 0.0" << endl;
    sf->seek_window(0);
    start = 0;
  }
  
  // analyse first window
  if (!sf->next_window(HIGH)) {
    cerr << "MaaateM: Warning: could not analyse first window." << endl;
    return mpl;
  }

  long columns = end - start;
  columns = (columns > sf->file_window_number()) ? sf->file_window_number() : columns;

  SegmentData *result = new SegmentData(startTime,
					endTime,
					columns,
					1);
  
  while (sf->at_window() <= end) {
    
    double sum=0.0;
   
    for (int sb = startSubband; sb <= endSubband; sb++) {
      sum += sf->subband_scalefactor(sb,HIGH);
    }
      
    // store the sum of subband mean
    result->data[result->colFilled++][0] = sum;
    sum=0.0;
    
    // analyse next window
    if (!sf->next_window(HIGH)) {
      break;
    }
    
  }
  
#ifdef DEBUG    
  cout << "Min sscf: " << result->min() << endl;
  cout << "Max sscf: " << result->max() << endl;
  cout << "Sum sscf: " << result->sum() << endl;
  cout << "Avg sscf: " << result->avg() << endl;
  cout << "Columns filled:" << result->colFilled << endl;
  cout << *result;
#endif
  
  mpl->push_back(ModuleParam(result));

  return mpl;
  
};


Module
loadSumscfModule (void)
{
  // init_sumscf gets executed during construction of module
  return Module(init_sumscf,
		NULL,
		suggest_sumscf,
		apply_sumscf,
		NULL,
		NULL);
}
