/*
    MPEG Maaate: An Australian MPEG audio analysis toolkit
    Copyright (C) 2000 Commonwealth Scientific and Industrial Research Organisation
    (CSIRO), Australia.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "plugins.H"

#include <dirent.h>

#if (defined (WIN32) || (_WIN32))
	#define DIRSEP "\\"
	#define PATHSEP ";"
#else
	#include <dlfcn.h>
	#define DIRSEP '/'
	#define PATHSEP ":"
#endif


#include <cstdlib>
#include <sys/types.h>

typedef list<Module> * (*loadFunc) ();
typedef void (*unloadFunc) ();

CSAPI_TIER2 
PluginLibrary::PluginLibrary (string filename) {
  list<Module> * tmp;
  // remember filename
  name = filename;
  
  // open the library and get a handle to it
#if (defined (WIN32) || (_WIN32))
  hLib = LoadLibrary(filename.c_str());
  if (!hLib) {
    cerr << "MaaateA: error opening " << filename << ": " << endl;
    return;
  }
#else
  plib = dlopen (filename.c_str(), RTLD_NOW);
  if (!plib) {
    cerr << "MaaateA: error opening " << filename << ": " << dlerror() << endl;
    return;
  }
#endif

  // get the function to load the modules of the library and call it
  // returning a list of modules; the initialisation functions are
  // also called during construction of modules
    loadFunc loadModules;
#if (defined (WIN32) || (_WIN32))
    loadModules = (loadFunc)GetProcAddress(hLib, "loadModules");
    if (!loadModules) {
      cerr << "MaaateA: error reading symbol from " << filename << ": " << endl;
      return;
    }
#else
    loadModules = (loadFunc) dlsym (plib, "loadModules");
    if (!loadModules) {
      cerr << "MaaateA: error reading symbol from " << filename << ": " << endl;
      cerr << dlerror () << endl;
      return;
    }
#endif
    tmp = loadModules();
    modList = *tmp;
	
  // set the plib/hLib member of each module that was loaded
    list<Module>::iterator iter;
    for (iter = modList.begin(); iter != modList.end(); ++iter) {
      (*iter).plib = this;
    }
}

CSAPI_TIER2 
PluginLibrary::~PluginLibrary() {
  // cleanup the modules
  unloadFunc unloadModules;
 
#if (defined (WIN32) || (_WIN32))
  unloadModules = (unloadFunc)GetProcAddress((HMODULE)hLib, "unloadModules");
  unloadModules();

  // close the library
  FreeLibrary((HMODULE)hLib);
#else
  unloadModules = (unloadFunc) dlsym (plib, "unloadModules");
  unloadModules();
  
  // close the library
  dlclose(plib);
#endif
}

CSAPI_TIER2 void
Plugins::AddStaticModules () {
}

CSAPI_TIER2 bool
Plugins::AddLibrary (string filename) {
  // find library either at given place or 
  // in MAAATE_PATH environment variable path or
  // in standard PACKAGE_PLUGIN_DIR path or
  // in standard /usr/local/lib
  struct stat *buf = (struct stat *) malloc (sizeof(struct stat));
  string filepath = string(filename);

  // Check that filename ends in ".so"
  //  string::size_type pos = filename.rfind ('.');
  string::size_type pos = filename.rfind ('.');
  if (pos == string::npos) {
    cerr << "MaaateA: warning " << filename << " not opened" << endl;
    free(buf);
    return false;
  }
  string extension(filename, pos, string::npos);
#if (defined (WIN32) || (_WIN32))
  if (extension != ".dll") {
    cerr << "MaaateA: warning " << filename << " not opened" << endl;
    free(buf);
    return false;
  }
  
#else
  if (extension != ".so") {
    cerr << "MaaateA: warning " << filename << " not opened" << endl;
    free(buf);
    return false;
  }
#endif


  if (stat(filename.c_str(), buf) != 0) {
    // some error occured when opening file - presume file doesn't exist
    // => try with different paths
    char * mpath = getenv("MAAATE_PATH");
    string crudpath;
    if (mpath != NULL && strlen(mpath) != 0) 
    {
      crudpath.append(mpath);
      crudpath.append(PATHSEP);
    }

    crudpath.append(PACKAGE_PLUGIN_DIR);
#if (defined (WIN32) || (_WIN32))
#ifdef _DEBUG
	crudpath.append(";.\\debug");
#else
	crudpath.append(";.\\release");
#endif
#else
    crudpath.append(":/usr/local/lib/");
#endif

    // go through colon-separated directory list and try library in each
    // always cutting out the first name
    bool found = false;
    string tmpdir;
    string::size_type pos = 0;
    while ((pos=crudpath.find(PATHSEP, pos)) != string::npos) {
      tmpdir = string(crudpath, (string::size_type) 0, pos);
      if (stat((tmpdir+DIRSEP+filename).c_str(), buf) == 0) {
	filepath = (tmpdir+DIRSEP+filename);
	found = true;
	break;
      }
      crudpath.erase ((string::size_type) 0, pos+1);
    }
    if (!found) {
      // try last dirname
      if (stat((crudpath+DIRSEP+filename).c_str(), buf) == 0) {
	filepath = (crudpath)+DIRSEP+filename;
      } else {
	free(buf);
	return false;
      }
    }
  }
  
  PluginLibrary * pl = new PluginLibrary (filepath);
  list<Module> * ml = pl->Modules();

  list<Module>::iterator iter;
  for (iter = ml->begin(); iter != ml->end(); ++iter) {
    AddModule (&(*iter));
  }
  free(buf);
  return true;
}

CSAPI_TIER2 void
Plugins::AddLibraries (string dirname) {
  DIR * dir;
  struct dirent * direntity;
  
  dir = opendir(dirname.c_str());
  if (dir == NULL) return;

  /* read directory entries one at a time */
  while ((direntity=readdir(dir)) != NULL) {
    AddLibrary(direntity->d_name);
  }
}

CSAPI_TIER2 void
Plugins::AddLibrariesPath (string pathlist) {
  string crudpath = string(pathlist);
  string tmpdir;
  string::size_type pos = 0;

  // go through colon-separated directory list and add libraries from there
  // always cutting out the first name
  while ((pos=crudpath.find(':', pos)) != string::npos) {
    tmpdir = string(crudpath, (string::size_type) 0, pos);
    AddLibraries (crudpath);
    crudpath.erase ((string::size_type) 0, pos+1);
  }
  // add last dirname
  AddLibraries (crudpath);
  
}

CSAPI_TIER2 void
Plugins::AddLibrariesMaaatePath () {
  AddStaticModules ();

  char * mpath = getenv("MAAATE_PATH");
  string crudpath = string();
  if (mpath != NULL && strlen(mpath) != 0) {
    crudpath.append(mpath);
    crudpath.append(":");
  }
  crudpath.append(PACKAGE_PLUGIN_DIR);

  AddLibrariesPath (crudpath);
}

CSAPI_TIER2 void
Plugins::AddModule (Module * module) {
  list<Module>::iterator iter;
  
  for (iter = removedList.begin(); iter != removedList.end(); ++iter) {
    if (&(*iter) == module) {
      modList.splice(modList.begin(), removedList, iter);
      return;
    }
  }

  modList.push_back (*module);
}

CSAPI_TIER2 void
Plugins::RemoveLibrary (string name) {
  list<Module> * lM;

  lM = LibraryModules (name);

  list<Module>::iterator iter;

  for (iter = lM->begin(); iter != lM->end(); ++iter) {
    RemoveModule (&(*iter));
  }
}

CSAPI_TIER2 void
Plugins::RemoveModule (Module * module) {
  list<Module>::iterator iter;
  
  for (iter = modList.begin(); iter != modList.end(); ++iter) {
    if (&(*iter) == module) {
      removedList.splice(removedList.begin(), modList, iter);
      return;
    }
  }
}

// get a specific module of a given name
CSAPI_TIER2 Module *
Plugins::GetModule (string name) {
  list<Module>::iterator iter;

  for (iter = modList.begin(); iter != modList.end(); ++iter) {
    if ((*iter).name() == name) {
      return &(*iter);
    }
  }

  return NULL;
}

CSAPI_TIER2 list<Module> *
Plugins::LibraryModules (string name) {
  list<Module>::const_iterator iter;

  for (iter = modList.begin(); iter != modList.end(); ++iter) {
    if ((*iter).plib->filename() == name) {
      return (*iter).plib->Modules();
    }
  }
  return NULL;
}




