/*
    MPEG Maaate: An Australian MPEG audio analysis toolkit
    Copyright (C) 2000 Commonwealth Scientific and Industrial Research Organisation
    (CSIRO), Australia.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "segmentTable.H"

// copy constructor
CSAPI_TIER2 
SegmentTable::SegmentTable (const SegmentTable& st) {
  segments.reserve(st.segments.size());
  vector<SegmentData>::const_iterator iter;
  for (iter = st.segments.begin(); iter<st.segments.end(); iter++) {
    segments.push_back(*iter);
  }
}

// invert SegmentTable into new SegmentTable
// give it the endtime of the analysed file to get the last segment, too 
CSAPI_TIER2 SegmentTable *
SegmentTable::invertTable (double endTime) {
  SegmentTable * st = new SegmentTable();
  double startTime = 0.0;
  vector<SegmentData>::const_iterator iter;
  for (iter = segments.begin(); iter<segments.end(); iter++) {
    if (((*iter).start() - startTime) > 0.0) {
      st->append(SegmentData(startTime, (*iter).start()));
    }
    startTime = (*iter).end();
  }
  if (endTime > startTime) {
    st->append(SegmentData(startTime, endTime));
  }
  return st;
}


// print table
CSAPI_TIER2 void
SegmentTable::print() {
  vector<SegmentData>::iterator iter;
  int i=0;
  for (iter = begin(); iter < end(); iter++) {
    cout << "---------- ENTRY " << i++ << " ----------" << endl;
    cout << (*iter);
  }
}
  
 
// print table as plain "starttime endtime duration"
// negative = print not the segments itself but the spaces between
CSAPI_TIER2 void
SegmentTable::printPlain(bool negative) {
  vector<SegmentData>::iterator iter;
  double negStart = 0.0;
  for (iter = begin(); iter < end(); iter++) {
    if (negative) {
      if (((*iter).start() - negStart) > 0.0 ) {
	cout << negStart << " "
	     << (*iter).start() << " "
	     << ((*iter).start() - negStart) << endl;
	negStart = (*iter).end();
      }
    } else {
      cout << (*iter).start() << " "
	   << (*iter).end() << " "
	   << (*iter).duration() << endl;
    }
  }
  return;
}


// overload friend operator <<
// main usage: debugging
CSAPI_TIER2 ostream&
operator<<(ostream& os, SegmentTable& st) {
  // print all contained data
  vector<SegmentData>::const_iterator iter;
  int i=0;
  for (iter = st.begin(); iter < st.end(); iter++) {
    os << "---------- ENTRY " << i++ << " ----------" << endl;
    os << (*iter);
  }

  return os;
}


// operator =
CSAPI_TIER2 SegmentTable&
SegmentTable::operator= (const SegmentTable& st) {
  if (this == &st)
    return *this;
  segments.clear();
  segments.reserve(st.segments.size());
  vector<SegmentData>::const_iterator iter;
  for (iter = st.segments.begin(); iter<st.segments.end(); iter++) {
    segments.push_back(*iter);
  }
  return *this;
}


// query table
// returns the SegmentData that is at the specified time
// either EXACT (i.e. containing the time),
// the last one at or BEFORE the specified time,
// or the first one at or AFTER the specified time
// It is presumed that:
// - the segments are ordered according to time
// - only the first segment fulfilling the requirements is returned.
CSAPI_TIER2 SegmentData *
SegmentTable::getSegment (double time, QuerySegmentType which) {
  vector<SegmentData>::iterator iter;

  for (iter = begin(); iter < end(); iter++) {
    if (which == EXACT) {
      if (((*iter).start() <= time) && ((*iter).end() >= time)) {
	return &(*iter);
      }
    } else if (which == AFTER) {
      if ((*iter).end() > time) {
	return &(*iter);
      }
    } else if (which == BEFORE) {
      if ((*iter).start() >= time) {
	return &(*(--iter));
      }
    }
  }
  if (size() > 0 && which == BEFORE) {
    iter = end();
    return &(*(--iter));
  }
  
  return NULL;
}

// query table
CSAPI_TIER2 int
SegmentTable::getSegmentIndex (double time) {
  int ndx = 0;

  vector<SegmentData>::const_iterator iter;
  for (iter = begin(); iter < end(); iter++, ndx++) {
    if (((*iter).start() <= time) && ((*iter).end() >= time)) {
      return ndx;
    }
  }
    
  return -1;
}


// insert an SegmentData according to its start time
CSAPI_TIER2 vector<SegmentData>::iterator
SegmentTable::insert (const SegmentData& sd) {
  // check capacity before inserting
  if (segments.capacity() == segments.size()) {
    segments.reserve(segments.size()+10);
  }
  // find insertion place
  vector<SegmentData>::iterator iter;
  for (iter = begin(); iter < end(); iter++) {
    if ((sd.start() < (*iter).start()) ||
	((sd.start() == (*iter).start()) && 
	 (sd.duration() <= (*iter).duration()))) {
      // insert before element
      return segments.insert(iter, sd);
    } 
  }
  // not inserted yet, append at end
  segments.push_back(sd);
  return segments.end();
}
    

// insert an SegmentTable positioning entries at right places
CSAPI_TIER2 void
SegmentTable::insert (const SegmentTable& st) {
  segments.reserve(segments.size()+st.segments.size());
  vector<SegmentData>::const_iterator iter;
  for (iter = st.segments.begin(); iter < st.segments.end(); iter++) {
    insert((*iter));
  }
}

// append a SegmentData
CSAPI_TIER2 void
SegmentTable::append (const SegmentData& sd) {
  // check capacity before inserting and make space for 10 more
  if (segments.capacity() == segments.size()) {
    segments.reserve(segments.size()+10);
  }
  // append at end
  segments.push_back(sd);
  return;
}
    
// erases an SegmentData at position ndx
CSAPI_TIER2 bool
SegmentTable::erase (int ndx) {
  vector<SegmentData>::iterator iter=begin();
  if ((ndx<0) || (ndx>=((int) segments.size()))) {
    return false;
  }
  for (int i=0; iter<end(); i++, iter++) {
    if (i == ndx) {
      segments.erase(iter);
      return true;
    }
  }
  return false;
}
    

// delete table
CSAPI_TIER2 void
SegmentTable::clear () {
  vector<SegmentData>::iterator iter;

  for (iter = begin(); iter < end(); iter++) {
    segments.erase(iter);
    iter--;
  }
}
   

// calculate statistics over all segments
CSAPI_TIER2 double 
SegmentTable::avgDuration() {
  vector<SegmentData>::iterator iter;
  double sum=0.0;
  int number=0;

  for (iter = begin(); iter < end(); iter++) {
    sum+=(*iter).duration();
    number++;
  }

  if (number > 0) {
    return (sum/number);
  } else {
    return 0;
  }

}

// calculate statistics over all segments
CSAPI_TIER2 double
SegmentTable::minDuration() {
  vector<SegmentData>::iterator iter;
  double min=DBL_MAX;

  for (iter = begin(); iter < end(); iter++) {
    if ( (*iter).duration() < min ) {
      min = (*iter).duration();
    }
  }

  return min;
}


// calculate statistics over all segments
CSAPI_TIER2 double
SegmentTable::maxDuration() {
  vector<SegmentData>::iterator iter;
  double max=DBL_MIN;

  for (iter = begin(); iter < end(); iter++) {
    if ( (*iter).duration() > max ) {
      max = (*iter).duration();
    }
  }

  return max;
}
